/* flashFsLib.c */

/* Copyright 1984 - 2001 Wind River Systems, Inc. */
/* modification history */
/*-----------------------------------------------------------------------------
08/12/2005 pcd
1. Added to support Fujitsu MBM29LV160T flash chip. This code will be used for 
   MS7751RSE01 CPU board
-----------------------------------------------------------------------------*/
/*
DESCRIPTION

*  Configure the RFA (flash) device for use by dosFs file system.
*/
#include "types.h"
#include "vxWorks.h"
#include "stdio.h"
#include "stdlib.h"
#include "errno.h"

#include "dosFsLib.h"
#include "rawFsLib.h"
#include "fioLib.h"
#include "blkIo.h"
#include "string.h"
#include "errnoLib.h" 
#include "flashFsLib.h"


/* Driver debug control */
#define  FLASH_DEBUGFS

/* Driver debug control */
#ifdef FLASH_DEBUGFS
#define FLASHFS_DEBUG_OFF		0x0000
#define FLASHFS_DEBUG_ERROR     0x0001
#define FLASHFS_DEBUG_BLKWRITE	0x0002
#define FLASHFS_DEBUG_BLKREAD 	0x0004
#define FLASHFS_DEBUG_IOCTL		0x0008
#define FLASHFS_DEBUG_GETPHYS	0x0010
#define FLASHFS_DEBUG_INIT 	    0x0020
#define FLASHFS_DEBUG_ALL    	0xffff

LOCAL int flashFsDebug = ( FLASHFS_DEBUG_ERROR /*| FLASHFS_DEBUG_BLKWRITE */);

#define FS_DRV_PRINT(FLG, X)						\
    do {								\
    if (flashFsDebug & FLG)						\
	printf X;							\
    } while (0)

#else /* FLASH_DEBUGFS */

#define FS_DRV_PRINT(FLG, X)

#endif /* FLASH_DEBUGFS */


/*
 * The 4MB flash part (Fujitsu MBM29LV160T) is divided into three sections - rfa0/rawFs, flash/dosFs, and rfa2/dosFs.
 * /rfa0 begins at sector 0  and ends at sector 7.  stored bootrom, total size 1MByte,       address 0xA0000000
 * flash begins at sector 8  and ends at sector 23. stored vxWorks, total size 2MByte,       address 0x00000000
 * /rfa2 begins at sector 24 and ends at sector 30. stored NVM+,    total size (1M-64K)Byte, address 0x00000000 
 * bootLine begins at sector 31, ends at sector 30. stored BootLine,total size (64K)Byte,    address 0xA03F8000 
 */


STATUS flashFsLibInit();
STATUS flashFsBlkRead (FLASH_DEV *, int, int, char *);
STATUS flashFsBlkWrite(FLASH_DEV *, int, int, char *);
STATUS flashFsIoctl(BLK_DEV *, int, int);
STATUS flashFsSync();
STATUS flashFsGetPhys(int, int *, int *);


LOCAL STATUS flashFsRfaConfig (char *, ulong_t, ulong_t, ulong_t, ulong_t);
LOCAL BLK_DEV *flashDevCreate (UINT32, FAST int, int, int);


IMPORT STATUS dosFsConfigInit ( DOS_VOL_CONFIG *, char,	UINT8, short, char, UINT16,	short, UINT, UINT );
extern STATUS	inflate(char *src, char *dst, int src_size);

/******************************************************************************
*
* flashFsLibInit - Initializes flash files system for /rfa0, /rfa1, /rfa2 
*
* This routine calls flashDrvLibInit and then calls flashFsRfaConfig to setup 
* individual rfa sections on the flash.
*
* RETURNS: ERROR if flashDrvLibInit fails, else OK.
*/


STATUS flashFsLibInit()
{
    char *name0 = BOOTROM_BASE_FS;
    char *name1 = APPL_BASE_FS;
    
#ifdef INCLUDE_NVM_SPACE
    char *name2 = NVM_BASE_FS;
#endif /* INCLUDE_NVM_SPACE */


    if( flashDrvLibInit() == ERROR ) 
    {
		FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
		          ( "flashFsLibInit(): flashDrvLibInit() failed\n" ));

		return( ERROR );
    }

	/* AMD AM29LV033 */
    /* rawFS (bootrom) starting at sector 0 boot sector (1MByte) */
    flashFsRfaConfig(name0, 0, BOOT_SECTOR_BASE_ADDRESS-FLASH_BASE_ADDRESS, 
		FLASH_FS_DOS_SECTOR_SIZE, FLASH_RAW_NUM_BLK);

    /* dosFS1 (vxWorks) starting at sector 32(2MBytes) */
    flashFsRfaConfig(name1, 1, APPLICATION_SECTOR_BASE_ADDRESS-FLASH_BASE_ADDRESS, 
		FLASH_FS_DOS_SECTOR_SIZE, FLASH_FS2_NUM_BLK);   

#ifdef INCLUDE_NVM_SPACE

    /* dosFS2 (NVM+) starting at sector 16 (1MBytes) */
    flashFsRfaConfig(name2, 2, NVM_SECTOR_BASE_ADDRESS-FLASH_BASE_ADDRESS, 
	    FLASH_FS_DOS_SECTOR_SIZE, FLASH_FS1_NUM_BLK); 
#endif	

	FS_DRV_PRINT (FLASHFS_DEBUG_INIT,
                  ("flashFsLibInit(): Success.\n" ));

    return( 0 );
}


/******************************************************************************
*
* flashFsRfaConfig - Configure the RFA (flash) device for use by file system.
*
* This routine creates the flash devices.  For unit 0, the flash device is attached 
* to rawFs and for units 1 and 2, the device is attached to the dosFs and mounted.
* 
*
* RETURNS: OK if operation succeeds, ERROR otherwise.
*/

LOCAL STATUS flashFsRfaConfig
    (
    char    * name,
    ulong_t   unit,
    ulong_t   offset,
    ulong_t   sectorSize,
    ulong_t   sectorCnt
    )
{
    STATUS            rc;
    BLK_DEV         * pBlkDev;
    DOS_VOL_DESC    * pDosVolDesc;
    RAW_VOL_DESC    * pRawVolDesc;
    static BOOL       rfaMounted[NUM_RFA_DEVICES] = { 0 };



    if ( (unit >= NUM_RFA_DEVICES) || (rfaMounted[unit] == TRUE) )
        return OK;
    
    rc = OK;
    
	FS_DRV_PRINT (FLASHFS_DEBUG_INIT,
                  ("flashFsRfaConfig(): Initializing device (%s) ... offset 0x%x, sectorSize %d, sectorCnt %d\n", 
				    name, (int)offset, (int)sectorSize, (int)sectorCnt ));
    
    /* Mount the file system */
    pBlkDev = flashDevCreate(offset, sectorSize, sectorCnt, 0);

    if (pBlkDev == NULL)
    {
		FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
                  ("flashFsRfaConfig(): flashDevCreate Flash Device Failure.\n"));
        return ERROR;
    }
    else 
		FS_DRV_PRINT (FLASHFS_DEBUG_INIT,
                  ("flashFsRfaConfig(): flashDevCreate Success on Unit %d %s.\n", (int)unit, name));

    /* The zero device is treated as a RAW file system. */
    if (unit == 0)
    {
        pRawVolDesc = rawFsDevInit(name, pBlkDev);
        if ( pRawVolDesc == NULL )
        {
		    FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
                      ("flashFsRfaConfig(): %s:%d rawFsDevInit(%s) failed\n",
                    __FILE__, __LINE__, name));
            rc = -1;
        }
        else
        {       
		    FS_DRV_PRINT (FLASHFS_DEBUG_INIT,
	                  ("flashFsRfaConfig(): Unit %d %s Success.\n", (int)unit, name));
            rc = 0;
            rfaMounted[unit] = TRUE;
        }
        
        return rc;
 
    }
    /* unit != 0 */
    else 
    {
	    pDosVolDesc = dosFsMkfs(name,pBlkDev);
      
        if (pDosVolDesc == NULL)
			rc = ERROR;

		if ( rc == OK )
		{
			FS_DRV_PRINT (FLASHFS_DEBUG_INIT,
                  ("flashFsRfaConfig(): Unit %ld %s Success.\n", unit, name));
			rfaMounted[unit] = TRUE;
		}
		else
			FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
                  ("flashFsRfaConfig(): Unit %ld %s Failure.\n", unit, name));

		return rc;
    }  /* else if (unit != 0) */

    return -1;

}


/*******************************************************************************
*
* flashDevCreate - create a FLASH disk device
*
* This routine creates a FLASH disk device.
*
* Memory for the FLASH disk can be pre-allocated separately; if so, the
* <flashAddr> parameter should be the address of the pre-allocated device
* memory.  Or, memory can be automatically allocated with malloc() by
* setting <flashAddr> to zero.
*
* The <bytesPerBlk> parameter specifies the size of each logical block
* on the FLASH disk.  If <bytesPerBlk> is zero, 512 is used.
*
* The <nBlocks> parameter specifies the size of the disk, in blocks.
* If <nBlocks> is zero, a default size is used.  The default is calculated
* using a total disk size of either 51,200 bytes or one-half of the size
* of the largest memory area available, whichever is less.  This default
* disk size is then divided by <bytesPerBlk> to determine the number
* of blocks.
*
* The <blkOffset> parameter specifies an offset, in blocks, from the start
* of the device to be used when writing or reading the FLASH disk.  This
* offset is added to the block numbers passed by the file system during
* disk accesses.  (VxWorks file systems always use block numbers beginning
* at zero for the start of a device.)  This offset value is typically useful
* only if a specific address is given for <flashAddr>.  Normally, <blkOffset>
* is 0.
*
* FILE SYSTEMS:
* Once the device has been created, it must be associated with a name and a
* file system (dosFs, rt11Fs, or rawFs).  This is accomplished using the
* file system's device initialization routine or make-file-system routine,
* e.g., dosFsDevInit() or dosFsMkfs().  The flashDevCreate() call returns a
* pointer to a block device structure (BLK_DEV).  This structure contains
* fields that describe the physical properties of a disk device and specify
* the addresses of routines within the flashDrv driver.  The BLK_DEV structure
* address must be passed to the desired file system (dosFs, rt11Fs or rawFs)
* via the file system's device initialization or make-file-system routine.
* Only then is a name and file system associated with the device, making it
* available for use.
*
* EXAMPLE:
* In the following example, a 200-Kbyte FLASH disk is created with
* automatically allocated memory, 512-byte blocks, a single track, and no
* block offset.  The device is then initialized for use with dosFs and assigned
* the name "DEV1:":
* .CS
*     BLK_DEV *pBlkDev;
*     DOS_VOL_DESC *pVolDesc;
*
*     pBlkDev = flashDevCreate (0, 512,  400,  0);
*     pVolDesc = dosFsMkfs ("DEV1:", pBlkDev);
* .CE
* The dosFsMkfs() routine calls dosFsDevInit() with default parameters and
* initializes the file system on the disk by calling ioctl() with the
* \%FIODISKINIT function.
*
* If the FLASH disk memory already contains a disk image created elsewhere,
* the first argument to flashDevCreate() should be the address in memory, and
* the formatting parameters -- <bytesPerBlk>, <blksPerTrack>, <nBlocks>, and
* <blkOffset> -- must be identical to those used when the image was
* created.  For example:
* .CS
*     pBlkDev = flashDevCreate (0xc0000, 512, 400, 0);
*     pVolDesc = dosFsDevInit ("DEV1:", pBlkDev, NULL);
* .CE
* In this case, dosFsDevInit() must be used instead of dosFsMkfs(), because
* the file system already exists on the disk and should not be
* re-initialized.  This procedure is useful if a FLASH disk is to be created
* at the same address used in a previous boot of VxWorks.  The contents of
* the FLASH disk will then be preserved.
*
* These same procedures apply when creating a FLASH disk with rt11Fs using
* rt11FsDevInit() and rt11FsMkfs(), or creating a FLASH disk with rawFs using
* rawFsDevInit().
*
* RETURNS:
* A pointer to a block device structure (BLK_DEV) or NULL if memory cannot
* be allocated for the device structure or for the FLASH disk.
*
* SEE ALSO: dosFsMkfs(), dosFsDevInit(), rt11FsDevInit(), rt11FsMkfs(),
* rawFsDevInit()
*/

LOCAL BLK_DEV *flashDevCreate
    (
    UINT32	flashOffset,    /* where it is in flash */
    FAST int	bytesPerBlk,	/* number of bytes per block */
    int		nBlocks,	/* number of blocks on this device */
    int		blkOffset	/* no. of blks to skip at start of device */
    )
{
    UINT32        largestSect;
    FAST FLASH_DEV	*pFlashDev;	/* ptr to created FLASH_DEV struct */
    FAST BLK_DEV	*pBlkDev;	/* ptr to BLK_DEV struct in FLASH_DEV */
    
    pBlkDev     = NULL;
    pFlashDev   = NULL;	/* ptr to created FLASH_DEV struct */


	FS_DRV_PRINT (FLASHFS_DEBUG_INIT,
	          ("flashDevCreate(): flashOffset=0x%x, bytesPerBlk=%d, nBlocks=%d, blkOffset=0x%x\n",
		                                (int)flashOffset, (int)bytesPerBlk, (int)nBlocks, (int)blkOffset));

    /* Set up defaults for any values not specified */

    if (bytesPerBlk == 0)
	    bytesPerBlk = FLASH_FS_DOS_SECTOR_SIZE;


    /* Scan to flash device sectors to find the area reserved, and note the */
    /* largest sector to allocate the sector write cache */

    largestSect = FLASH_BANK_SECTOR_SIZE;

    /* If the requested device will fit in flash */ 
    if (largestSect)
    {
        /* Allocate a FLASH_DEV structure for device */
        pFlashDev = (FLASH_DEV *) calloc (1, sizeof (FLASH_DEV));
        
        /* If memory available, create device */
        if (pFlashDev != NULL)
        {

           /* Done in flashDrvInit */
            /* Allocate cache for writing to a flash sector */
          /* pFlashDev->flash_cache = (char *)malloc(largestSect); */

            
            /* Initialize BLK_DEV structure (in FLASH_DEV) */
            pBlkDev = &pFlashDev->flash_blkdev;
    
            pBlkDev->bd_nBlocks      = nBlocks;		/* number of blocks */
            pBlkDev->bd_bytesPerBlk  = bytesPerBlk;	/* bytes per block */
            pBlkDev->bd_blksPerTrack = nBlocks;	    /* blocks per track */
    
            pBlkDev->bd_nHeads       = 1;		    /* one "head" */
            pBlkDev->bd_removable    = FALSE;		/* not removable */
            pBlkDev->bd_retry	     = 1;		    /* retry count */
            pBlkDev->bd_mode	     = O_RDWR;		/* initial mode for device */
            pBlkDev->bd_readyChanged = TRUE;		/* new ready status */
    
            pBlkDev->bd_blkRd	     = flashFsBlkRead;	/* read block function */
            pBlkDev->bd_blkWrt	     = flashFsBlkWrite;	/* write block function */
            pBlkDev->bd_ioctl	     = flashFsIoctl;	/* ioctl function */
            pBlkDev->bd_reset	     = NULL;		/* no reset function */
            pBlkDev->bd_statusChk    = NULL;		/* no check-status function */
            
            /* Initialize remainder of device struct */
    
            pFlashDev->flash_offset = flashOffset;  /* Start offset */
            pFlashDev->flash_blkOffset = blkOffset;	/* block offset */
        }
    }
    else
		FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
                  ("flashFsRfaConfig(): %s:%d Sector Count invalid\n", __FILE__, __LINE__));

    return (pBlkDev);
}



/******************************************************************************
*
* flashFsBlkRead - Block read routine.
*
* This routine is shared by the RFAs.  Flash offset is used to distinguish reads  
* between the RFAs. 
*
* RETURNS: OK if operation succeeds, ERROR otherwise.
*/


STATUS flashFsBlkRead (FLASH_DEV *pFlashDev, int startBlk, int numBlks, char *pBuf)
{
    int	blkIndx, phySectorNum, offset;


	FS_DRV_PRINT (FLASHFS_DEBUG_BLKREAD,
		      ( "flashFsBlkRead(): startBlk = %d, numBlks = %d\n", (int)startBlk, (int)numBlks ));

    /* Add in the block offset */
    startBlk += pFlashDev->flash_blkOffset;

    /* Calculate flash offset */
    startBlk += (pFlashDev->flash_offset/FLASH_FS_DOS_SECTOR_SIZE);

	FS_DRV_PRINT (FLASHFS_DEBUG_BLKREAD,
		      ( "flashFsBlkRead(): pFlashDev->flash_blkOffset = 0x%x, pFlashDev->flash_offset = 0x%x, startBlk = 0x%x\n", 
			   (int)pFlashDev->flash_blkOffset, (int)pFlashDev->flash_offset, (int)startBlk ));


    for( blkIndx = 0; blkIndx < numBlks; blkIndx++ ) 
	{
		if( flashFsGetPhys( (startBlk + blkIndx), &phySectorNum, &offset ) == ERROR ) {
			FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
	              ( "flashFsBlkRead(): flashFsGetPhys() failed\n" ));
			return( ERROR );
		}

	    FS_DRV_PRINT (FLASHFS_DEBUG_BLKREAD,
		      ( "flashFsBlkRead(): phySectorNum = 0x%x, offset = 0x%x\n", (int)phySectorNum, (int)offset ));

		/*
		d ( (void *) (FLASH_BASE_ADDRESS + (phySectorNum * FLASH_BANK_SECTOR_SIZE) + offset), FLASH_FS_DOS_SECTOR_SIZE, 1);
		*/

		if( flashBlkRead( phySectorNum, pBuf, offset, FLASH_FS_DOS_SECTOR_SIZE ) == ERROR ) {
			FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
	              ( "flashFsBlkRead(): flashBlkRead() failed\n" ));
			return( ERROR );
		}

		/*
		d ( (void *) pBuf, FLASH_FS_DOS_SECTOR_SIZE, 1);
		*/

		pBuf += FLASH_FS_DOS_SECTOR_SIZE;

    }

    return( OK );
   
}


/******************************************************************************
*
* flashFsBlkWrite - Block write routine.
*
* This routine is shared by the RFAs.  Flash offset is used to distinguish writes  
* between the RFAs.  flashSyncFilesystem is called at the exit of the function to 
* flush the flash buffer to flash part. 
*
* RETURNS: OK if operation succeeds, ERROR otherwise.
*/


STATUS flashFsBlkWrite(FLASH_DEV  *pFlashDev, int startBlk, int numBlks, char *pBuf)
{
    int	blkIndx, phySectorNum, offset;


	FS_DRV_PRINT (FLASHFS_DEBUG_BLKWRITE,
	          ( "flashFsBlkWrite(): startBlk = %d, numBlks = %d\n",	(int)startBlk, (int)numBlks ));

    /* Add in the block offset */
    startBlk += pFlashDev->flash_blkOffset;
    
    /* Calculate flash offset */
    startBlk += (pFlashDev->flash_offset/FLASH_FS_DOS_SECTOR_SIZE);

	FS_DRV_PRINT (FLASHFS_DEBUG_BLKWRITE,
		      ( "flashFsBlkWrite(): pFlashDev->flash_blkOffset = 0x%x, pFlashDev->flash_offset = 0x%x, startBlk = 0x%x\n", 
			   (int)pFlashDev->flash_blkOffset, (int)pFlashDev->flash_offset, (int)startBlk ));

 
    for( blkIndx = 0; blkIndx < numBlks; blkIndx++ ) 
	{
		if( flashFsGetPhys( (startBlk + blkIndx), &phySectorNum, &offset ) == ERROR ) {

			FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
	              ( "flashFsBlkWrite(): flashFsGetPhys() failed. the startblk is %x, the blkindex is %x\n", (int)startBlk, (int)blkIndx));
			return( ERROR );
		}

		/*
		d ( (void *) pBuf, FLASH_FS_DOS_SECTOR_SIZE, 1);
		*/

		if( flashBlkWrite( phySectorNum, pBuf, offset, FLASH_FS_DOS_SECTOR_SIZE ) == ERROR ) {
			FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
	              ( "flashFsBlkWrite(): flashBlkWrite() failed\n" ));
			return( ERROR );
		}

	    FS_DRV_PRINT (FLASHFS_DEBUG_BLKWRITE,
		      ( "flashFsBlkWrite(): phySectorNum = 0x%x, offset = 0x%x\n", (int)phySectorNum, (int)offset ));

		/*
		d ( (void *) (FLASH_BASE_ADDRESS + (phySectorNum * FLASH_BANK_SECTOR_SIZE) + offset), FLASH_FS_DOS_SECTOR_SIZE, 1);
		*/

		pBuf += FLASH_FS_DOS_SECTOR_SIZE;

    }

	flashSyncFilesystem ();
    return( OK );
}



/******************************************************************************
*
* flashFsIoctl - Ioctl routines.
*  
* Currently disk format and files sytem sync are supported. 
*
* RETURNS: OK.
*/


STATUS flashFsIoctl(BLK_DEV *pDev, int funcCode, int arg)
{

	FS_DRV_PRINT (FLASHFS_DEBUG_IOCTL,
	          ( "flashFsBlkIoctl(): called, funCode=%d\n",  (int)funcCode));

    switch (funcCode)
    {
		case FIODISKFORMAT:
			flashEraseBank();
			FS_DRV_PRINT (FLASHFS_DEBUG_IOCTL,
                  ("flashFsBlkIoctl(): I am here erasing awayt\n"));
			break;
   
		case FLASH_FS_SYNC:
			flashSyncFilesystem();
			break;

		default:
			FS_DRV_PRINT (FLASHFS_DEBUG_ERROR,
                  ("flashFsBlkIoctl(): UNKNOW function code\n"));
			errnoSet( S_ioLib_UNKNOWN_REQUEST );
			return( ERROR );
			break;
    }

    return( OK );
}




/* sal lib needs it */

STATUS flashFsSync()
{
 flashSyncFilesystem();
 return 0;
}



/******************************************************************************
*
* flashFsGetPhys - Determines the physical sector and offset.
*
* This routine takes the blknum from dosFs and translates it into physical sector
* number and offset.  The returned value is used by flashBlkWrite.
*
* RETURNS: OK.
*/


STATUS flashFsGetPhys( int blkNum, int *sectorNum, int *offset )
{

    *sectorNum = (blkNum / (FLASH_BANK_SECTOR_SIZE / FLASH_FS_DOS_SECTOR_SIZE));
    *offset = ((blkNum % (FLASH_BANK_SECTOR_SIZE / FLASH_FS_DOS_SECTOR_SIZE)) \
				* FLASH_FS_DOS_SECTOR_SIZE);

	FS_DRV_PRINT (FLASHFS_DEBUG_GETPHYS,
		      ("flashFsGetPhys(): blkNum = %d, *sectorNum = %d, *offset = 0x%08X\n",
				(int)blkNum, (int)*sectorNum, (int)*offset ));

    return( OK );
}


/*---------------------------------------------------------------------------*/

/*****************************************************************************
* netFlash - Download bootrom_uncmp.bin from the network FTP server into flash.
*/
#define FLASH_BUF_SIZE		(FLASH_RAW_NUM_BLK * FLASH_FS_DOS_SECTOR_SIZE)

STATUS loadBoot
(
    char *hostName,
    char * fileName
)
{
    int             fd = -1;
    int             errFd = -1;
    FILE           *fp = 0;
    int             i, bootImageSize=0;
    char           *flashBuf = 0;


    if ((flashBuf = malloc(FLASH_BUF_SIZE)) == 0) 
    {
	   printf("\nNot enough memory\n");
	   goto fail;
    }

    /* 
     * Using tftp to get boot image 
     */    
    printf ("Loading Boot ROM '%s'... \n", fileName);

#ifdef INCLUDE_TFTP_CLIENT
    if (tftpXfer (hostName, 0, fileName, "get", "binary", &fd,
		      &errFd) != OK)
    {
	    printf("\nTFTP Error occurred\n");
        goto fail;
    }
#endif

    if ((fp = fdopen(fd, "r")) == 0) 
    {
	   printf("\nCould not open fd\n");
	   goto fail;
    }

    /* 
     * Read boot image data
     */    

    /* file is *.bin */
    if ((bootImageSize = fread(flashBuf, sizeof (char), FLASH_BUF_SIZE, fp)) == 0 ) 
    {
	   printf("\nFailed reading from binary file %s\n", fileName);
	   goto fail;

    }
    printf("\nTFTP Transfered %d Bytes\n", (int)bootImageSize);

    fclose (fp);       /* closes fd */
    close (errFd);

    
    /* 
     * Save to flash 
     */    

    printf ("Saving Boot ROM into flash.\n");

    for (i=0; (i < FLASH_RAW_NUM_SEC) & ( bootImageSize > 0 ); i++)
    {
        printf("\nErasing boot sector %d ...", (BOOT_BEGIN_SECTOR+i) );
        flashEraseSector(BOOT_BEGIN_SECTOR + i);
        printf("\nWriting boot data to sector %d...", (BOOT_BEGIN_SECTOR+i) );
        taskDelay (sysClkRateGet ());
        flashBlkWrite(BOOT_BEGIN_SECTOR+i, flashBuf+i*FLASH_BANK_SECTOR_SIZE,0,FLASH_BANK_SECTOR_SIZE);
        bootImageSize -= FLASH_BANK_SECTOR_SIZE;
    }
    
    printf("\nDone\n");
    
    return (OK);


 fail:
    if (fp)
	   fclose(fp);		/* closes fd */
    else if (fd >= 0)
	   close (fd);
    if (errFd >= 0)
	   close (errFd);
    if (flashBuf)
	   free(flashBuf);

    return (ERROR);

} /* netFlash() */



/*********************************************************************************
* loadApp - Download Application Image from the network FTP server into flash.
*/
#define MAX_APP_SIZE    (2024 * 1024) /* FLASH_FS2_NUM_SEC * FLASH_BANK_SECTOR_SIZE) */   /* 1MB */

STATUS loadApp
(
    char * host,
    char * fileName
)
{
    int             fd = -1;
    int             errFd = -1;
    int             fd_flash = -1;
    char            flashAppFileName[128];
    char           *flashBuf = 0, *pflashBuf;          /* buffer to hold image data from TFTP */
	ulong_t         appImageSize=0, flashAppImageSize; /* size of image data from TFTP */
	char            buf[1024];
	ulong_t         writeLen=0;
	int  		    ret, tot, twirl;



    if ((flashBuf = malloc(MAX_APP_SIZE)) == 0) {
	   printf("\nNot enough memory\n");
	   goto fail;
    }

    /* 
     * Using tftp to get Application image 
     */    
    printf ("Loading Application image '%s' from host IP %s ...", fileName, host);

#ifdef INCLUDE_TFTP_CLIENT
    if (tftpXfer (host, 0, fileName, "get", "binary", &fd, &errFd) != OK)
    {
		printf("\nTFTP Error occurred\n");
        goto fail;
    }
#endif

    /* 
     * Read Application image data
     */    
	tot = 0;
	twirl = 0;
    pflashBuf = flashBuf;
    printf ("\nRead Application image '%s' to buf 0x%08x...", fileName, (int)pflashBuf);
	while ((ret = read(fd, buf, sizeof(buf))) > 0)
	{
		if( (tot + ret) > MAX_APP_SIZE )
		{
			printf("File size (%d) larger than (%d):\n", tot+ret, MAX_APP_SIZE);
	        goto fail;
		}
		if ((twirl++ & 0x0f) == 0)
			printf(".");

		bcopy(buf, pflashBuf, ret);
		pflashBuf += ret;
		tot += ret;
	}
	printf("\n");
	appImageSize =tot;

    if ( appImageSize <= 0)
    {
        printf("File size (%d) less than zero.\n", tot+ret);
	    goto fail;
	}

	close(fd);

	memset(buf, 0, sizeof(buf));
	if ( read(errFd, buf, sizeof(buf)) > 0 )
	{
		printf("TFTP Error occurred:%s\n", (char *)buf);
	    goto fail;
	}
	else 
		printf("Transfered %d Bytes\n", (int)appImageSize);

    /* 
     * Save to flash 
     */    

    /* get Application file name in flash */
    memset(flashAppFileName, 0, sizeof(flashAppFileName) );
    strcpy( flashAppFileName, VX_FILENAME );

    pflashBuf = flashBuf;
    flashAppImageSize = appImageSize;
    if ((fd_flash = open(flashAppFileName, O_WRONLY | O_CREAT, 0644)) != ERROR)
    {
        printf ("Saving Application image into flash '%s' ...", flashAppFileName);
    
        while ( appImageSize > 0 ) 
        {
            writeLen = FLASH_SECTOR_SIZE < appImageSize ? FLASH_SECTOR_SIZE : appImageSize;
            if ( ( ret = write(fd_flash, pflashBuf, writeLen) ) != writeLen)
            {
			    printf("\nWrite to Flash error, written length (%d) is not equal to write length (%d):", 
                        (int)ret, (int)writeLen);
	            goto fail;
            }

		    pflashBuf += ret;
            appImageSize -= ret;

			printf("\nwritten length (%d), remain length (%d) ...", (int)ret, (int)appImageSize);
        }
	    printf("\n");

        close(fd_flash);
    }
    else
    {
        printf("%s:%d open %s failed Errno %08x\n",
                    __FILE__, __LINE__, flashAppFileName, errnoGet());
	    goto fail;
    }

    printf("Done\n");

    /* Display Application image */
#if (0)
    printf("\n");
    pflashBuf = flashBuf;
	d ( (void *) pflashBuf, flashAppImageSize, 4);
#endif
    
    if (fd >= 0)
	   close (fd);
    if (fd_flash >= 0)
	   close (fd_flash);
    if (errFd >= 0)
	   close (errFd);
    if (flashBuf)
	   free(flashBuf);

    return (OK);


 fail:
    if (fd >= 0)
	   close (fd);
    if (fd_flash >= 0)
	   close (fd_flash);
    if (errFd >= 0)
	   close (errFd);
    if (flashBuf)
	   free(flashBuf);

    return (ERROR);

} /* loadApp() */



#if (1)
#define DECOMP_BUF_SIZE		0x280000               /* 2.5MB */
#endif
#define COMP_BUF_SIZE		(DECOMP_BUF_SIZE / 3)

STATUS decompressFile(int zfd)
{
    char		*imageBuf = NULL;
    char		*compBuf = NULL;
    int			fd = -1;
    int			rv = ERROR;
    int			compSize, r;


    /*
     * Read comprssed file to buffer 'compBuf'.
     */

    if (((char *)compBuf = malloc(COMP_BUF_SIZE)) == NULL) 
    {
	    printf("Can't alloc enough memory for image buffer\n");
	    goto done;
    }

    compSize = 0;

    while ((r = read(zfd,	/* Read loop required to support network */
		     compBuf + compSize,
		     COMP_BUF_SIZE - compSize)) > 0)
	compSize += r;

    if (r < 0) 
    {
	    printf("Read failed: errno = %d\n", errnoGet());
	    goto done;
    }

    if (compSize == COMP_BUF_SIZE) 
    {
	    printf("Compressed image too large\n");
	    goto done;
    }

	printf("Sucessful read comprssed file to compBuf 0x%08x, size %d\n", (int)compBuf, compSize );

    /*
     * Decomprssed image from buffer 'compBuf', and put into buffer 'imageBuf'.
     */

    if (((char *)imageBuf = malloc(DECOMP_BUF_SIZE)) == NULL) 
    {
	    printf("Can't alloc enough memory for decompression buffer\n");
	    goto done;
    }

    printf("Uncompressing %d bytes to imageBuf 0x%08x... ", compSize, (int)imageBuf);

    if ((r = inflate(compBuf, imageBuf, compSize)) < 0) 
    {
	    printf("\nUncompress failed\n");
	    goto done;
    }

    printf("\nUncompress done" );


    printf("\n");

    rv = OK;

 done:
    if (fd >= 0)
	close(fd);
    if (imageBuf)
	cfree(imageBuf);
    if (compBuf)
	cfree(compBuf);

    return rv;

} /* decompressFile() */

  
/******************************************************************************
*
* flashLoad - load a vxWorks image from FLASH memory
*
* RETURNS: OK, or ERROR if file can not be loaded.
*/

STATUS flashLoadApp
(
    char *fileName
)
{
    int			fd;

    /*
     * Load the boot file.
     * Support loading deflated files.
     */

    if ((fd = open(fileName, O_RDONLY, 0)) == ERROR) 
    {
	    printf("Cannot open \"%s\".\n", fileName);
	    return ERROR;
    }

	printf("Loading %s  %d... \n", fileName, fd);
    /* if (DEFLATED_FILE(fileName))  */
    {
	    if (decompressFile( fd ) != OK) 
        {
	        /* Error message already printed */
	        close(fd);
	        return ERROR;
	    }
    } 
    printf("Done\n");
    close(fd);

    return OK;

} /* flashLoad() */


