/* ks884xEnd.c - END style Micrel KS8841 or KS8842 PCI Bus Ethernet driver */

/* Copyright 1984-2002 Wind River Systems, Inc., and Cetia Inc. */

#ifdef KS_PCI_BUS

#include "copyright_wrs.h"

/*
modification history
--------------------
23Jul04,pcd  Create for Micrel KS8841 or KS8842 PCI Bus Ethernet driver.
*/

/*
DESCRIPTION
This module implements the Micrel KS8841, or KS8842 PCI Ethernet 32-bit 
network interface driver.

The PCI ethernet controller is inherently little-endian because
the chip is designed to operate on a PCI bus which is a little-endian
bus. The software interface to the driver is divided into three parts.
The first part is the PCI configuration registers and their set up. 
This part is done at the BSP level in the various BSPs which use this
driver. The second and third part are dealt with in the driver. The second
part of the interface is comprised of the I/O control registers and their
programming. The third part of the interface is comprised of the descriptors
and the buffers. 

This driver is designed to be moderately generic, operating unmodified
across the range of architectures and targets supported by VxWorks.  To
achieve this, the driver must be given several target-specific parameters,
and some external support routines must be provided. These target-specific
values and the external support routines are described below.

This driver supports multiple units per CPU.  The driver can be
configured to support big-endian or little-endian architectures.  It
contains error recovery code to handle known device errata related to DMA
activity. 

Some big-endian processors may be connected to a PCI bus through a
host/PCI bridge which performs byte swapping during data phases.  On such
platforms, the PCnet-PCI controller need not perform byte swapping during a
DMA access to memory shared with the host processor.

BOARD LAYOUT
This device is on-board.  No jumpering diagram is necessary.  

EXTERNAL INTERFACE
The drvier provides one standard external interface, ks884xEndLoad().  As
input, this routine takes a string of colon-separated parameters.  The
parameters should be specified in hexadecimal (optionally preceded by '0x'
or a minus sign '-').  The parameter string is parsed using strtok_r().

TARGET-SPECIFIC PARAMETERS
The format of the parameter string is:

<unit:devMemAddr:devIoAddr:pciMemBase:vecNum:intLvl:
memAdrs:memSize:memWidth:boardBusEndianMode:offset:flags>

\is
\i <unit>
The unit number of the device.  Unit numbers start at zero and increase for
each device controlled by the same driver.  The driver does not use this
value directly.  The unit number is passed through the MUX API where it is
used to differentiate between multiple instances of a particular driver.

\i <devMemAddr>
This parameter is the memory mapped I/O base address of the device registers
in the memory map of the CPU.  The driver will locate device registers as
offsets from this base address.
This is a required parameter since ks884x only support memory mapping.

\i <devIoAddr>
This parameter specifies the I/O base address of the device registers in the
I/O map of some CPUs. 
This parameter is no longer used, but is retained so that the
load string format will be compatible with legacy initialization routines.
The driver will always use memory mapped I/O registers specified via the
<devMemAddr> parameter.

\i <pciMemBase>
This parameter is the base address of the host processor memory as seen
from the PCI bus. This parameter is zero for most Intel architectures.

\i <vecNum>
This parameter is the vector associated with the device interrupt.
This driver configures the PCnet device to generate hardware interrupts
for various events within the device; thus it contains
an interrupt handler routine.  The driver calls pciIntConnect() to connect
its interrupt handler to the interrupt vector generated as a result of the
PCnet interrupt.

\i <intLvl>
Some targets use additional interrupt controller devices to help organize
and service the various interrupt sources.  This driver avoids all
board-specific knowledge of such devices.  During the driver's
initialization, the external routine sysKs884xIntEnable() is called to
perform any board-specific operations required to allow the servicing of a
PCnet interrupt.  For a description of sysKs884xIntEnable(), see "External
Support Requirements" below.

\i <memAdrs>
This parameter gives the driver the memory address to carve out its
buffers and data structures. If this parameter is specified to be
NONE then the driver allocates cache coherent memory for buffers
and descriptors from the system memory pool.
The PCnet device is a DMA type of device and typically shares access to
some region of memory with the CPU.  This driver is designed for systems
that directly share memory between the CPU and the PCnet.  It
assumes that this shared memory is directly available to it
without any arbitration or timing concerns.

\i <memSize>
This parameter can be used to explicitly limit the amount of shared
memory (bytes) this driver will use.  The constant NONE can be used to
indicate no specific size limitation.  This parameter is used only if
a specific memory region is provided to the driver.

\i <memWidth>
Some target hardware that restricts the shared memory region to a
specific location also restricts the access width to this region by
the CPU.  On these targets, performing an access of an invalid width
will cause a bus error.

This parameter can be used to specify the number of bytes of access
width to be used by the driver during access to the shared memory.
The constant NONE can be used to indicate no restrictions.

Current internal support for this mechanism is not robust; implementation 
may not work on all targets requiring these restrictions.

\i <boardBusEndianMode>
This parameter specifies big-endian and little-endian modes of operation.  
When big-endian mode is selected, the PCnet-PCI controller will swap the
order of bytes on the AD bus during a data phase on access to the FIFOs
only:  AD[31:24] is byte 0, AD[23:16] is byte 1, AD[15:8] is byte 2 and
AD[7:0] is byte 3.  In order to select the big-endian mode, set this
parameter to (1).  Most implementations, including natively
big-endian host architectures, should set this parameter to (1) in
order to select little-endian access to the FIFOs, as the driver is
currently designed to perform byte swapping as appropriate to the host
architecture.

\i <offset>
This parameter specifies a memory alignment offset.  Normally this
parameter is zero except for architectures which can only access 32-bit
words on 4-byte aligned address boundaries.  For these architectures the
value of this offset should be 2.

\i <flags>
This is parameter is used for future use.  Currently its value should be
zero.
\ie

EXTERNAL SUPPORT REQUIREMENTS
This driver requires five externally defined support functions that can
be customized by modifying global pointers.  The function pointer types
and default "bindings" are specified below.  To change the defaults, the
BSP should create an appropriate routine and set the function pointer
before first use.  This would normally be done within sysHwInit2().

Note that all of the pointers to externally defined functions <must> be
set to a valid executable code address.  Also, note that
sysKs884xIntEnable(), sysKs884xIntDisable(), and sysKs884xEnetAddrGet()
must be defined in the BSP.  This was done so that the driver would be
compatible with initialization code and support routines in existing BSPs.

The function pointer convention has been introduced to facilitate future
driver versions that do not explicitly reference a named BSP-defined
function.  Among other things, this would allow a BSP designer to define,
for example, one endIntEnable() routine to support multple END drivers.

\is
\i 'ks884xIntConnect'
\cs
    IMPORT STATUS (* ks884xIntConnect)
        (
        VOIDFUNCPTR * vector,     /@ interrupt vector to attach to     @/
        VOIDFUNCPTR   routine,    /@ routine to be called              @/
        int           parameter   /@ parameter to be passed to routine @/
        );

    /@ default setting @/

    ks884xIntConnect = pciIntConnect;
\ce
The ks884xIntConnect pointer specifies a function used to connect the
driver interrupt handler to the appropriate vector.  By default it is
the <pciIntLib> routine pciIntConnect().

\i 'ks884xIntDisconnect'
\cs
    IMPORT STATUS (* ks884xIntDisconnect)
        (
        VOIDFUNCPTR * vector,     /@ interrupt vector to attach to     @/
        VOIDFUNCPTR   routine,    /@ routine to be called              @/
        int           parameter   /@ routine parameter                 @/
        );

    /@ default setting @/

    ks884xIntDisconnect = pciIntDisconnect2;
\ce
The ks884xIntDisconnect pointer specifies a function used to disconnect
the interrupt handler prior to unloading the driver.  By default it is
the <pciIntLib> routine pciIntDisconnect2().

\i 'ks884xIntEnable'
\cs
    IMPORT STATUS (* ks884xIntEnable)
        (
        int level                 /@ interrupt level to be enabled @/
        );

    /@ default setting @/

    ks884xIntEnable = sysKs884xIntEnable;
\ce
The ks884xIntEnable pointer specifies a function used to enable the
interrupt level for the END device.  It is called once during
initialization.  By default it is a BSP routine named sysKs884xIntEnable(). 
The implementation of this routine can vary between architectures, and
even between BSPs for a given architecture family.  Generally, the
parameter to this routine will specify an interrupt <level> defined for an
interrupt controller on the host platform.  For example, MIPS and PowerPC
BSPs may implement this routine by invoking the WRS intEnable() library
routine.  WRS Intel Pentium BSPs may implement this routine via
sysIntEnablePIC().

\i 'ks884xIntDisable'
\cs
    IMPORT STATUS (* ks884xIntDisable)
        (
        int level                 /@ interrupt level to be disabled @/
        );

    /@ default setting @/

    ks884xIntDisable = sysKs884xIntDisable;
\ce
The ks884xIntDisable pointer specifies a function used to disable the
interrupt level for the END device.  It is called during stop.  By default
it is a BSP routine named sysKs884xIntDisable().  The implementation of
this routine can vary between architectures, and even between BSPs for a
given architecture family.  Generally, the parameter to this routine will
specify an interrupt <level> defined for an interrupt controller on the
host platform.  For example, MIPS and PowerPC BSPs may implement this
routine by invoking the WRS intDisable() library routine.  WRS Intel
Pentium BSPs may implement this routine via sysIntDisablePIC().

\i 'ks884xEnetAddrGet'
\cs
    IMPORT STATUS (* ks884xEnetAddrGet)
        (DRV_CTRL * pDrvCtrl, char * pStationAddr);

\ce
The ks884xEnetAddrGet pointer specifies a function used to get the Ethernet
(IEEE station) address of the device.  
\ie

SYSTEM RESOURCE USAGE
When implemented, this driver requires the following system resources:

    - one mutual exclusion semaphore
    - one interrupt vector
    - 14240 bytes in text for a PENTIUM3 target
    - 120 bytes in the initialized data section (data)
    - 0 bytes in the uninitialized data section (BSS)

The driver allocates clusters of size 1520 bytes for receive frames and
and transmit frames.

SEE ALSO: muxLib, endLib, netBufLib

\tb "Network Protocol Toolkit User's Guide"

\tb "PCnet-PCI II Single-Chip Full-Duplex Ethernet Controller
for PCI Local Bus Product"

\tb "PCnet-FAST Single-Chip Full-Duplex 10/100 Mbps Ethernet Controller
for PCI Local Bus Product"

definition TEST_THROUGHPUT        
To test throughput from ks884x chip. In the packets receive routine ks884xRecv(), 
drop the packets right away (not pass to protocol stack) when received packets.

definition BRINGUP_BOARD
For debug purpose, just to bring up the demo board, without software initilize 
the device. All the ks884x device registers are the hardware default value.

definition NOTCHECK_CHIPID
For debug purpose, don't check the ks884x device ID.

*/

#include "vxWorks.h"
#include "taskLib.h"
#include "stdio.h"
#include "stdlib.h"
#include "sysLib.h"
#include "intLib.h"
#include "iv.h"
#include "netLib.h"
#include "inetLib.h"

#ifdef WR_IPV6
#include "adv_net.h"
#endif /*WR_IPV6*/

#define END_MACROS

#include "etherMultiLib.h"
#include "endLib.h"
#include "ks884xEnd.h"
#include "hardware.h"
#include "hw_ver.h"
#include "ks_ver.h"
#include "ks_ver_put.h"


#undef   BRINGUP_BOARD          /* Just to bring up the demo board, without init the device */

/* defines */

/*
 * If LN_KICKSTART_TX is TRUE the transmitter is kick-started to force a
 * read of the transmit descriptors, otherwise the internal polling (1.6msec)
 * will initiate a read of the descriptors.  This should be FALSE if there
 * is any chance of memory latency or chip accesses detaining the PCnet DMA,
 * which results in a transmitter UFLO error.  This can be changed with the
 * global lnKickStartTx below.
 */

#define LN_KICKSTART_TX TRUE


/* board level/bus specific and architecture specific macros */

#if _BYTE_ORDER==_BIG_ENDIAN
#   define PCI_SWAP(x)  (( (pDrvCtrl)->boardBusEndianMode) ? LONGSWAP((int)(x)) : (int)(x) ) 
#else
#   define PCI_SWAP(x)  (x)
#endif


/* Cache macros */

#define KS_CACHE_INVALIDATE(address, len) \
        CACHE_DRV_INVALIDATE (&pDrvCtrl->cacheFuncs, (address), (len))

#define KS_CACHE_FLUSH(address, len)				    \
        CACHE_DRV_FLUSH (&pDrvCtrl->cacheFuncs, (address), (len))

#define KS_CACHE_USER_FLUSH(address, len)				    \
        CACHE_USER_FLUSH((address), (len))

#define KS_CACHE_VIRT_TO_PHYS(address) \
        CACHE_DRV_VIRT_TO_PHYS (&pDrvCtrl->cacheFuncs, (address))

#define KS_CACHE_PHYS_TO_VIRT(address) \
        CACHE_DRV_PHYS_TO_VIRT (&pDrvCtrl->cacheFuncs, (address))

#define END_FLAGS_ISSET(pEnd, setBits) \
            ((pEnd)->flags & (setBits))

/* memory to PCI address translation macros */

#define PCI_TO_MEM_PHYS(pciAdrs) \
    ((UINT)(pciAdrs) - (pDrvCtrl->pciMemBase))

#define MEM_TO_PCI_PHYS(memAdrs) \
    ((UINT)(memAdrs) + (pDrvCtrl->pciMemBase))
    
#define SYS_OUT_LONG(pDrvCtrl, addr, value)     \
    if (pDrvCtrl->flags & KS_MODE_MEM_IO_MAP)   \
        *((volatile UINT32 *)(addr)) = (value); \
    else                                        \
        sysOutLong ((int)(addr), (value));      \

#define SYS_IN_LONG(pDrvCtrl, addr, data)       \
    if (pDrvCtrl->flags & KS_MODE_MEM_IO_MAP)   \
        ((data) = *((volatile UINT32 *)(addr)));\
    else                                        \
        ((data) = sysInLong ((int) (addr)));    \


/* A shortcut for getting the hardware address from the MIB II stuff. */

/* RFC 1213 (old) */

#define END_HADDR(pEnd) \
    ((pEnd)->mib2Tbl.ifPhysAddress.phyAddress)

#define END_HADDR_LEN(pEnd) \
    ((pEnd)->mib2Tbl.ifPhysAddress.addrLength)

/* RFC 2233 (new - not implemented in this driver yet)
 *
 * #define END_HADDR(pEnd) \
 * ((pEnd)->pMib2Tbl->m2Data.mibIfTbl.ifPhysAddress.phyAddress)
 *
 * #define END_HADDR_LEN(pEnd) \
 * ((pEnd)->pMib2Tbl->m2Data.mibIfTbl.ifPhysAddress.addrLength)
 */

/*** END MACROS ***/	 
#define END_M2_INUCAST(pEnd)			(pEnd)->mib2Tbl.ifInUcastPkts++
#define END_M2_INNUCAST(pEnd)		    (pEnd)->mib2Tbl.ifInNUcastPkts++
#define END_M2_INERRORS(pEnd)		    (pEnd)->mib2Tbl.ifInErrors++
#define END_M2_INDISCARDS(pEnd)		    (pEnd)->mib2Tbl.ifInDiscards++
#define END_M2_INOCTETS(pEnd,bytes)	    (pEnd)->mib2Tbl.ifInOctets += bytes
#define END_M2_OUTUCAST(pEnd)		    (pEnd)->mib2Tbl.ifOutUcastPkts++
#define END_M2_OUTNUCAST(pEnd)		    (pEnd)->mib2Tbl.ifOutNUcastPkts++
#define END_M2_OUTDISCARDS(pEnd)		(pEnd)->mib2Tbl.ifOutDiscards++
#define END_M2_OUTERRORS(pEnd)		    (pEnd)->mib2Tbl.ifOutErrors++
#define END_M2_OUTOCTETS(pEnd,bytes)	(pEnd)->mib2Tbl.ifOutOctets += bytes

/* imports */



/* bind external support interface to default BSP implementation */
IMPORT int    endMultiLstCnt (END_OBJ *);

IMPORT STATUS pciIntConnect (VOIDFUNCPTR *, VOIDFUNCPTR, int);
IMPORT STATUS pciIntDisconnect2 (VOIDFUNCPTR *, VOIDFUNCPTR, int);
IMPORT STATUS sysKs884xIntEnable (int);
IMPORT STATUS sysKs884xIntDisable (int);

STATUS (* ks884xIntConnect)
       (VOIDFUNCPTR *, VOIDFUNCPTR, int) = pciIntConnect;
STATUS (* ks884xIntDisconnect)
       (VOIDFUNCPTR *, VOIDFUNCPTR, int) = pciIntDisconnect2;

STATUS (* ks884xIntEnable)(int)   = sysKs884xIntEnable;
STATUS (* ks884xIntDisable)(int)  = sysKs884xIntDisable;


#define DRV_DEBUG
/*
 */
#ifdef DRV_DEBUG        /* if debugging driver */
#    include "logLib.h"
#    define DRV_DEBUG_OFF           0x000000
#    define DRV_DEBUG_RX            0x000001
#    define DRV_DEBUG_TX            0x000002
#    define DRV_DEBUG_INT           0x000004
#    define DRV_DEBUG_POLL_RX       0x000008
#    define DRV_DEBUG_POLL_TX       0x000010
#    define DRV_DEBUG_POLL          (DRV_DEBUG_POLL_RX | DRV_DEBUG_POLL_TX)
#    define DRV_DEBUG_LOAD          0x000020
#    define DRV_DEBUG_UNLOAD        0x000040
#    define DRV_DEBUG_IOCTL         0x000080
#    define DRV_DEBUG_RESET         0x000100
#    define DRV_DEBUG_MCAST         0x000200
#    define DRV_DEBUG_CSR           0x000400
#    define DRV_DEBUG_RX_PKT        0x000800
#    define DRV_DEBUG_START         0x001000
#    define DRV_DEBUG_STOP          0x002000
#    define DRV_DEBUG_TX_DONE       0x004000
#    define DRV_DEBUG_PHY           0x010000
#    define DRV_DEBUG_BDs           0x020000
#    define DRV_DEBUG_CHIP_INIT     0x040000
#    define DRV_DEBUG_WATCHDOG      0x080000
#    define DRV_DEBUG_ERROR         0x800000
#    define DRV_DEBUG_ALL           0xffffff

LOCAL int ksDebug = ( DRV_DEBUG_ERROR | DRV_DEBUG_STOP  /*| DRV_DEBUG_INT | DRV_DEBUG_POLL_TX */);

#define DRV_LOG(FLG, X0, X1, X2, X3, X4, X5, X6)       \
    if (ksDebug & FLG)                              \
        logMsg((char *) X0, (int)X1, (int)X2, (int)X3, \
                            (int)X4, (int)X5, (int)X6);

#define DRV_PRINT(FLG,X)                                                \
	if (ksDebug & FLG) printf X;

#define DUMPPACKET_DEFAULT          ( DRV_DEBUG_DUMP_TX | DRV_DEBUG_DUMP_RX ) 

#else /* DRV_DEBUG */

#define DRV_LOG(DBG_SW, X0, X1, X2, X3, X4, X5, X6)
#define DRV_PRINT(FLG,X)                                                \

#define DUMPPACKET_DEFAULT          ( 0 ) 

#endif /* DRV_DEBUG */

#define DRV_DEBUG_DUMP_TX           0x100000
#define DRV_DEBUG_DUMP_RX           0x200000


/* globals */

IMPORT int ks884x_max_if;  

DRV_CTRL 	*gpDrvCtrl[KS884X_MAX_IF];


/* locals */

LOCAL volatile int ks884xBusEndianess   = KS_LITTLE_ENDIAN;
LOCAL volatile int ks884xInterruptLevel = KS_ACTIVE_LOW;


/* default description for the END device */
LOCAL char * pDescription = "Micrel %s PCI Enhanced Network Driver";
BYTE 	     fKS884x = 0;      /* 1: KS8841, 2: KS8842 */
char         ksIpAddress[INET_ADDR_LEN];

const char *deviceName[] =
{
/*  0 */  "UNKNOWN",
/*  1 */  "KSZ8841",
/*  2 */  "KSZ8842",
/*  3 */  "KSZ8861",
/*  4 */  "KSZ8862",
	      ""
};

/* forward declarations */

LOCAL STATUS    ks884xReset (DRV_CTRL * pDrvCtrl);
LOCAL void      ks8842xInt (DRV_CTRL * pDrvCtrl);
LOCAL void      ks884xHandleRecvInt (DRV_CTRL * pDrvCtrl, UINT32 uIER);
LOCAL STATUS    ks884xRecv (DRV_CTRL * pDrvCtrl, M_BLK_ID, RXDESC * );
LOCAL void      ks884xHandleTxDoneInt (DRV_CTRL * pDrvCtrl, UINT32 uIER);
LOCAL void      ks884xTxDoneProcess ( DRV_CTRL * pDrvCtrl );
LOCAL void      ks884xHandleError (DRV_CTRL * pDrvCtrl, UINT32 uIER);
LOCAL void      ks884xRestart (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xRestartSetup (DRV_CTRL * pDrvCtrl);
LOCAL void      ks884xAddrFilterSet (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xMemInit (DRV_CTRL * pDrvCtrl);

LOCAL void      ks884xSetDefaults ( DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xAllocTxBDs ( DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xAllocRxBDs ( DRV_CTRL * pDrvCtrl);
LOCAL void      ks884xAllocRxBuffers ( DRV_CTRL * pDrvCtrl);
      M_BLK_ID  ks884xAllocMblk ( DRV_CTRL *, UCHAR **, M_BLK_ID );
LOCAL void      ks884xFreeTxBDs ( DRV_CTRL * pDrvCtrl);
LOCAL void      ks884xFreeRxBDs ( DRV_CTRL * pDrvCtrl);
LOCAL void      CleanTxRing ( DRV_CTRL * pDrvCtrl);
LOCAL void      CleanRxRing ( DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ksSetApiFunctions ( DRV_CTRL * pDrvCtrl	);

LOCAL STATUS    ks884xEnetAddrGet ( DRV_CTRL *, char * );
LOCAL STATUS    ks884xEnetAddrSet ( DRV_CTRL *, char * );
LOCAL STATUS    ks884xGetPhyStatus ( DRV_CTRL * pDrvCtrl, PORT_KS_INFO *, ULONG);
LOCAL void      ks884xAutoNegoStart ( DRV_CTRL * pDrvCtrl, PORT_KS_INFO *, ULONG );
LOCAL void      ks884xWatchdog ( DRV_CTRL * pDrvCtrl);
LOCAL void      getDeviceName ( USHORT );
      void      showKsBspInfo ( void );
      void      showKsLinkStatus ( void );


/* END-Specific interfaces */

LOCAL STATUS    ks884xEndStart (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xEndStop (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xEndUnload (DRV_CTRL * pDrvCtrl);
LOCAL int       ks884xEndIoctl (DRV_CTRL * pDrvCtrl, int cmd, caddr_t data);
LOCAL STATUS    ks884xEndMCastAddrAdd (DRV_CTRL* pDrvCtrl, char * pAddress);
LOCAL STATUS    ks884xEndMCastAddrDel (DRV_CTRL * pDrvCtrl, char * pAddress);
LOCAL STATUS    ks884xMCastAddrGet (DRV_CTRL * pDrvCtrl, MULTI_TABLE * pTable);
      STATUS    ks884xEndSend (DRV_CTRL * pDrvCtrl, M_BLK_ID pBuf);
      STATUS    ks884xEndPollSend (DRV_CTRL * pDrvCtrl, M_BLK_ID pBuf);
      STATUS    ks884xEndPollReceive (DRV_CTRL * pDrvCtrl, M_BLK_ID pBuf);
      STATUS    ks884xEndPollStart (DRV_CTRL * pDrvCtrl);
      STATUS    ks884xEndPollStop (DRV_CTRL * pDrvCtrl);
      STATUS    ks884xInitParse ( DRV_CTRL *  pDrvCtrl, char * );

/*
 * Declare our function table.  This is static across all driver
 * instances.
 */
LOCAL NET_FUNCS ks884xFuncTable =
    {
    (FUNCPTR) ks884xEndStart,          /* Function to start the device. */
    (FUNCPTR) ks884xEndStop,           /* Function to stop the device. */
    (FUNCPTR) ks884xEndUnload,         /* Unloading function for the driver. */
    (FUNCPTR) ks884xEndIoctl,          /* Ioctl function for the driver. */
    (FUNCPTR) ks884xEndSend,           /* Send function for the driver. */
    (FUNCPTR) ks884xEndMCastAddrAdd,   /* Multicast address add  */
    (FUNCPTR) ks884xEndMCastAddrDel,   /* Multicast address delete */
    (FUNCPTR) ks884xMCastAddrGet,      /* Multicast table retrieve */
    (FUNCPTR) ks884xEndPollSend,       /* Polling send function  */
    (FUNCPTR) ks884xEndPollReceive,    /* Polling receive function */
    endEtherAddressForm,               /* Put address info into a packet.  */
    endEtherPacketDataGet,             /* Get a pointer to packet data. */
    endEtherPacketAddrGet              /* Get packet addresses. */
    };


/* inline functions */


/******************************************************************************
*
* rbdNextAddrSet - set a receive descriptor's next descriptor address (RDES3)
*
* This routine sets the next descriptor address specified by <pNextRBD> 
* to a Rx Descriptor, specified by <pRBD>.
*
* RETURNS: N/A
*
* NOMANUAL
*/
__inline__ LOCAL void rbdNextAddrSet
(
    const DRV_CTRL  * pDrvCtrl,  /* specifies device control structure */
    volatile RXDESC * pRBD,      /* specifies a current descriptor RXDESC structure */
    volatile RXDESC * pNextRBD   /* specifies a next descriptor RXDESC structure */
)
{
    UINT32 pTmp  = MEM_TO_PCI_PHYS (KS_CACHE_VIRT_TO_PHYS ((UINT32) pNextRBD));
    pRBD->RxDMANextPtr = (UINT32) PCI_SWAP (pTmp);
}

/******************************************************************************
*
* rbdBuffAddrSet - set a receive descriptor's packet buffer address (RDES2)
*
* This routine sets the input packet buffer address specified by <pBuf> 
* to a Rx Descriptor, specified by <pRBD>.
*
* RETURNS: N/A
*
* NOMANUAL
*/
__inline__ LOCAL void rbdBuffAddrSet
(
    const DRV_CTRL  * pDrvCtrl,  /* specifies device control structure */
    volatile RXDESC * pRBD,      /* specifies an RXDESC structure */
    const char *      pBuf       /* buffer addr to associate with RXDESC */
)
{
    UINT32 pTmp  = MEM_TO_PCI_PHYS (KS_CACHE_VIRT_TO_PHYS ((UINT32) pBuf));
    pRBD->RxDMABufAddr = (UINT32) PCI_SWAP (pTmp);
}


/******************************************************************************
*
* rxDescSet - set a receive descriptor's control (RDES0, RDES1).
*
* This routine sets the receive control (RDES1) to a Rx Descriptor, 
* specified by <pTBD>. 
* And sets the receive control (RDES0) OWN bit to be owned by ks884x.
*
* RETURNS: N/A
*
* NOMANUAL
*/
__inline__ LOCAL void rxDescSet
(
    const DRV_CTRL  * pDrvCtrl,  /* specifies device control structure */
    volatile RXDESC * pRBD       /* specifies the RBD to "clean" */
)
{
    pRBD->RxFrameControl= pRBD->RxFrameControl | ( PCI_SWAP ( DESC_OWN_BIT ) );
}


/******************************************************************************
*
* rbdNextAddrGet - get a receive descriptor's next descriptor address (RDES3).
*
* RETURNS:  The next descriptor address associated with the specified RXDESC,
*           specified by <pRBD>.
*
* NOMANUAL
*/
__inline__ LOCAL RXDESC * rbdNextAddrGet
(
    const DRV_CTRL * pDrvCtrl,         /* specifies device control structure */
    const volatile RXDESC * pRBD       /* specifies a RXDESC structure */
)
{
    UINT32 pTemp = PCI_TO_MEM_PHYS (PCI_SWAP (pRBD->RxDMANextPtr));
    return (RXDESC *)(KS_CACHE_PHYS_TO_VIRT (pTemp));
}

/******************************************************************************
*
* rbdBuffAddrGet - get a receive descriptor's packet buffer address (RDES2).
*
* RETURNS:  The packet buffer address associated with the specified RXDESC,
*           specified by <pRBD>.
*
* NOMANUAL
*/
__inline__ LOCAL char * rbdBuffAddrGet
(
    const DRV_CTRL * pDrvCtrl,         /* specifies device control structure */
    const volatile RXDESC * pRBD       /* specifies a RXDESC structure */
)
{
    UINT32 pTemp = PCI_TO_MEM_PHYS (PCI_SWAP (pRBD->RxDMABufAddr));
    return (char *)(KS_CACHE_PHYS_TO_VIRT (pTemp));
}

/******************************************************************************
*
* tbdNextAddrSet - set a transmit descriptor's next descriptor address (TDES3).
*
* This routine sets the next descriptor address specified by <pNextRBD> 
* to a Tx Descriptor, specified by <pTBD>.
*
* RETURNS: N/A
*
* NOMANUAL
*/
__inline__ LOCAL void tbdNextAddrSet
(
    const DRV_CTRL  * pDrvCtrl,  /* specifies device control structure */
    volatile TXDESC * pTBD,      /* specifies a current TXDESC structure */
    volatile TXDESC * pNextRBD   /* specifies a next descriptor TXDESC structure */
)
{
    UINT32 pTmp  = MEM_TO_PCI_PHYS (KS_CACHE_VIRT_TO_PHYS ((UINT32) pNextRBD));
    pTBD->TxDMANextPtr = (UINT32) PCI_SWAP (pTmp);
}

/******************************************************************************
*
* tbdBuffAddrSet - set a transmit descriptor's packet buffer address (TDES2).
*
* This routine sets the output packet buffer address specified by <pBuf> 
* to a Tx Descriptor, specified by <pTBD>.
*
* RETURNS: N/A
*
* NOMANUAL
*/
__inline__ LOCAL void tbdBuffAddrSet
(
    const DRV_CTRL  * pDrvCtrl,  /* specifies device control structure */
    volatile TXDESC * pTBD,      /* specifies a TXDESC structure */
    const char *      pBuf       /* buffer addr to associate with TMD */
)
{
    UINT32 pTmp  = MEM_TO_PCI_PHYS (KS_CACHE_VIRT_TO_PHYS ((UINT32) pBuf));
    pTBD->TxDMABufAddr = (UINT32) PCI_SWAP (pTmp);
}

/******************************************************************************
*
* txDescSet - set a transmit descriptor's control (TDES0, TDES1).
*
* This routine sets the transmit control (TDES1) specified by txFrameCntlMask 
* to a Tx Descriptor, specified by <pTBD>. 
* And sets the transmit control (TDES0) OWN bit to be owned by ks884x.
*
* RETURNS: N/A
*
* NOMANUAL
*/
__inline__ LOCAL void txDescSet
(
    const DRV_CTRL  * pDrvCtrl,         /* specifies device control structure */
    volatile TXDESC * pTBD,             /* specifies the TBD to "clean" */
    UINT32            txFrameCntlMask
)
{
    pTBD->TxFrameControl = PCI_SWAP ( txFrameCntlMask ) ;
    pTBD->TxOwnBit       = PCI_SWAP ( DESC_OWN_BIT );
}


__inline__ LOCAL UINT32 txDescGet
(
    const DRV_CTRL  * pDrvCtrl,         /* specifies device control structure */
    volatile TXDESC * pTBD              /* specifies the TBD to "clean" */
)
{
    return ( PCI_SWAP ( pTBD->TxFrameControl ) );
}


/******************************************************************************
*
* tbdNextAddrGet - get a transmit descriptor's next descriptor address (TDES3).
*
* RETURNS:  The next descriptor address associated with the specified Tx Descriptor.
*
* NOMANUAL
*/
__inline__ LOCAL TXDESC * tbdNextAddrGet
(
    const DRV_CTRL * pDrvCtrl,         /* specifies device control structure */
    const volatile TXDESC * pTBD       /* specifies a TXDESC structure */
)
{
    UINT32 pTemp = PCI_TO_MEM_PHYS (PCI_SWAP (pTBD->TxDMANextPtr));
    return (TXDESC *)(KS_CACHE_PHYS_TO_VIRT (pTemp));
}

/******************************************************************************
*
* tbdBuffAddrGet - get a transmit descriptor's packet buffer address (TDES2).
*
* RETURNS:  The packet buffer address associated with the specified Tx Descriptor.
*
* NOMANUAL
*/
__inline__ LOCAL char * tbdBuffAddrGet
(
    const DRV_CTRL * pDrvCtrl,         /* specifies device control structure */
    const volatile TXDESC * pTBD       /* specifies a TXDESC structure */
)
{
    UINT32 pTemp = PCI_TO_MEM_PHYS (PCI_SWAP (pTBD->TxDMABufAddr));
    return (char *)(KS_CACHE_PHYS_TO_VIRT (pTemp));
}

/******************************************************************************
*
* rxBDHasErrors - test a Received DMA Descriptor's error bits
*
* RETURNS:
* TRUE if the any receive DMA descriptor ERR bit is set 
*
* NOMANUAL
*/
__inline__ LOCAL BOOL rbdHasErrors
(
    DRV_CTRL * pDrvCtrl,         /* specifies device control structure */
    const volatile RXDESC * pRBD   /* specifies the RBD to "check" */
)
{
 	UINT32 uFrameCtrl = PCI_SWAP(pRBD->RxFrameControl); 
    PHARDWARE     phw = pDrvCtrl->pHW;
    BOOL       status = FALSE;
    ULONG      portIndex=0;

	if (uFrameCtrl & RFC_ERROR_MASK ) 
    {
        /* Update driver receive error counters */
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR ]++;
		if (uFrameCtrl & RFC_TL) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_TOOLONG ]++;
		if (uFrameCtrl & RFC_CRC) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_CRC ]++;
		if (uFrameCtrl & RFC_RF) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_RUNT ]++;

		if (uFrameCtrl & RFC_RE) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_MII ]++;
		if (uFrameCtrl & RFC_IPE) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_IP ]++;
		if (uFrameCtrl & RFC_TCPE) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_TCP ]++;
		if (uFrameCtrl & RFC_UDPE) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_ERROR_UDP ]++;

		status = TRUE;

    } /* if (uFrameCtrl & RFC_ERROR_MASK) */

    return (status);
}


/*
 * ks884x_dump_packet
 *	This function is use to dump given packet for debugging.
 *
 * Argument(s)
 *	pDrvCtrl	pointer to DRV_CTRL structure.
 *	data		pointer to the beginning of the packet to dump
 *	len			length of the packet
 *
 * Return(s)
 *	NONE.
 */
  
__inline void ks884x_dump_packet 
( 
    UINT8         *data, 
    int            len,
    UINT32         port,
    UINT32         flag
)
{
    if  ( ((flag & DRV_DEBUG_DUMP_TX) != DRV_DEBUG_DUMP_TX) &&
          ((flag & DRV_DEBUG_DUMP_RX) != DRV_DEBUG_DUMP_RX) )
        return;


    if ( flag & DRV_DEBUG_DUMP_TX )
         printf ("Tx On port %d \n", port);
    else if ( flag & DRV_DEBUG_DUMP_RX )
         printf ("Rx On port %d \n", port);


	if (len >= 18) 
    {
		printf("Pkt Len=%d\n", len);
		printf("DA=%02x:%02x:%02x:%02x:%02x:%02x\n", 
				*data, *(data + 1), *(data + 2), *(data + 3), *(data + 4), *(data + 5));
		printf("SA=%02x:%02x:%02x:%02x:%02x:%02x\n", 
				*(data + 6), *(data + 7), *(data + 8), *(data + 9), *(data + 10), *(data + 11));
		printf("Type=%04x (%d)\n", ntohs(*(unsigned short *)(data + 12)), ntohs(*(unsigned short *)(data + 12)));

        {
			int	j = 0, k;

			do {
				printf("\n %04d   ", j);
				for (k = 0; (k < 16 && len); k++, data++, len--) 
					printf("%02x  ", *data);
				j += 16;
			} while (len > 0);
			printf("\n");
		}
	}
}


/******************************************************************************
*
* ks884xEndLoad - initialize the driver and device
*
* This routine initializes the driver and the device to the operational state.
* All of the device-specific parameters are passed in <initString>, which
* expects a string of the following format:
*
* <unit:devMemAddr:devIoAddr:pciMemBase:vecnum:intLvl:memAdrs
* :memSize:memWidth:boardBusEndianMode:offset:flags>
*
* This routine can be called in two modes. If it is called with an empty but
* allocated string, it places the name of this device (that is, "ks") into 
* the <initString> and returns 0.
*
* If the string is allocated and not empty, the routine attempts to load
* the driver using the values specified in the string.
*
* CAVEATS
* For ks8842 device (two port switch), the two END interfaces are represented for 
* the two switch port. But, ks8842 only has one set of DMA (same as ks8841).
* So both END interface's "pDrvCtrl->pDMAInfo" point to same DMA structure.
*
* RETURNS: An END object pointer, or NULL on error, or 0 and the name of the
* device if the <initString> was NULL.
*/
END_OBJ * ks884xEndLoad
(
    char * initString            /* string to be parse by the driver */
)
{
    DRV_CTRL * pDrvCtrl;
    PHARDWARE  phw;
    struct hw_fn *pks_fn=NULL;
    char       ksEndDescription[100];
    USHORT   ChipID;
    UCHAR    DevRevisionID;



    if (initString == NULL)
    {
        return (NULL);
    }

    if (initString [0] == 0)
    {
        bcopy ((char *)KS_DEV_NAME, initString, KS_DEV_NAME_LEN);
        return ((END_OBJ *) OK);
    }

    /* allocate the device structure */

    if ((pDrvCtrl = (DRV_CTRL *) calloc (sizeof (DRV_CTRL), 1))
        == NULL)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): allocate pDrvCtrl failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* parse the init string, filling in the device structure */

    if (ks884xInitParse (pDrvCtrl, initString) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): ks884xInitParse failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* Record pDrvCtrl as Global */
    gpDrvCtrl[pDrvCtrl->unit] = pDrvCtrl;

    /* 
     *  There is only one ks884x HW structure. 
     *  Always point to the same HW structure, no matter is is kx8841 or ks8842
     */
    if ( pDrvCtrl->unit == 0)
    {
        /* allocate the ks884x HW structure */
        pDrvCtrl->pHW = (PHARDWARE)calloc (sizeof (HARDWARE), 1);
        if (pDrvCtrl->pHW == NULL)
        {
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): allocate pHW failed\n", 1, 2, 3, 4, 5, 6);
	        goto errorExit;
        }
        memset( pDrvCtrl->pHW , 0, sizeof( HARDWARE ));

        /* allocate the ks884x internal set API structure */
        pks_fn = (struct hw_fn *)calloc (sizeof (struct hw_fn), 1);
        if (pks_fn == NULL)
        {
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): allocate pks_fn failed\n", 1, 2, 3, 4, 5, 6);
	        goto errorExit;
        }
        memset( pks_fn , 0, sizeof( struct hw_fn ));
        pDrvCtrl->pHW->m_hwfn = pks_fn;

        /* allocate the ks884x DMA structure */
        pDrvCtrl->pDMAInfo = (PDMA_INFO)calloc (sizeof (DMA_INFO), 1);
        if (pDrvCtrl->pDMAInfo == NULL)
        {
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): allocate DMA failed\n", 1, 2, 3, 4, 5, 6);
	        goto errorExit;
        }
        memset( pDrvCtrl->pDMAInfo , 0, sizeof( DMA_INFO ));
    }
    else
    {
        /* it is second END interface, point to the same ks884x HW staructure */
        pDrvCtrl->pHW =  gpDrvCtrl[0]->pHW;
        pDrvCtrl->pDMAInfo = gpDrvCtrl[0]->pDMAInfo; 
    }

    /* 
     * Assign "memory mapped I/O base address of the ks884x device registers
     * in the memory map of the CPU" to ks884x HW structure.
     * And PCI Conroller endian mode.
     *
     * We only can access ks884x registers after here.
     */
    phw = pDrvCtrl->pHW;
    phw->m_ulVIoAddr  = pDrvCtrl->devAdrs; 
    phw->m_boardBusEndianMode = pDrvCtrl->boardBusEndianMode; 

	/* set some others default values. */
	ks884xSetDefaults( pDrvCtrl );


    /* Have the BSP hand us our address. */

    #if (0)
    	printf("stMacStation MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", 
		    DI->stMacStation[0], 
		    DI->stMacStation[1],
		    DI->stMacStation[2],
		    DI->stMacStation[3], 
		    DI->stMacStation[4],
		    DI->stMacStation[5]); 

    #endif

    /* Get ks884x station MAC address */
    (* ks884xEnetAddrGet) (pDrvCtrl, (char *) &(pDrvCtrl->enetAddr [0]));

	/* if device's MAC address is not assigned, use driver's default MAC address instead */
	if (
       ((0 == pDrvCtrl->enetAddr[0] && 0 == pDrvCtrl->enetAddr[1] &&
		 0 == pDrvCtrl->enetAddr[2] && 0 == pDrvCtrl->enetAddr[3] &&
		 0 == pDrvCtrl->enetAddr[4] && 0 == pDrvCtrl->enetAddr[5])) ||
	   ((0xff == pDrvCtrl->enetAddr[0] && 0xff == pDrvCtrl->enetAddr[1] &&
		 0xff == pDrvCtrl->enetAddr[2] && 0xff == pDrvCtrl->enetAddr[3] &&
		 0xff == pDrvCtrl->enetAddr[4] && 0xff == pDrvCtrl->enetAddr[5])) 
       )
    {
        /* Set driver's default MAC to ks884x station MAC address */
        if ( ks884xEnetAddrSet(pDrvCtrl, (char *) &DI->stMacStation[0]) == ERROR )
        {
	        DRV_PRINT(DRV_DEBUG_ERROR, 
                     ("ks884xEndLoad(): fail to set MAC address to ks884x.\n"));
            goto errorExit; 
        }
	}
    else
    {
		memcpy(&DI->stMacStation, pDrvCtrl->enetAddr, ETH_LENGTH_OF_ADDRESS);  
    }

    /* initialize the END and MIB2 parts of the structure */

    HardwareReadChipID ( phw, &ChipID, &DevRevisionID );
    getDeviceName ( ChipID );

    memset(ksEndDescription, 0, sizeof(ksEndDescription));
    sprintf( ksEndDescription, pDescription, deviceName[fKS884x]  );

    if (END_OBJ_INIT (&pDrvCtrl->endObj, (DEV_OBJ *)pDrvCtrl, KS_DEV_NAME,
                      pDrvCtrl->unit, &ks884xFuncTable, ksEndDescription) == ERROR
     || END_MIB_INIT (&pDrvCtrl->endObj, M2_ifType_ethernet_csmacd,
                      &pDrvCtrl->enetAddr[0], 6, ETHERMTU, LN_SPEED) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): END_OBJ_INIT failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* Perform memory allocation */

    if (ks884xMemInit (pDrvCtrl) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): ks884xMemInit failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* set the flags to indicate readiness */

    END_OBJ_READY (&pDrvCtrl->endObj, IFF_NOTRAILERS | IFF_BROADCAST |
                                      IFF_MULTICAST  | IFF_SIMPLEX);

    /* reset and reconfigure the device */

    if (ks884xRestartSetup (pDrvCtrl) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndLoad(): ks884xRestartSetup failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }


    /* Create a watchdog timer */
	pDrvCtrl->watchdog = wdCreate();
	if (NULL == pDrvCtrl->watchdog) 
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "wdCreate failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
	}

	if (OK != wdStart(pDrvCtrl->watchdog, pDrvCtrl->nWatchdogDelay, (FUNCPTR)ks884xWatchdog, (int)pDrvCtrl)) 
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "wdStart failed\n", 1, 2, 3, 4, 5, 6);
		wdDelete(pDrvCtrl->watchdog);
		pDrvCtrl->watchdog = NULL;
	}

 	DRV_LOG(DRV_DEBUG_LOAD, "ks884xEndLoad(unit=%d) done.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    return (&pDrvCtrl->endObj);

errorExit:
    if (pks_fn != NULL)
        free ((char *)pks_fn);

    if (pDrvCtrl->pHW != NULL)
        free ((char *)pDrvCtrl->pHW);

    if (pDrvCtrl->pDMAInfo != NULL)
        free ((char *)pDrvCtrl->pDMAInfo);

    if (pDrvCtrl != NULL)
        free ((char *)pDrvCtrl);

    return ((END_OBJ *)NULL);
}

/*******************************************************************************
*
* ks884xInitParse - parse the initialization string
*
* Parse the input string. This routine is called from ks884xEndLoad() which
* intializes some values in the driver control structure with the values
* passed in the intialization string.
*
* The initialization string format is:
*
* <unit:devMemAddr:devIoAddr:pciMemBase:vecNum:intLvl:memAdrs
* :memSize:memWidth:boardBusEndianMode:offset:flags>
*
* \is
* \i <unit>
* The device unit number.  Unit numbers are integers starting at zero and
* increasing for each device contolled by the driver.
* \i <devMemAddr>
* The device memory mapped I/O register base address.  Device registers
* must be mapped into the host processor address space in order for the
* driver to be functional.  Thus, this is a required parameter.
* \i <devIoAddr>
* Device register base I/O address (obsolete).
* \i <pciMemBase>
* Base address of PCI memory space.
* \i <vecNum>
* Interrupt vector number.
* \i <intLvl>
* Interrupt level.  Generally, this value specifies an interrupt level
* defined for an external interrupt controller.
* \i <memAdrs>
* Memory pool address or NONE.
* \i <memSize>
* Memory pool size or zero.
* \i <memWidth>
* Memory system size, 1, 2, or 4 bytes (optional).
* \i <boardBusEndianMode>
* Board Bus Endian mode options (1- board bus is big endian mode).
* \i <offset>
* Memory alignment offset.
* \i <flags>
* Device specific flags reserved for future use.
* \ie
*
* RETURNS: OK, or ERROR if any arguments are invalid.
*/
STATUS ks884xInitParse
(
    DRV_CTRL *  pDrvCtrl,       /* pointer to the control structure */
    char *      initString      /* initialization string */
)
{
    char *      tok;
    char *      pHolder = NULL;
    UINT32      devMemAddr;
    UINT32      devIoAddr;



    /* unit number */

    if ((tok = strtok_r (initString, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->unit = atoi (tok);
    DRV_LOG (DRV_DEBUG_LOAD, "Unit : %d ...\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    /* memory mapped IO address base */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    devMemAddr = (UINT32) strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_LOAD, "devMemAddr : 0x%X ...\n", devMemAddr,
             2, 3, 4, 5, 6);

    /* Intel IO space address base */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    devIoAddr = (UINT32) strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_LOAD, "devIoAddr : 0x%X ...\n", devIoAddr,
             2, 3, 4, 5, 6);


    /* always use memory mapped IO if provided, else use IO map */
    
    if ((devMemAddr == NONE) && (devIoAddr == NONE))
    {
        return (ERROR);
    }
    else if (devMemAddr != NONE)
    {
        pDrvCtrl->devAdrs = devMemAddr;
        pDrvCtrl->flags  |= KS_MODE_MEM_IO_MAP;
    }
    else
    {
        pDrvCtrl->devAdrs = devIoAddr;
    }


    /* PCI memory base address as seen from the CPU */
    
    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->pciMemBase = strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_LOAD, "Pci : 0x%X ...\n", pDrvCtrl->pciMemBase,
             2, 3, 4, 5, 6);

    /* interrupt vector */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->inum = atoi (tok);
    DRV_LOG (DRV_DEBUG_LOAD, "inum : 0x%X ...\n", pDrvCtrl->inum,
             2, 3, 4, 5, 6);

    /* interrupt level */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->ilevel = atoi (tok);
    DRV_LOG (DRV_DEBUG_LOAD, "ilevel : 0x%X ...\n", pDrvCtrl->ilevel,
             2, 3, 4, 5, 6);

    /* caller supplied memory address */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->memAdrs = (char *)strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_LOAD, "memAdrs : 0x%X ...\n", (int)pDrvCtrl->memAdrs,
             2, 3, 4, 5, 6);

    /* caller supplied memory size */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->memSize = strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_LOAD, "memSize : 0x%X ...\n", pDrvCtrl->memSize,
             2, 3, 4, 5, 6);

    /* caller supplied memory width */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->memWidth = atoi (tok);
    DRV_LOG (DRV_DEBUG_LOAD, "memWidth : 0x%X ...\n", pDrvCtrl->memWidth,
             2, 3, 4, 5, 6);

    /* boardBusEndianMode value */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->boardBusEndianMode = atoi (tok);
    DRV_LOG (DRV_DEBUG_LOAD, "boardBusEndianMode : 0x%X ...\n", pDrvCtrl->boardBusEndianMode,
             2, 3, 4, 5, 6);

    /* caller supplied alignment offset */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->offset = atoi (tok);
    DRV_LOG (DRV_DEBUG_LOAD, "Offset : 0x%X ...\n", pDrvCtrl->offset,
             2, 3, 4, 5, 6);

    /* caller supplied flags */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->flags |= strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_LOAD, "flags : 0x%X ...\n", pDrvCtrl->flags,
             2, 3, 4, 5, 6);


    return OK;
}

/*******************************************************************************
*
* ks884xMemInit - initialize memory for Lance chip
*
* Using data in the control structure, setup and initialize the memory
* areas needed.  If the memory address is not already specified, then allocate
* cache safe memory.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xMemInit
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    UINT32            sz;       /* temporary size holder */


    /***** Establish size of shared memory region we require *****/

  	sz = (DI->nTxDescTotal * TXBD_SIZE) + (DI->nRxDescTotal * RXBD_SIZE) + 
          _CACHE_ALIGN_SIZE * 2;


	DRV_LOG (DRV_DEBUG_LOAD, "ks884xMemInit(unit=%d): sz=%d\n",
			                  pDrvCtrl->unit, sz, 3, 4, 5, 6);

    /***** Establish a region of shared memory *****/

    /* OK. We now know how much shared memory we need.  If the caller
     * provides a specific memory region, we check to see if the provided
     * region is large enough for our needs.  If the caller did not
     * provide a specific region, then we attempt to allocate the memory
     * from the system, using the cache aware allocation system call.
     */

    switch ((int) pDrvCtrl->memAdrs)
    {
	    default :       /* caller provided memory */
	        if (pDrvCtrl->memSize < sz)     /* not enough space */
		    {
		        DRV_LOG (DRV_DEBUG_ERROR, "ks884xMemInit(unit=%d): not enough memory provided\n"
			        "ks884x: need %ul got %d\n",
			        pDrvCtrl->unit, sz, pDrvCtrl->memSize, 4, 5, 6);
		        return (ERROR);
            }

	        /* set the beginning of pool */

	        pDrvCtrl->pShMem = pDrvCtrl->memAdrs;

            DRV_LOG (DRV_DEBUG_LOAD, "(default)pShMem : 0x%08X\n", 
                                     pDrvCtrl->pShMem, 2, 3, 4, 5, 6);

	        /* assume pool is cache coherent, copy null structure */

	        pDrvCtrl->cacheFuncs = cacheNullFuncs;
	        break;

        case NONE :     /* get our own memory */

	        /* Because the structures that are shared between the device
	        * and the driver may share cache lines, the possibility exists
	        * that the driver could flush a cache line for a structure and
	        * wipe out an asynchronous change by the device to a neighboring
	        * structure. Therefore, this driver cannot operate with memory
	        * that is not write coherent.  We check for the availability of
	        * such memory here, and abort if the system did not give us what
	        * we need.
	        */

	        if (!CACHE_DMA_IS_WRITE_COHERENT ())
		    {
		        printf ( "ks884x: device requires cache coherent memory\n" );
		        return (ERROR);
		    }

	        pDrvCtrl->pShMem = (char *) cacheDmaMalloc (sz);

	        if (pDrvCtrl->pShMem == NULL)
		    {
		        printf ( "ks884x: system memory unavailable\n" );
		        return (ERROR);
		    }
            DRV_LOG (DRV_DEBUG_LOAD, "(NONE)pShMem: 0x%08X, size: 0x%08X\n", 
                                     pDrvCtrl->pShMem, sz, 3, 4, 5, 6);

	        /* copy the DMA structure */

	        pDrvCtrl->cacheFuncs = cacheDmaFuncs;

            break;
    }


    /* Save some things */

    pDrvCtrl->memBase = (char *)((UINT32) pDrvCtrl->pShMem & 0xff000000);

    if ((int) pDrvCtrl->memAdrs == NONE)
        pDrvCtrl->flags |= KS_MEM_ALLOC_FLAG;

    /* first let's clear memory */

    bzero ((char *) pDrvCtrl->pShMem, (int) sz);

    /* allocate pool structure for mblks, clBlk, and clusters */

    if ((pDrvCtrl->endObj.pNetPool = malloc (sizeof (NET_POOL))) == NULL)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xMemInit(unit=%d): malloc for pNetPool failed !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
	    return (ERROR);
    }

    
    pDrvCtrl->clDesc.clNum    = (DI->nRxDescTotal + 2) * 2;  
    pDrvCtrl->mClCfg.clBlkNum = pDrvCtrl->clDesc.clNum;
    pDrvCtrl->mClCfg.mBlkNum  = pDrvCtrl->mClCfg.clBlkNum * 2; 

    /* total memory size for mBlks and clBlks */
    
    pDrvCtrl->mClCfg.memSize =
        (pDrvCtrl->mClCfg.mBlkNum  * (MSIZE + sizeof (long))) +
        (pDrvCtrl->mClCfg.clBlkNum * (CL_BLK_SZ + sizeof (long)));

    /* total memory for mBlks and clBlks */

    if ((pDrvCtrl->mClCfg.memArea = (char *) memalign
        (sizeof (long), pDrvCtrl->mClCfg.memSize)) == NULL)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xMemInit(unit=%d): memalign failed !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return (ERROR);
    }

    DRV_LOG (DRV_DEBUG_LOAD, "mClCfg.memArea: 0x%08X, mClCfg.memSize: 0x%08X\n", 
                             pDrvCtrl->mClCfg.memArea, pDrvCtrl->mClCfg.memSize, 3, 4, 5, 6);

    /* total memory size for all clusters */

    pDrvCtrl->clDesc.clSize  = ROUND_UP(DI->uRxBufferLen, _CACHE_ALIGN_SIZE);
    pDrvCtrl->clDesc.memSize =
        (pDrvCtrl->clDesc.clNum * (pDrvCtrl->clDesc.clSize + 8)) + sizeof (int);

    /* Do we hand over our own memory? */

    if (pDrvCtrl->memAdrs != (char *) NONE)
        pDrvCtrl->clDesc.memArea =
                  (char *)(pDrvCtrl->pShMem + (DI->nTxDescTotal * TXBD_SIZE));
    else
        pDrvCtrl->clDesc.memArea = cacheDmaMalloc (pDrvCtrl->clDesc.memSize);

    if (pDrvCtrl->clDesc.memArea == NULL)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xMemInit(unit=%d): system memory unavailable !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return (ERROR);
    }

    DRV_LOG (DRV_DEBUG_LOAD, "clDesc.memArea: 0x%08X, clDesc.memSize: 0x%08X\n", 
                             pDrvCtrl->clDesc.memArea, pDrvCtrl->clDesc.memSize, 3, 4, 5, 6);

    /* initialize the device net pool */
    
    if (netPoolInit (pDrvCtrl->endObj.pNetPool, &pDrvCtrl->mClCfg,
                     &pDrvCtrl->clDesc, 1, NULL) == ERROR)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xMemInit(unit=%d): netPoolInit fail !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return (ERROR);
    }
    
    /* Store the cluster pool ID as others need it later. */

    pDrvCtrl->pClPoolId = clPoolIdGet (pDrvCtrl->endObj.pNetPool,
                                       pDrvCtrl->clDesc.clSize, FALSE);

#if (0)
    printf ("ks884x: BDs address(pDrvCtrl->pShMem) = 0x%08x\n", (int)pDrvCtrl->pShMem );
    printf ("ks884x: BDs buf address(pDrvCtrl->clDesc.memArea) = 0x%08x\n", (int)pDrvCtrl->clDesc.memArea  );
#endif

    return OK;
 }

/*******************************************************************************
*
* ks884xEndStart - start the device
*
* This function calls BSP functions to connect interrupts and start the
* device running in interrupt mode.
*
* CAVEATS
* For ks8842 device (two port switch), the two END interfaces are represented for 
* the two switch port. But, ks8842 only has one set of DMA (same as ks8841)  
* need to be started.
*
* RETURNS: OK or ERROR if the interrupt handler could not be connected.
*/
LOCAL STATUS ks884xEndStart
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    static int initialized = 0;
    PHARDWARE  phw = pDrvCtrl->pHW;
    STATUS     result = OK;
    UINT32     tempData;


    pDrvCtrl->txCleaning = FALSE;
    pDrvCtrl->txBlocked  = FALSE;
	pDrvCtrl->rxHandling = FALSE;



    /* only execute the following code once. */
    if(initialized == 0)
    {
 	    DRV_LOG(DRV_DEBUG_START, "ks884xEndStart: pDrvCtrl->inum=%d.\n", pDrvCtrl->inum, 2, 3, 4, 5, 6);

        result = (* ks884xIntConnect)((VOIDFUNCPTR *) INUM_TO_IVEC(pDrvCtrl->inum),
                                 ks8842xInt, (int)(pDrvCtrl));

        if (result == OK)
        {

            /* initialise the hardware interrupt mask for proper operation. */
            HardwareSetupInterrupt( phw );
            phw->m_ulInterruptMask &= ~( INT_TX_UNDERRUN );
            phw->m_ulInterruptMask |= INT_RX_OVERRUN;

            CACHE_PIPE_FLUSH (); /* Flush the write pipe */

            /* Clear interrupts on the device-side */
            HardwareAcknowledgeInterrupt( phw, 0xFFFFFFFF );
            CACHE_PIPE_FLUSH (); /* Flush the write pipe */

            /* enables the interrupts of the hardware. */
            HardwareEnableInterrupt( phw );
            CACHE_PIPE_FLUSH (); /* Flush the write pipe */

            /* Enable system level interrupt */
            (* ks884xIntEnable) (pDrvCtrl->ilevel);

	        /* Enable MAC DMA Tx/Rx */
            HardwareEnable( phw );
            CACHE_PIPE_FLUSH (); /* Flush the write pipe */

	        /* Start MAC DMA Rx */
	        HW_WRITE_DWORD ( phw, REG_DMA_RX_START, DMA_START ); 
            CACHE_PIPE_FLUSH (); /* Flush the write pipe */

 	        DRV_LOG(DRV_DEBUG_START, "ks884xEndStart(unit=%d) done.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

        }
        else
 	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndStart(unit=%d): IntConnect fail.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);


        /* Display Device Driver info on UART */
        showKsBspInfo (); 
        showKsLinkStatus ();

        initialized ++;

    }   /* if(initialized == 0) */

    if (result == OK)
    {

        /* mark the interface as up */

        END_FLAGS_SET (&pDrvCtrl->endObj, (IFF_UP | IFF_RUNNING));

    }


    return (result);
}

/*******************************************************************************
*
* ks884xHandleError - handle controller interrupt errors
*
* This routine performs task level error handling for controller errors
* detected in interrupt context.
*
* RETURNS: N/A.
*/
LOCAL void ks884xHandleError
(
    DRV_CTRL * pDrvCtrl,   /* pointer to DRV_CTRL structure */
    UINT32     uIER        /* holds error bits read from interrupt status */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    ULONG      portIndex=0;


    DRV_LOG(DRV_DEBUG_ERROR, "ks884xHandleError: %s Process Stop\n", 
            ((uIER & INT_TX_STOP) ? "Transmit":"Receive"), 2, 3, 4, 5, 6);

    pDrvCtrl->errorHandling = FALSE;

	/* increase counter */
    if ( uIER & INT_TX_STOP )
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT_STOP ]++;
    else
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_INT_STOP ]++;

	/* will implement error recovery later
	 */

	/* enable corresponding int bit again */
    HardwareEnableInterruptBit ( phw, uIER );

}

/*******************************************************************************
*
* ks8842xInt - handle controller interrupt
*
* This routine is called at interrupt level in response to an interrupt from
* the controller.
*
* RETURNS: N/A
*/
LOCAL void ks8842xInt
(
    DRV_CTRL * pDrvCtrl  /* pointer to DRV_CTRL structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
	ULONG      uISR=0;
    ULONG      errs = 0;   /* assume no errors */
    int        portIndex = 0;




    /* Read the device interrupt status register */
    HardwareReadInterrupt( phw, &uISR );

    DRV_LOG(DRV_DEBUG_INT, "\n ks8842xInt(unit=%d), uISR=0x%08X\n", pDrvCtrl->unit, uISR, 3, 4, 5, 6); 

    /* If false interrupt, return. */
    if (! (uISR & phw->m_ulInterruptMask))
    {
        return;
    }

    uISR &= phw->m_ulInterruptMask;

    /* Acknowledge the interrupt. */
    HardwareAcknowledgeInterrupt( phw, uISR );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    /* Disable the interrupt bit */
    HardwareDisableInterruptBit( phw, uISR );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    /* 
     * Process transmitted frame descriptors 
     */

    if ( (uISR & INT_TX) || (uISR & INT_TX_UNDERRUN) )
    {
        if (uISR & INT_TX_UNDERRUN) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT_UNDERRUN ]++;

        if ( (uISR & INT_TX) )
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT ]++;

        if ( !(pDrvCtrl->txCleaning) )
        { 
            if (netJobAdd ((FUNCPTR) ks884xHandleTxDoneInt,
                (int) pDrvCtrl, (int)(uISR & (INT_TX | INT_TX_UNDERRUN)), 0, 0, 0) == OK)
            {
                pDrvCtrl->txCleaning = TRUE;
            }
        }
    }

    /* 
     * Process received frame descriptors 
     */

    if ( (uISR & INT_RX) || (uISR & INT_RX_OVERRUN) )
    {
	    ULONG  uIER = uISR | (INT_RX | INT_RX_OVERRUN);

        if ( uISR & INT_RX ) 
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_INT ]++;
        if ( uISR & INT_RX_OVERRUN )
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_NO_BUFFER ]++;

        if ( !(pDrvCtrl->rxHandling) )
        {
            /* put a task level input packet handler on the netTask job queue */

            if (netJobAdd ((FUNCPTR) ks884xHandleRecvInt,
                           (int) pDrvCtrl, (int)(uIER),0,0,0) == OK)
            {
                pDrvCtrl->rxHandling = TRUE;
            }
        }
    }


	/* 
	 *  PHY Link Change
	 */

    if ( ( uISR & INT_PHY ) ) 
    {
        /* Checking Link status */
        SwitchGetLinkStatus( phw );

        /* enable corresponding int bit again */
        HardwareEnableInterruptBit ( phw, INT_PHY );
    }


    /* 
     * Check for any DMA Stop 
     */

    if ( (uISR & INT_TX_STOP) || (uISR & INT_RX_STOP) )
    {

	    DRV_LOG(DRV_DEBUG_ERROR, "ks8842xInt: %s Process Stop\n", 
                ((uISR & INT_TX_STOP) ? "Transmit":"Receive"), 2, 3, 4, 5, 6);

        if ( uISR & INT_TX_STOP )
        {
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT_STOP ]++;

			/* Todo: user can modify DMA transmit configuration now. */

			/* Enable DMA Transmit */
            HardwareStartTransmit( phw );

	        /* enable corresponding int bit again */
            HardwareEnableInterruptBit ( phw, INT_TX_STOP );
        }
        else
        {
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_INT_STOP ]++;

			/* Todo: user can modify DMA receive configuration now. */

			/* Enable DMA Receive */
            HardwareStartReceive( phw );

	        /* enable corresponding int bit again */
            HardwareEnableInterruptBit ( phw, INT_RX_STOP );
        }
	}


#if (0)
    if (pDrvCtrl->txBlocked)    /* cause a MUX restart */
    {
        if (netJobAdd ((FUNCPTR) muxTxRestart,
           (int) &pDrvCtrl->endObj, 0, 0, 0, 0) == OK)
        {
            pDrvCtrl->txBlocked = FALSE;
        }
    }
#endif
}

/*******************************************************************************
*
* ks884xHandleRecvInt - task level interrupt service for input packets
*
* This routine is called at task level indirectly by the interrupt
* service routine to do any message received processing.
*/
LOCAL void ks884xHandleRecvInt
(
    DRV_CTRL * pDrvCtrl,  /* points to DRV_CTRL structure */
    UINT32     uIER       /* holds INT_RX_FRAME bit read from interrupt status */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    M_BLK_ID 	pMblk;
	RXDESC     *CurrentDescriptor, *firstDescriptor;
	UINT32      uFrameCtrl;
	int	        i;
	int         oldLevel = 0;



    DRV_LOG(DRV_DEBUG_RX, "ks884xHandleRecvInt (%d), uIER=0x%08X\n", pDrvCtrl->unit, uIER, 3, 4, 5, 6);

 
	i = DI->nRxDescNextAvail;
	CurrentDescriptor = &DI->pRxDescBaseAddr[i];
	KS_CACHE_INVALIDATE(CurrentDescriptor, sizeof(RXDESC));


	/* keep it for later use */
	firstDescriptor = CurrentDescriptor;

	while ( !(( uFrameCtrl = PCI_SWAP(CurrentDescriptor->RxFrameControl) ) & DESC_OWN_BIT) ) 
    {
		/* if it is a valid Rx descriptor, the pMblk should point to a pMblk buffer!!! */
		pMblk = DI->pRxMblk[i].pMblk;
		if (NULL == pMblk) 
        {
			DRV_LOG(DRV_DEBUG_ERROR, "ks884xHandleRecvInt (%d), null pMblk\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);
			break;
		}

        /* precess receive packet */
        ks884xRecv (pDrvCtrl, pMblk, CurrentDescriptor);

        oldLevel = intLock();	/* protect it */

        /* go to next descriptor */
		i = (++i % DI->nRxDescTotal);
		CurrentDescriptor = &DI->pRxDescBaseAddr[i];

		intUnlock(oldLevel);

        /* only process one round of Rx dexcriptors */
		if (firstDescriptor == CurrentDescriptor) 
        {
			DRV_LOG(DRV_DEBUG_RX, "one round break\n", 1, 2, 3, 4, 5, 6);
			break;
		}

		KS_CACHE_INVALIDATE(CurrentDescriptor, sizeof(RXDESC));

    } /* while (!((uFrameCtrl = CurrentDescriptor->RxFrameControl) & DESC_OWN_BIT)) */


	DI->nRxDescNextAvail = i;

    pDrvCtrl->rxHandling = FALSE;

	/* enable Rx engine again */
	HW_WRITE_DWORD ( phw, REG_DMA_RX_START, DMA_START ); 
    
	/* 
     * More stuff to do - re-queue ourselves 
     */
#if (1)
	uFrameCtrl = PCI_SWAP (CurrentDescriptor->RxFrameControl); 
	if  (!(uFrameCtrl & DESC_OWN_BIT)) 
    {
        if ( !(pDrvCtrl->rxHandling) )
        {
           if (netJobAdd ((FUNCPTR) ks884xHandleRecvInt,
                          (int) pDrvCtrl, (int)uIER,0,0,0) == OK)
           {
               pDrvCtrl->rxHandling = TRUE;
           }
        }
    }

	/* enable Rx engine again */
	HW_WRITE_DWORD ( phw, REG_DMA_RX_START, DMA_START ); 
#endif

	/* enable Rx interrupt bit again */
    HardwareEnableInterruptBit ( phw, uIER );
}



/*
 * ReceiveBufferClone
 *	This function is use to get a new mblk/clblk/cluster and copy recevied data to 
 *  this new Mblk buffer in order to make IP header 32 bit alignment.

 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *	pSrcMblk	pointer to M_BLK_ID with source mblk to copy data from.
 *	len			lenght of data packet received
 *
 * Return(s)
 *	NONE.
 */
LOCAL M_BLK_ID ReceiveBufferClone
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
    M_BLK_ID   pSrcMblk, 
    int        len
)
{
	M_BLK_ID pMblk;
	UCHAR	*pCluster = NULL;

	DRV_LOG(DRV_DEBUG_RX, "ReceiveBufferClone\n", 1, 2, 3, 4, 5, 6);

	pMblk = ks884xAllocMblk(pDrvCtrl, &pCluster, NULL);
	if (NULL == pMblk) 
    {
	    DRV_LOG(DRV_DEBUG_RX, "ReceiveBufferClone(): Out of mblk/clblk/cluster!\n", 
                              1, 2, 3, 4, 5, 6);
		return NULL;
    }

	/* shift data to make IP header dword aligned if required */
	pMblk->mBlkHdr.mData += pDrvCtrl->offset;
	pSrcMblk->mBlkPktHdr.len = len;
	pSrcMblk->mBlkHdr.mLen = len;
    netMblkToBufCopy(pSrcMblk, pMblk->mBlkHdr.mData, NULL);

    /* make the packet data coherent */
    KS_CACHE_INVALIDATE(pMblk->mBlkHdr.mData, len);

	pMblk->mBlkHdr.mLen = len;
	pMblk->mBlkHdr.mFlags |= M_PKTHDR;
	pMblk->mBlkPktHdr.len = len;

	pSrcMblk->mBlkPktHdr.len = 0;
	pSrcMblk->mBlkHdr.mLen = 0;

	return pMblk;
}


/*******************************************************************************
*
* ks884xRecv - process the incoming packet
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xRecv
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
    M_BLK_ID   pMblk, 
	RXDESC   * CurrentDescriptor
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    UINT32     uPort;
    M_BLK_ID   pCloneMblk;
	UINT32     Length;
	UINT32     uFrameCtrl;
    int        portIndex = 0;



    DRV_LOG (DRV_DEBUG_RX, "ks884xRecv \n", 1, 2, 3, 4, 5, 6);

 	uFrameCtrl = PCI_SWAP (CurrentDescriptor->RxFrameControl); 

	/* make the packet data coherent */
	KS_CACHE_INVALIDATE(pMblk->mBlkHdr.mData, DI->uRxBufferLen);

    /* Get the source port that packet come from */
    uPort = (uFrameCtrl & DESC_RX_PORT_MASK) >> 20;
    phw->m_bPortRX = uPort;
    portIndex = phw->m_bPortRX;
    if ( portIndex > 0 )
         portIndex -- ;

    if ( phw->fDebugDumpRx )
    {
		/* length with CRC bytes included */
		Length = (uFrameCtrl & RFC_FL_MASK);

        /* for debug use */
        printf("RDES0:0x%08X, RDES1:0x%08X, RDES2:0x%08X, RDES3:0x%08X\n", 
                PCI_SWAP(CurrentDescriptor->RxFrameControl), 
                PCI_SWAP(CurrentDescriptor->RxDMABufLen), 
                rbdBuffAddrGet(pDrvCtrl, CurrentDescriptor), 
                (UINT32)CurrentDescriptor->RxDMANextPtr);
        ks884x_dump_packet( (UINT8 *)pMblk->mBlkHdr.mData, Length, uPort, (DI->uDebugDumpPkt & DRV_DEBUG_DUMP_RX));
    }

	/* check if error happened!!! */
    if  ( rbdHasErrors (pDrvCtrl,CurrentDescriptor) )
    {
#if (0)
		DRV_LOG(DRV_DEBUG_ERROR, "%s> error found (framectrl=0x%08x)\n", 
                __FUNCTION__, uFrameCtrl, 3, 4, 5, 6);
#endif

        END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
		goto CLEAN_UP;
    }


	/* 
     * if we are here, means a valid packet received!!! Get length of the pacekt 
     */

    /* get the packet length */

	if ( (uFrameCtrl & RFC_LS ) &&  (uFrameCtrl & RFC_FS) ) 
    {
	    /* received frame is in single buffers  */

		/* length with CRC bytes included */
		Length = (uFrameCtrl & RFC_FL_MASK);

		/* offset CRC bytes! */
		Length -= ETH_CRC_LENGTH;
    }
	else 
    {
	    /* received frame in multiple buffers  */

		if (uFrameCtrl & RFC_FS) 
        {
			/* first segment */
			Length = DI->uRxBufferLen;

			/* compensite offset CRC */
			Length += ETH_CRC_LENGTH;
		}
		else if (uFrameCtrl & RFC_LS) 
        {
			/* last segment */
			Length += (uFrameCtrl & RFC_FL_MASK);
		}

        /* Should nerve goes here, since we set received buffer is big enough to hold entire frame */
		DRV_LOG(DRV_DEBUG_ERROR, "%s> entire frame not in one buffer (framectrl=0x%08x)\n", 
                __FUNCTION__, uFrameCtrl, 3, 4, 5, 6);
		goto CLEAN_UP;

    } /* if ((uFrameCtrl & (RFC_LS | RFC_FS)) != (RFC_LS | RFC_FS)) */

    phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_RCV ] += Length;

	/* make the packet data coherent */

	KS_CACHE_INVALIDATE( pMblk->mBlkHdr.mData, Length );

	if (uFrameCtrl & RFC_MF) 
    {
		END_M2_INNUCAST(&pDrvCtrl->endObj);
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_MULTICAST_FRAMES_RCV ]++;
    }
	else 
    {
		END_ERR_ADD(&pDrvCtrl->endObj, MIB2_IN_UCAST, +1);
		END_M2_INUCAST(&pDrvCtrl->endObj);
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_UNICAST_FRAMES_RCV ]++;
	}

	END_M2_INOCTETS(&pDrvCtrl->endObj, Length);
    phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_RCV ]++;


#ifdef TEST_THROUGHPUT
    /*
     *  We loopback received packets to transmit.
     */
    if ( phw->fLoopbackStart )
    {
        M_BLK_ID 	pNewMblkTemp= NULL;
        UCHAR	  * pClusterTemp = NULL;

	    /* copy recevied data to a new Mblk buffer for Transmit */
        if ( (pNewMblkTemp = (M_BLK_ID)ks884xAllocMblk(pDrvCtrl, &pClusterTemp, NULL) ) != NULL ) 
        {
             bcopy(pMblk->mBlkHdr.mData, pNewMblkTemp->mBlkHdr.mData, Length);

	         pNewMblkTemp->mBlkHdr.mLen = Length;
	         pNewMblkTemp->mBlkHdr.mFlags |= M_PKTHDR;
	         pNewMblkTemp->mBlkPktHdr.len = Length;
             pNewMblkTemp->mBlkHdr.mNext = NULL;
  
	         if (! (uFrameCtrl & RFC_MF) ) 
                  pNewMblkTemp->mBlkHdr.mData[5] = pNewMblkTemp->mBlkHdr.mData[11] ;  /* DA = SA */
             pNewMblkTemp->mBlkHdr.mData[11]= DI->stMacStation[5] ;              /* SA = MacStation */

             /* make the packet data coherent */
             KS_CACHE_INVALIDATE(pNewMblkTemp->mBlkHdr.mData, Length);

             /* the clean descriptor to the device's DMA controller  */
             rxDescSet ( pDrvCtrl, CurrentDescriptor );
	         KS_CACHE_FLUSH(CurrentDescriptor, sizeof(RXDESC));

             /* Set "fPollSendFromCLI" to indicate that ks884xEndPollSend from CLI command */
             pDrvCtrl->fPollSendFromCLI = TRUE; 

             /* Set trasnmit destination port */
             phw->m_bPortTX = uPort;

             /* Send the received packet to the device */
             ks884xEndSend ( pDrvCtrl, pNewMblkTemp );

             return (OK);
        }
        else
        {
             phw->m_cnCounter[ uPort-1][ OID_COUNTER_XMIT_ALLOC_FAIL ]++;

	         /* 
              * No available TxBufs 
              */
             if (DI->fTxNoResource) 
             {
                /* clean up processed tx descriptors */
                pDrvCtrl->txBlocked = TRUE;        /* transmitter not ready */

                /* clean up processed tx descriptors */
	            ks884xTxDoneProcess(pDrvCtrl);
	         }

	         goto CLEAN_UP;               
        }
    }
    else 
    {
        /* for the device Rx direction throughput test, not pass packets to the upper layer */
	    goto CLEAN_UP;               
    }

#endif /* #ifdef TEST_THROUGHPUT */

	/* 
	 * copy recevied data to a new Mblk buffer in order to make IP header 32 bit alignment.
	 */
	pCloneMblk = ReceiveBufferClone(pDrvCtrl, pMblk, Length);
	if (NULL == pCloneMblk) 
    {
		/* drop it */
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_DROPPED ]++;
		END_ERR_ADD(&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
	    goto CLEAN_UP;
	}

	/* Call the upper layer's receive routine. */
	END_RCV_RTN_CALL(&pDrvCtrl->endObj, pCloneMblk);


CLEAN_UP:

	/* 
	 * done with this descriptor, 
     * return the clean descriptor to the device's DMA controller.
   	 */

    rxDescSet ( pDrvCtrl, CurrentDescriptor );
	KS_CACHE_FLUSH(CurrentDescriptor, sizeof(RXDESC));

    return (OK);
}

/*******************************************************************************
*
* ks884xEndSend - the driver send routine
*
* This routine takes a M_BLK_ID sends off the data in the M_BLK_ID.
* The buffer must already have the addressing information properly installed
* in it.  This is done by a higher layer.
*
*
* RETURNS: OK or ERROR.
*/
STATUS ks884xEndSend
(
    DRV_CTRL * pDrvCtrl,  /* device to be initialized */
    M_BLK_ID   pMblk      /* data to send */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    UINT32     uPort;
    UINT32     txFrameCntlMask;
	TXDESC	  *pTxDesc;
	TXDESC	  *firstDescriptor, *CurrentDescriptor;
	int		   i, dataLen;
	char	  *data;
	BOOL	   fFisrtSegment = TRUE;
	int		   oldLevel = 0;
    ULONG      portIndex=0;



    /* check device mode */

    if (pDrvCtrl->flags & KS_POLLING)
    {
        netMblkClChainFree (pMblk);
        errno = EINVAL;

        return (ERROR);
    }

#if (1)
	/* 
     * No available TxBufs 
     */
    if (DI->fTxNoResource) 
    {
        /* clean up processed tx descriptors */
        pDrvCtrl->txBlocked = TRUE;        /* transmitter not ready */

        /* clean up processed tx descriptors */
	    ks884xTxDoneProcess(pDrvCtrl);

		errno = EINVAL;
		/* return END_ERR_BLOCK; */ /* just return without freeing mBlk chain */
	}
#endif

    /*
     * Obtain exclusive access to transmitter.  This is necessary because
     * we might have more than one stack transmitting at once.
     */

    END_TX_SEM_TAKE (&pDrvCtrl->endObj, WAIT_FOREVER);

    /*
     * One packet from vxWorks IP Stack may occupy multiple clusters.
     */

    if ( pDrvCtrl->fPollSendFromCLI )
        uPort = phw->m_bPortTX;
    else
        uPort = pDrvCtrl->unit;

    portIndex = uPort;

    i = DI->nTxDescNextAvail;
  	firstDescriptor = &DI->pTxDescBaseAddr[i];

    while ( ( pMblk ) /* && (DI->nTxDescAvail >0) */)
    {
	    /* get a Tx Frame Descriptor */
	    i = DI->nTxDescNextAvail;
	    pTxDesc = &DI->pTxDescBaseAddr[i];

	    /* invalidate to make consistency between device and host */
	    KS_CACHE_INVALIDATE(pTxDesc, sizeof(TXDESC));

	    /*
	     * If resources are not available, release the semaphore
	     * and return END_ERR_BLOCK. Do not free packet if use END_ERR_BLOCK.
	     */
	    if ( PCI_SWAP(pTxDesc->TxOwnBit) & DESC_OWN_BIT ) 
        {
		    DRV_LOG(DRV_DEBUG_TX, "ks884xEndSend(): Current Tx descriptor is not owned by host\n", 1, 2, 3, 4, 5, 6);

            pDrvCtrl->txBlocked = TRUE;

		    /* note that if return END_ERR_BLOCK, don't need to free Mblk */
            END_TX_SEM_GIVE (&pDrvCtrl->endObj);

		    return END_ERR_BLOCK;
        }

	    if (DI->pTxMblk[i].pMblk) 
        {
		    DRV_LOG(DRV_DEBUG_TX, "%s: pMblk is not null, txBD=%d\n", 
                    __FUNCTION__, (int)i, 3, 4, 5, 6);

            /* free this DI->pTxMblk[i].pMblk, 
               since ks884xTxDoneProcess() has not got the chance to do it.*/
			netMblkClChainFree(DI->pTxMblk[i].pMblk);
			DI->pTxMblk[i].pMblk = NULL;
        }

        /* get data pointer and data length in this mBlk */ 
	    data = pMblk->mBlkHdr.mData;
        dataLen = pMblk -> mBlkHdr.mLen;
	    KS_CACHE_USER_FLUSH(data, dataLen);

	    DI->pTxMblk[i].pMblk = pMblk;

	    /* set Tx DMA buffer address */
        tbdBuffAddrSet (pDrvCtrl, pTxDesc, data);

        /* set DMA tx descriptor control */

        txFrameCntlMask = txDescGet (pDrvCtrl, pTxDesc);

        txFrameCntlMask &= (0x1E000000);
		txFrameCntlMask |= (dataLen & TFC_TBS_MASK) | TFC_IC;
        txFrameCntlMask |= (fFisrtSegment) ?           (TFC_FS) : 0 ;
        txFrameCntlMask |= (!pMblk -> mBlkHdr.mNext) ? (TFC_LS) : 0 ;
        txFrameCntlMask |= (uPort << 20);                /* Set destination port */

        txDescSet ( pDrvCtrl, pTxDesc, txFrameCntlMask );

	    /* update it for device to read */
	    KS_CACHE_FLUSH(pTxDesc, sizeof(TXDESC));

        if ( portIndex == 3 )
        {
            while ( portIndex & 3)
            {
               portIndex >>= 1;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;
            }
        }
        else
        {
            if ( portIndex > 0 ) 
               portIndex --;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;
        }

        if ( phw->fDebugDumpTx )
        {
            printf("TDES0:0x%08X, TDES1:0x%08X, TDES2:0x%08X, TDES3:0x%08X\n", 
                    PCI_SWAP(pTxDesc->TxOwnBit), 
                    PCI_SWAP(pTxDesc->TxFrameControl), 
                    tbdBuffAddrGet(pDrvCtrl, pTxDesc), 
                    (int)(pTxDesc->TxDMANextPtr));
            ks884x_dump_packet((UINT8 *) data, dataLen, uPort, (DI->uDebugDumpPkt & DRV_DEBUG_DUMP_TX)); 
        }

	    /* update pending transimt packet count */
	    DI->nTxDescAvail--;

	    if (DI->nTxDescAvail < 1) 
        {
		    /* this will block next transmit until we clear the flag */
		    DI->fTxNoResource = TRUE;
		    DRV_LOG(DRV_DEBUG_TX, "ks884xEndSend(): Oops, no Tx Resource (unit=%d)\n", 
                    pDrvCtrl->unit, 2, 3, 4, 5, 6);
        }

	    /* kick off Tx transfer */
	    HW_WRITE_DWORD ( phw, REG_DMA_TX_START, DMA_START ); 

	    /* flush write pipe */
	    CACHE_PIPE_FLUSH();

        /* increment the unicast output counter */
        END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_UCAST, +1);
        if ( portIndex == 3 )
        {
            while ( portIndex & 3)
            {
               portIndex >>= 1;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;
            }
        }
        else
        {
            if ( portIndex > 0 ) 
                portIndex --;
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;
        }

 		oldLevel = intLock();	/* protect it */

	    /* adv to next available descriptor */
	    DI->nTxDescNextAvail = ++DI->nTxDescNextAvail % DI->nTxDescTotal;

        /* Reference next cluster of the Mblk chain */
        pMblk = pMblk -> mBlkHdr.mNext ;
        fFisrtSegment = FALSE;

	    DI->nTransmitCount++;
        /* DBG_PRINT ("ks884xEndSend() nTransmitCount = %d"NEWLINE, DI->nTransmitCount); */

		intUnlock(oldLevel);

        /* only process one round of Tx dexcriptors */
  	    CurrentDescriptor = &DI->pTxDescBaseAddr[DI->nTxDescNextAvail];
		if (firstDescriptor == CurrentDescriptor) 
        {
			DRV_LOG(DRV_DEBUG_TX, "one round break\n", 1, 2, 3, 4, 5, 6);
			break;
		}

    } /* while ( !pMblk ) */

    /* Advance our management index(es) */
    END_TX_SEM_GIVE (&pDrvCtrl->endObj);

    return (OK);
}

/*******************************************************************************
*
* ks884xEndIoctl - the driver I/O control routine
*
* Process an ioctl request.
*
* RETURNS OK or ERROR value
*/
LOCAL int ks884xEndIoctl
(
    DRV_CTRL * pDrvCtrl, /* device to be initialized */
    int        cmd,      /* ioctl command to execute */
    caddr_t    data      /* data to get or set */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    int    error = 0;

    switch ((UINT)cmd)
    {
        case EIOCSADDR:
            if (data == NULL)
                return (EINVAL);

            bcopy ((char *)data, (char *)END_HADDR(&pDrvCtrl->endObj),
                   END_HADDR_LEN (&pDrvCtrl->endObj));
            break;

        case EIOCGADDR:
            if (data == NULL)
                return (EINVAL);

            bcopy ((char *)END_HADDR (&pDrvCtrl->endObj), (char *)data,
                   END_HADDR_LEN (&pDrvCtrl->endObj));
            break;

        case EIOCSFLAGS:
            {
            long value    = (long) data;
            long oldFlags = END_FLAGS_GET (&pDrvCtrl->endObj);
            long newFlags;

            if (value < 0)
            {
                value = -value;
                --value;
                END_FLAGS_CLR (&pDrvCtrl->endObj, value);
            }
            else
            {
                END_FLAGS_SET (&pDrvCtrl->endObj, value);
            }

            /* Handle IFF_XXX flags set or cleared by command */
              
            newFlags = END_FLAGS_GET (&pDrvCtrl->endObj);

            /* IFF_ALLMULTI */

            if ((newFlags & IFF_ALLMULTI) && !(oldFlags & IFF_ALLMULTI))
            {
                /* Enable Multicast frames mode to MAC Rx DMA control */
                HardwareSetMulticast( phw, TRUE );
                pDrvCtrl->flags |= KS_MULTICAST_FLAG; 
            }
            else if ((oldFlags & IFF_ALLMULTI) && !(newFlags & IFF_ALLMULTI))
            {
                /* Disable Multicast frames mode to MAC Rx DMA control */
                HardwareSetMulticast( phw, FALSE );
                pDrvCtrl->flags &= ~KS_MULTICAST_FLAG; 
            }

            /* IFF_PROMISC */
              
            if ((newFlags & IFF_PROMISC) && !(oldFlags & IFF_PROMISC))
            {
                /* Enable Promiscuous mode to device's MAC Rx DMA control */
                HardwareSetPromiscuous( phw, TRUE );
                pDrvCtrl->flags |= KS_PROMISCUOUS_FLAG; 
            }
            else if ((oldFlags & IFF_PROMISC) && !(newFlags & IFF_PROMISC))
            {
                /* Disable Promiscuous mode to device's MAC Rx DMA control */
                HardwareSetPromiscuous( phw, FALSE );
                pDrvCtrl->flags &= ~KS_PROMISCUOUS_FLAG; 
            }

            }
            break;

        case EIOCGFLAGS:
            *(int *)data = END_FLAGS_GET (&pDrvCtrl->endObj);
            break;

        case EIOCMULTIADD:
            error = ks884xEndMCastAddrAdd (pDrvCtrl, (char *) data);
            break;

        case EIOCMULTIDEL:
            error = ks884xEndMCastAddrDel (pDrvCtrl, (char *) data);
            break;

        case EIOCMULTIGET:
            error = ks884xMCastAddrGet (pDrvCtrl, (MULTI_TABLE *) data);
            break;

        case EIOCPOLLSTART:
            error = ks884xEndPollStart (pDrvCtrl);
            break;

        case EIOCPOLLSTOP:
            error = ks884xEndPollStop (pDrvCtrl);
            break;

        case EIOCGMIB2:
            if (data == NULL)
                return (EINVAL);
            bcopy((char *)&pDrvCtrl->endObj.mib2Tbl, (char *)data,
                  sizeof(pDrvCtrl->endObj.mib2Tbl));
            break;

        case EIOCGFBUF:
            if (data == NULL)
                return (EINVAL);
            *(int *)data = DI->uRxBufferLen;
            break;

        case EIOCGMWIDTH:
            if (data == NULL)
                return (EINVAL);
            *(int *)data = pDrvCtrl->memWidth;
            break;

        case EIOCGHDRLEN:
            if (data == NULL)
                return (EINVAL);
            *(int *)data = SIZEOF_ETHERHEADER;
            break;

        default:
            error = EINVAL;
    }

    return (error);
}

/**
 *	Set internal KS884x API function for ISA bus.
 *
 */ 
/*******************************************************************************
*
* ksSetApiFunctions
*
*	Set internal KS884x API function for ISA bus.
*
* RETURNS: NONE.
*/
LOCAL STATUS ksSetApiFunctions 
(    
    DRV_CTRL * pDrvCtrl	/* device ID */
)
{
    struct hw_fn *pks_fn = pDrvCtrl->pHW->m_hwfn;


    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ksSetApiFunctions(): Start... \n"));

    if ( pks_fn == NULL )
    {
        DRV_PRINT(DRV_DEBUG_ERROR, 
                 ("ksSetApiFunctions(): pks_fn is NULL. \n"));
        return ERROR;
    }

    HardwareSetupFunc( pks_fn );

    return OK;
}

/*******************************************************************************
*
* ks884xReset - hardware reset of chip (stop it)
*
* This routine is responsible for resetting the device
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xReset
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    PHARDWARE     phw    = pDrvCtrl->pHW;

    /* Generate a soft-reset of the controller */

#if (1)
    if ( !HardwareReset( phw ) ) 
        return (ERROR);
#endif

    return (OK);
}


/*******************************************************************************
*
* ks884xRestartSetup - 
*
* Setup memory descriptors and initializes hardware chip.
*
* CAVEATS
* For ks8842 device (two port switch), the two END interfaces are represented for 
* the two switch port. But, ks8842 only has one set of DMA (same as ks8841)  
* need to be initialized.
*
* RETURNS OK or ERROR.
*/
LOCAL STATUS ks884xRestartSetup
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    static int initialized = 0;
    PHARDWARE  phw = pDrvCtrl->pHW;
    UINT32     rxDescBase, txDescBase;
    UINT32     delayCount=20;



    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xRestartSetup(): Start initializes hardware chip... \n"));

    /* set promiscuous mode if it's asked for */

    if (END_FLAGS_GET (&pDrvCtrl->endObj) & IFF_PROMISC)
    {
        pDrvCtrl->flags |= KS_PROMISCUOUS_FLAG; 
        DI->bRxPromiscuous = TRUE;
        phw->m_bPromiscuous = TRUE;
    }
    else
    {
        pDrvCtrl->flags &= ~KS_PROMISCUOUS_FLAG; 
        DI->bRxPromiscuous = FALSE;
        phw->m_bPromiscuous = FALSE;
    }

    /* set receive all multicast frames mode if it's asked for */

    if (END_FLAGS_GET (&pDrvCtrl->endObj) & (IFF_ALLMULTI | IFF_MULTICAST))
    {
        pDrvCtrl->flags |= KS_MULTICAST_FLAG; 
        DI->bRxMulticast = TRUE;
        phw->m_bAllMulticast = TRUE;
    }
    else
    {
        pDrvCtrl->flags &= ~KS_MULTICAST_FLAG; 
        DI->bRxMulticast = FALSE;
        phw->m_bAllMulticast = FALSE;
        phw->m_bMulticastListSize = 0;
    }

    /* only execute the following code once. */
    if(initialized == 0)
    {

       /* 
        * Allocate Tx\Rx descriptors and Rx buffers 
        */

        /* initialize the Transmit Buffer Descriptors ring */
	    if ( ks884xAllocTxBDs(pDrvCtrl) == ERROR ) 
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                     ("ks884xRestartSetup(unit=%d): ERROR on initialize the TBD ring. \n", pDrvCtrl->unit));
            return ERROR ;
        }

        /* initialize the Receive Buffer Descriptors ring */
	    if ( ks884xAllocRxBDs(pDrvCtrl) == ERROR ) 
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                      ("ks884xRestartSetup(unit=%d): ERROR on initialize the RBD ring. \n", pDrvCtrl->unit ));
            return ERROR ;
        }

        
       /* 
        * Set ks884x internal API function point 
        */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling ksSetApiFunctions()... \n"));
        if ( ksSetApiFunctions ( pDrvCtrl) == ERROR )
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                     ("ks884xRestartSetup(unit=%d): ERROR on set ks884x internal API. \n", pDrvCtrl->unit));
            return ERROR ;
        }
        
       /* 
        * reset the device  
        */
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling ks884xReset()... \n"));
        if ( ks884xReset (pDrvCtrl) == ERROR )
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                    ("ks884xRestartSetup(unit=%d): ERROR on reset ks884x. \n", pDrvCtrl->unit ));
            return ERROR ;
        }


        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling HardwareInitialize()... \n"));
        if ( !HardwareInitialize( phw ) ) 
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                      ("ks884xRestartSetup(unit=%d): ERROR on incorrect ks884x chip ID. \n", pDrvCtrl->unit ));
            return ERROR ;
        }

	    /* Stop MAC DMA first for reconfiguration */
        HardwareDisable ( phw );


        /* set up address filter for multicasting */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                ("ks884xRestartSetup(): calling ks884xAddrFilterSet()... \n"));
        if (END_MULTI_LST_CNT (&pDrvCtrl->endObj) > 0)
        {
            ks884xAddrFilterSet (pDrvCtrl);
        }

	    /* Set base address of Tx\Rx descriptor */
	    HardwareSetDescriptorBase (phw, (ULONG)DI->pTxDescBaseAddr, (ULONG)DI->pRxDescBaseAddr);

        /* 
         * Clear MIB's Counters 
         */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling HardwareClearCounters(1)... \n"));
        phw->m_bPortSelect = 0;         /* clear Port 1 */
        HardwareClearCounters( phw );
#ifdef DEF_KS8842
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling HardwareClearCounters(2)... \n"));
        phw->m_bPortSelect = 1;         /* clear Port 2 */
        HardwareClearCounters( phw );
#endif
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling HardwareClearCounters(3)... \n"));
        phw->m_bPortSelect = 2;         /* clear Port 3 */
        HardwareClearCounters( phw );

        /* 
         * Initialize device's default configuration setting. 
         */ 

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling HardwareSetup()... \n"));
        HardwareSetup( phw );
        HardwareSwitchSetup ( phw );

#ifdef DEF_KS8842

        /* Set Switch function address */
        SwitchSetAddress ( phw, phw->m_bOverrideAddress );

        /* Enable Switch */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): calling SwitchEnable()... \n"));
        SwitchEnable ( phw, TRUE );
#endif

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xRestartSetup(): finished calling HardwareSetup()... \n"));

        initialized ++;
    }


    CACHE_PIPE_FLUSH ();

    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xRestartSetup(): Finished initializes hardware chip. \n"));

    return (OK);
}


/*******************************************************************************
*
* ks884xRestart - restart the device after a fatal error
*
* This routine takes care of all the messy details of a restart.  The device
* is reset and re-initialized.  The driver state is re-synchronized.
*
* RETURNS: N/A
* TODO: call this routine if something fail 
*/
LOCAL void ks884xRestart
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    if (ks884xRestartSetup (pDrvCtrl) == OK)
    {
        /* set the flags to indicate readiness */

        END_OBJ_READY (&pDrvCtrl->endObj, IFF_UP | IFF_RUNNING |
                       IFF_NOTRAILERS | IFF_BROADCAST | IFF_MULTICAST);
    }
}


/******************************************************************************
*
* ks884xAddrFilterSet - set the address filter for multicast addresses
*
* This routine goes through all of the multicast addresses on the list
* of addresses and sets the
* device's logical address filter in the initialization block.
*
* RETURNS: N/A
*
* SEE ALSO:  
*/
LOCAL void ks884xAddrFilterSet
(
    DRV_CTRL * pDrvCtrl             /* pointer to DRV_CTRL structure */
)
{
    UINT8             mcast [MAC_ADDRESS_LENGTH];   /* temporary logical address */
    PHARDWARE         phw = pDrvCtrl->pHW;
    ETHER_MULTI *     pCurr = END_MULTI_LST_FIRST (&pDrvCtrl->endObj);



    DRV_LOG(DRV_DEBUG_MCAST, "ks884xAddrFilterSet(unit=%d) is called.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    /* clear all the MAC address filters */

    HardwareAddrFilterClear ( phw );

    while (pCurr != NULL)
    {
        /* copy the specified address, as we'll destroy it with shifts below */

        mcast[0] = pCurr->addr[0];
        mcast[1] = pCurr->addr[1];
        mcast[2] = pCurr->addr[2];
        mcast[3] = pCurr->addr[3];
        mcast[4] = pCurr->addr[4];
        mcast[5] = pCurr->addr[5];


        /* Set up MAC address filter. */

        if ( !( HardwareAddrFilterAdd (phw, mcast) ) )
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xAddrFilterSet: exceed MAC address filter table.\n", 1, 2, 3, 4, 5, 6);

        pCurr = END_MULTI_LST_NEXT(pCurr);
    }
}

/*******************************************************************************
*
* ks884xEndPollReceive - routine to receive a packet in polled mode.
*
* This routine is called by a user to try and get a packet from the
* device. This routine return OK if it is successful in getting the packet
*
* RETURNS: OK or EAGAIN.
*/
STATUS ks884xEndPollReceive
(
    DRV_CTRL * pDrvCtrl,  /* points to a device control structure */
    M_BLK_ID   pMblk      /* mBlk chain stores received data */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    UINT32      uPort;
    M_BLK_ID 	pDIMblk;
	RXDESC     *CurrentDescriptor;
	UINT32      uFrameCtrl;
	UINT32      Length;
	int	        i;
    STATUS      status = OK;
    int         portIndex = 0;



    if ((pMblk->mBlkHdr.mFlags & M_EXT) != M_EXT)
    {
        return (EAGAIN);
    }

	DRV_LOG(DRV_DEBUG_POLL_RX, "ks884xEndPollReceive (unit=%d)\n", 
                                pDrvCtrl->unit, 2, 3, 4, 5, 6);


	i = DI->nRxDescNextAvail;
	CurrentDescriptor = &DI->pRxDescBaseAddr[i];
	KS_CACHE_INVALIDATE(CurrentDescriptor, sizeof(RXDESC));

    #if (0)
    printf("Rx Desc %02d addr:0x%08X, RDES0:0x%08X, RDES1:0x%08X, RDES2:0x%08X, RDES3:0x%08X\n", 
                i, CurrentDescriptor, 
                PCI_SWAP(CurrentDescriptor->RxFrameControl), 
                PCI_SWAP(CurrentDescriptor->RxDMABufLen), 
                rbdBuffAddrGet(pDrvCtrl, CurrentDescriptor), 
                CurrentDescriptor->RxDMANextPtr);
    #endif

    /* get the next available (host owned) receive descriptor */

	uFrameCtrl = PCI_SWAP (CurrentDescriptor->RxFrameControl);
	if ( uFrameCtrl & DESC_OWN_BIT ) /* Current Rx descriptor is not owned by host */
    {
		 DRV_LOG(DRV_DEBUG_POLL_RX, "ks884xEndPollReceive(): Current Rx descriptor is not owned by host\n", 
                                   1, 2, 3, 4, 5, 6);
         return (EAGAIN);
    }
	else 	                         
    /* 
     * Current Rx descriptor is owned by host 
     */
    {
	    /* if it is a valid Rx descriptor, the pMblk should point to a pMblk buffer!!! */
	    pDIMblk = DI->pRxMblk[i].pMblk;
	    if (pDIMblk == NULL) 
        {
		    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndPollReceive(): null pMblk\n", 
                                      1, 2, 3, 4, 5, 6);
            status = EAGAIN;
            goto CLEARUP;
        }

        /* 
         * precess receive packet 
         */

	    /* check error status of this transmit descriptor */
        if  ( rbdHasErrors (pDrvCtrl, CurrentDescriptor) )
        {
	        /* error happened!!! */
		    DRV_LOG(DRV_DEBUG_ERROR, "%s> error found (framectrl=0x%08x)\n", 
                __FUNCTION__, uFrameCtrl, 3, 4, 5, 6);

            END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
            status = EAGAIN;
            goto CLEARUP;
        } 

        /* get the packet length (including CRC bytes) */
		Length = (uFrameCtrl & RFC_FL_MASK);

		/* offset CRC bytes! */
		Length -= ETH_CRC_LENGTH;

        /* Get the source port that packet come from */
        uPort = (uFrameCtrl & DESC_RX_PORT_MASK) >> 20;
        phw->m_bPortRX = uPort;

	    /* make the packet data coherent */

	    KS_CACHE_INVALIDATE( pDIMblk->mBlkHdr.mData, Length );

        if ( phw->fDebugDumpRx )
        {
            /* for debug use */
            printf("RDES0:0x%08X, RDES1:0x%08X, RDES2:0x%08X, RDES3:0x%08X\n", 
                    PCI_SWAP(CurrentDescriptor->RxFrameControl), 
                    PCI_SWAP(CurrentDescriptor->RxDMABufLen), 
                    rbdBuffAddrGet(pDrvCtrl, CurrentDescriptor), 
                    (int)CurrentDescriptor->RxDMANextPtr);
	        ks884x_dump_packet( (UINT8 *)pDIMblk->mBlkHdr.mData, (Length+ETH_CRC_LENGTH), uPort, (DI->uDebugDumpPkt & DRV_DEBUG_DUMP_RX));
        }

        /* 
         * Copy data from received descripotr buffer to 'pMblk'.
         */

        pMblk->mBlkHdr.mLen    = Length;
        pMblk->mBlkHdr.mFlags |= M_PKTHDR;
        pMblk->mBlkPktHdr.len  = Length;
        bcopy ((char *)(pDIMblk->mBlkHdr.mData), (char *)(pMblk->mBlkHdr.mData), Length);

        END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_UCAST, +1);

        phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_RCV ]++;
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_RCV ] += Length;

    } /* if ( (uFrameCtrl & DESC_OWN_BIT) )  */


	/* 
	 * done with this descriptor, 
     * return the clean descriptor to the device's DMA controller.
   	 */

CLEARUP:

    rxDescSet ( pDrvCtrl, CurrentDescriptor );

    CACHE_PIPE_FLUSH ();

    /* Increment index to next descriptor */
	i = (++i % DI->nRxDescTotal);
	DI->nRxDescNextAvail = i;

    return (status);
}

/*******************************************************************************
*
* ks884xEndPollSend - send a packet in polled mode.
*
* This routine is called by a user to try and send a packet on the
* device.
*
* INTERNAL
* A return value of OK indicates that the packet has been sent.  In this
* case, the upper level protocol (WDB for example) will free the mBlk and
* associated cluster.
*
* RETURNS: OK or EAGAIN.
*/
STATUS ks884xEndPollSend
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
    M_BLK_ID   pMblk      /* mBlk chain stores transmitted data */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    UINT32     uPort;
    UINT32     txFrameCntlMask;
	TXDESC	  *pTxDesc;
	int		   i, dataLen;
	char	  *data;
    STATUS     status = OK;
	M_BLK_ID   pNewMblk = NULL;
	UCHAR	  *pCluster = NULL;
    ULONG      portIndex=0;



	DRV_LOG(DRV_DEBUG_POLL_TX, "ks884xEndPollSend (unit=%d)\n", 
                                pDrvCtrl->unit, 2, 3, 4, 5, 6);

	/* 
     * No available TxBufs 
     */
    if (DI->fTxNoResource) 
    {

        /* clean up processed tx descriptors */
        pDrvCtrl->txBlocked = TRUE;        /* transmitter not ready */
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_ALLOC_FAIL ]++;

        /* clean up processed tx descriptors */
	    ks884xTxDoneProcess(pDrvCtrl);

        return (EAGAIN);
	}

    if ( pDrvCtrl->fPollSendFromCLI )
        uPort = phw->m_bPortTX;
    else
        uPort = pDrvCtrl->unit;

    /*
     * Copy <pMblk> to a new <pNewMblk>, 
     * and send <pNewMblk>'s buffer data to device's transmit descriptor.
     */

	/* get a Tx Frame Descriptor */
	i = DI->nTxDescNextAvail;
	pTxDesc = &DI->pTxDescBaseAddr[i];

	/* invalidate to make consistency between device and host */
	KS_CACHE_INVALIDATE(pTxDesc, sizeof(TXDESC));

    #if (0)
    printf("Tx Desc %02d addr:0x%08X, TDES0:0x%08X, TDES1:0x%08X, TDES2:0x%08X, TDES3:0x%08X\n", 
                i, pTxDesc, 
                PCI_SWAP(pTxDesc->TxOwnBit), 
                PCI_SWAP(pTxDesc->TxFrameControl), 
                tbdBuffAddrGet(pDrvCtrl, pTxDesc), 
                pTxDesc->TxDMANextPtr);
    #endif

    /* Get New mblk/clblk/cluster */

	pNewMblk = ks884xAllocMblk(pDrvCtrl, &pCluster, NULL);
    if ( pNewMblk == NULL )
    {
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_DROPPED ]++;
		DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndPollSend(): ks884xAllocRxBuffers failed \n", 1, 2, 3, 4, 5, 6);
        return (EAGAIN);
    }

	/*
	 * If resources are not available, return.
	 * Do not free packet if use END_ERR_BLOCK.
	 */
	if ( ( PCI_SWAP(pTxDesc->TxOwnBit) & DESC_OWN_BIT ) )
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xEndPollSend(): Current Tx descriptor is not owned by host\n", 
                                   1, 2, 3, 4, 5, 6);
        return (EAGAIN);
    }

    /* 
     * copies all the data in the entire `mBlk' chain referenced in <pMblk> to 
     * the buffer referenced in 'pNewMblk'.
     */

    dataLen = netMblkToBufCopy ( pMblk, pNewMblk->mBlkHdr.mData, NULL );
	data    = pNewMblk->mBlkHdr.mData;
    pNewMblk->mBlkHdr.mLen += dataLen;


	DI->pTxMblk[i].pMblk = pNewMblk;

	/* set Tx DMA buffer address */
    tbdBuffAddrSet (pDrvCtrl, pTxDesc, data);

    /* set DMA tx descriptor control */
    txFrameCntlMask = txDescGet (pDrvCtrl, pTxDesc);

    txFrameCntlMask &= (0x1E000000);
	txFrameCntlMask |= (dataLen & TFC_TBS_MASK) | ( TFC_IC | TFC_FS | TFC_LS) ;
    txFrameCntlMask |= (uPort << 20);                /* Set destination port */

    txDescSet ( pDrvCtrl, pTxDesc, txFrameCntlMask );

	/* flush write pipe */
	CACHE_PIPE_FLUSH();

    phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;

    if ( phw->fDebugDumpTx )
    {
       printf("TDES0:0x%08X, TDES1:0x%08X, TDES2:0x%08X, TDES3:0x%08X\n", 
                PCI_SWAP(pTxDesc->TxOwnBit), 
                PCI_SWAP(pTxDesc->TxFrameControl), 
                tbdBuffAddrGet(pDrvCtrl, pTxDesc), 
                (int)pTxDesc->TxDMANextPtr);
       printf("Tx Buffer:0x%08X\n", (int)data);
       ks884x_dump_packet( (UINT8 *)data, dataLen, uPort, (DI->uDebugDumpPkt & DRV_DEBUG_DUMP_TX)); 
    }

	DI->nTransmitCount++;

	/* update pending transimt packet count */
	DI->nTxDescAvail--;

	if (DI->nTxDescAvail < 1) 
    {
	    /* this will block next transmit until we clear the flag */
	    DI->fTxNoResource = TRUE;
	    DRV_LOG(DRV_DEBUG_POLL_TX, "ks884xEndPollSend(): Oops, no Tx Resource (unit=%d), nTxDescAvail=%d \n", 
                    pDrvCtrl->unit, DI->nTxDescAvail, 3, 4, 5, 6);
    }

	/* adv to next available descriptor */
	DI->nTxDescNextAvail = ++DI->nTxDescNextAvail % DI->nTxDescTotal;


	/* kick off Tx transfer */
	HW_WRITE_DWORD ( phw, REG_DMA_TX_START, DMA_START ); 

	/* flush write pipe */
	CACHE_PIPE_FLUSH();

	DRV_LOG(DRV_DEBUG_POLL_TX, "ks884xEndPollSend (unit=%d): kick off Tx transfer.\n", 
                                pDrvCtrl->unit, 2, 3, 4, 5, 6);

    /* increment the unicast output counter */
    END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_UCAST, +1);
    phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;

    return (status);
}

/*****************************************************************************
*
* ks884xEndMCastAddrAdd - add a multicast address for the device
*
* This routine adds a multicast address to whatever the driver
* is already listening for.  It then resets the address filter.
*
* NOTE
* etherMultiAdd() returns EINVAL for invalid multicast addresses.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xEndMCastAddrAdd
(
    DRV_CTRL * pDrvCtrl,  /* pointer to DRV_CTRL structure */
    char *     pAddress   /* address to be added */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;

    /*
     * ENETRESET is returned to inform the driver that the list has
     * changed and its reception filter should be adjusted accordingly.
     */
 
    int retVal = etherMultiAdd (&pDrvCtrl->endObj.multiList, pAddress);

    if (retVal == ENETRESET)
    {
        /* edit Logical Addresss Filter Registers */

       if ( HardwareAddrFilterAdd (phw, pAddress) )
          retVal = OK;
       else
          retVal = ERROR;
    }

    return ((retVal == OK) ? OK : ERROR);
}

/*****************************************************************************
*
* ks884xEndMCastAddrDel - delete a multicast address for the device
*
* This routine removes a multicast address from whatever the driver
* is listening for.  It then resets the address filter.
*
* NOTE
* etherMultiAdd() returns EINVAL for invalid multicast addresses.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xEndMCastAddrDel
(
    DRV_CTRL * pDrvCtrl,  /* pointer to DRV_CTRL structure */
    char *     pAddress   /* address to be deleted */
)
{
    PHARDWARE         phw = pDrvCtrl->pHW;
    /*
     * ENETRESET is returned to inform the driver that the list has
     * changed and its reception filter should be adjusted accordingly.
     */
 
    int retVal = etherMultiDel (&pDrvCtrl->endObj.multiList, pAddress);

    if (retVal == ENETRESET)
    {
        /* edit Logical Addresss Filter Registers */

       if ( HardwareAddrFilterDel (phw, pAddress) )
           retVal = OK;
       else
           retVal = ERROR;
    }

    return ((retVal == OK) ? OK : ERROR);
}

/*****************************************************************************
*
* ks884xMCastAddrGet - get the multicast address list for the device
*
* This routine gets the multicast list of whatever the driver
* is already listening for.
*
* RETURNS: OK/ERROR
*/
LOCAL STATUS ks884xMCastAddrGet
    (
    DRV_CTRL * 	pDrvCtrl,  /* pointer to DRV_CTRL structure */
    MULTI_TABLE * 	pTable     /* table where address will be copied */
    )
    {
    return (etherMultiGet (&pDrvCtrl->endObj.multiList, pTable));
    }

/*******************************************************************************
*
* ks884xEndStop - stop the device
*
* This function calls BSP functions to disconnect interrupts and stop
* the device from operating in interrupt mode.
*
*
* RETURNS: OK or ERROR if the interrupt handler could not be disconnected.
*/
LOCAL STATUS ks884xEndStop
(
    DRV_CTRL * pDrvCtrl  /* device control structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    STATUS     result = OK;



	DRV_LOG(DRV_DEBUG_STOP, "ks884xEndStop (unit=%d) is called.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

	if (NULL != pDrvCtrl->watchdog) 
    {
		wdDelete(pDrvCtrl->watchdog);
		pDrvCtrl->watchdog = NULL;
	}

    /* Disable interrupts on the device-side */
    HardwareDisableInterrupt ( phw );

    /* Disable system level interrupt */
    (* ks884xIntDisable) (pDrvCtrl->ilevel);

	/* Stop MAC DMA */
    HardwareDisable ( phw );

#ifdef DEF_KS8842
    /* Disable ks884x swtich */
    SwitchEnable ( phw, FALSE );
#endif

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    /* mark the interface as down */

    END_FLAGS_CLR (&pDrvCtrl->endObj, (IFF_UP | IFF_RUNNING));

    /* disconnect the interrupt handler */

    result = (* ks884xIntDisconnect) ((VOIDFUNCPTR *)
             INUM_TO_IVEC (pDrvCtrl->inum), ks8842xInt, (int)(pDrvCtrl));


    return (result);
}

/******************************************************************************
*
* ks884xEndUnload - unload a driver from the system
*
* This function first brings down the device, and then frees any
* stuff that was allocated by the driver in the load function. The controller
* structure should be free by who ever is calling this function.
*
* RETURNS: OK
*/
LOCAL STATUS ks884xEndUnload
(
    DRV_CTRL * pDrvCtrl  /* device control structure */
)
{


 	DRV_LOG(DRV_DEBUG_UNLOAD, "ks884xEndUnload(unit=%d).\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    END_OBJECT_UNLOAD (&pDrvCtrl->endObj);

    /* Free device's Transmit\Receive Descriptors */

	ks884xFreeTxBDs(pDrvCtrl);
	ks884xFreeRxBDs(pDrvCtrl);

    /* Free the shared DMA memory. */

    if (pDrvCtrl->flags & KS_MEM_ALLOC_FLAG)
        cacheDmaFree ((void *)(pDrvCtrl->pShMem));

    /* Free the shared DMA memory allocated for clusters */
    
    if (pDrvCtrl->flags & KS_MEM_ALLOC_FLAG)
        cacheDmaFree (pDrvCtrl->clDesc.memArea);

    /* Free the memory allocated for mBlks and clBlks */
    
    if (pDrvCtrl->mClCfg.memArea != NULL)
        free (pDrvCtrl->mClCfg.memArea);

    /* Free the memory allocated for driver pool structure */

    if (pDrvCtrl->endObj.pNetPool != NULL)
        free (pDrvCtrl->endObj.pNetPool);

    return (OK);
}

/*******************************************************************************
*
* ks884xEndPollStart - start polled mode operations
*
* RETURNS: OK always.
*/
STATUS ks884xEndPollStart
(
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;

 

    /* Disable interrupts on the device-side */
    HardwareDisableInterrupt ( phw );

    /* Clear interrupts on the device-side */
    HardwareAcknowledgeInterrupt( phw, phw->m_ulInterruptMask );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    pDrvCtrl->flags |= KS_POLLING;

    DRV_LOG (DRV_DEBUG_POLL, "POLLING STARTED\n", 0, 0, 0, 0, 0, 0);

    return (OK);
}

/*******************************************************************************
*
* ks884xEndPollStop - stop polled mode operations
*
* RETURNS: OK always.
*/
STATUS ks884xEndPollStop
(
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    int intLevel = intLock();


    /* Clear interrupts on the device-side */
    HardwareAcknowledgeInterrupt( phw, phw->m_ulInterruptMask );

    /* Enable interrupts on the device-side */
    HardwareEnableInterrupt( phw );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    pDrvCtrl->flags &= ~KS_POLLING;

    intUnlock (intLevel);

    DRV_LOG (DRV_DEBUG_POLL, "POLLING STOPPED\n", 0, 0, 0, 0, 0, 0);

    return (OK);
}

/******************************************************************************
*
* ks884xHandleTxDoneInt - clean the transmit ring
*
* This routine is called at task level indirectly by the interrupt
* service routine to  clean the transmit ring.

* This routine frees driver clusters associated with Tx descriptor that have used
* the clusters for transmitting packets.  
*
* RETURNS: N/A
*
* NOMANUAL
*/
LOCAL void ks884xHandleTxDoneInt
(
    DRV_CTRL * pDrvCtrl,  /* pointer to DRV_CTRL structure */
    UINT32     uIER       /* holds INT_TX_DONE bit read from interrupt status */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;


    DRV_LOG(DRV_DEBUG_TX_DONE, "ks884xHandleTxDoneInt (%d), uIER=0x%08X\n", pDrvCtrl->unit, uIER, 3, 4, 5, 6);

    pDrvCtrl->txCleaning = TRUE; 
    
    END_TX_SEM_TAKE (&pDrvCtrl->endObj, WAIT_FOREVER);

    /* clean up processed tx descriptors */
	ks884xTxDoneProcess(pDrvCtrl);

    pDrvCtrl->txCleaning = FALSE;

    END_TX_SEM_GIVE (&pDrvCtrl->endObj);

	/* enable Tx interrupts */
    HardwareEnableInterruptBit ( phw, uIER ); 

}


/*******************************************************************************
*
* ks884xTxDoneProcess - 
*
* Clear the Transmit descriptors when chip has completed the transmit the packets.
*
* RETURNS: N/A.
*/
LOCAL void ks884xTxDoneProcess
(
    DRV_CTRL * pDrvCtrl  /* points to device control structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
	int     i;
	int     oldLevel = 0;
	TXDESC *TransmitDescriptor;


	i = DI->nTxDescUsed;
	TransmitDescriptor = &DI->pTxDescBaseAddr[i];
    KS_CACHE_INVALIDATE(TransmitDescriptor, sizeof(TXDESC));

	while ( !( PCI_SWAP(TransmitDescriptor->TxOwnBit) & DESC_OWN_BIT)  && (DI->nTransmitCount > 0) ) 
    {

#ifndef TEST_THROUGHPUT
		if (DI->pTxMblk[i].pMblk) 
        {
			netMblkClChainFree(DI->pTxMblk[i].pMblk);
			DI->pTxMblk[i].pMblk = NULL;
		}
		else 
        {                                                                                            
			DRV_LOG(DRV_DEBUG_TX_DONE, "ks884xTxDoneProcess(): Oops, TXBD(%d) with null pMblk\n", i, 2, 3, 4, 5, 6); 
		}

#else  /* #ifndef TEST_THROUGHPUT */
        /* need to free buffer if doing the loopback under TEST_THROUGHPUT */
        if ( phw->fLoopbackStart )
        {

		    if (DI->pTxMblk[i].pMblk) 
            {
			    netMblkClChainFree(DI->pTxMblk[i].pMblk);
			    DI->pTxMblk[i].pMblk = NULL;
            }
        }

#endif /* #ifndef TEST_THROUGHPUT */

        oldLevel = intLock();	/* protect it */

		DI->nTxDescAvail++;
		DI->nTransmitCount--;

		/* increase to next Tx descriptor */
		i = ++i % DI->nTxDescTotal;
		TransmitDescriptor = &DI->pTxDescBaseAddr[i];
    	KS_CACHE_INVALIDATE(TransmitDescriptor, sizeof(TXDESC));
		DI->nTxProcessedCount++;

		intUnlock(oldLevel);

	}

	DI->nTxDescUsed = i;

	if ( DI->nTxDescAvail > 0 ) 
		DI->fTxNoResource = FALSE;

}


/*******************************************************************************
*
* ks884xParamInit - set global parameters for endianess and interrupt level
*
* ???
*
* RETURNS: OK Always.
*
* NOMANUAL
*/
STATUS ks884xParamInit
(
    int pciBusEndianess,
    int intLevel
)
{
    ks884xBusEndianess   = pciBusEndianess;
    ks884xInterruptLevel = intLevel;

    return (OK);
}


/*
 * ks884xAutoNegoStart
 *	This function is used to start auto negotiation
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xAutoNegoStart
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
	PORT_KS_INFO *DPI,     /* ks884x port related (switch port only) */
    ULONG      uPort
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;


    /* 
     * auto-negotiation mode 
     */
	if (SW_PHY_AUTO == DPI->myLinkType) 
    {
        /* if auto-negotiation not in progress, and Link state has changed */
	    if ( !DPI->bAutoNegoInProgress && DPI->bLinkChanged) 
        {
            /* Start auto-negotiation on this port */
		    SwitchRestartAutoNego( phw );

           	DPI->bAutoNegoInProgress = TRUE;
		    DPI->bLinkChanged = FALSE;
		    DPI->bLinkActive  = FALSE;
        }
    }

	ks884xGetPhyStatus ( pDrvCtrl, DPI, uPort );
}

/*
 * ks884xGetPhyStatus
 *	This function is used to get the status of auto negotiation.
 *
 * Argument(s)
 *	pDrvCtrl	pointer to DRV_CTRL structure.
 *
 * Return(s)
 *	TRUE	link up
 *	FALSE	link down
 */
LOCAL STATUS ks884xGetPhyStatus
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
	PORT_KS_INFO *DPI,     /* ks884x port related (switch port only) */
    ULONG      uPort
)
{
    PHARDWARE   phw  = pDrvCtrl->pHW;
    PPORT_INFO  pPortInfo = NULL;


    DRV_LOG(DRV_DEBUG_PHY, "ks884xGetPhyStatus\n", 1, 2, 3, 4, 5, 6);

    pPortInfo = &phw->m_PortInfo[ uPort ];

    if ( (phw == NULL) || (pPortInfo == NULL) )
    {
        DRV_LOG(DRV_DEBUG_ERROR, "ks884xGetPhyStatus(): phw or pPortInfo is NULL.\n", 1, 2, 3, 4, 5, 6);
        return FALSE;
    }

    /* Get port's PHY link status */
    SwitchGetLinkStatus( phw );

	/* 
     * PHY link is down 
     */

    if (pPortInfo->ulHardwareState == MediaStateDisconnected ) 
    {
		DPI->bLinkActive = FALSE;
		DPI->myLinkSpeed = SPEED_UNKNOWN;
		DPI->myLinkDuplex = 0;

        /* TODO : set link LED */

		DRV_LOG(DRV_DEBUG_PHY, "Link down.\n", 1, 2, 3, 4, 5, 6);
		return FALSE;
	}

	/* 
     * PHY link is up
     */

	DPI->bLinkActive = TRUE;

    /* 
     * auto-negotiation mode 
     */
	if (SW_PHY_AUTO == DPI->myLinkType) 
    {

        /* Link is full duplex */ 
		if (pPortInfo->bDuplex == FULL_DUPLEX ) 
        {
            /* force enable flow control, regardless auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, TRUE );

            /* disable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, FALSE );
        } 
        /* Link is half duplex */ 
        else 
        {
            /* flow control is disable based on auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, FALSE );

            /* enable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, TRUE );
		}

		DRV_LOG(DRV_DEBUG_PHY, "Link up.\n", 1, 2, 3, 4, 5, 6);
	}
	/* 
     * manually connection mode 
     */
	else   /* if (SW_PHY_AUTO == DI->myLinkType[uPort]) */
    {
		if ( (SW_PHY_10BASE_T_FD == DPI->myLinkType) || 
             (SW_PHY_100BASE_TX_FD == DPI->myLinkType ) )
        {
            /* manually select link for full duplex */ 
			DPI->myLinkDuplex = FULL_DUPLEX;

            /* force enable flow control, regardless auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, TRUE );

            /* disable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, FALSE );
		}
		else 
        {
            /* manually selected link for half duplex */ 
			DPI->myLinkDuplex = HALF_DUPLEX;

            /* flow control is enable based on auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, FALSE );

            /* enable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, TRUE );
		}

		if  ( (SW_PHY_100BASE_TX_FD == DPI->myLinkType) || 
              (SW_PHY_100BASE_TX    == DPI->myLinkType) )
        {
            /* manually selected link speed is 100Mbps */
			DPI->myLinkSpeed = SPEED_100;

            /* TODO : set link LED */
		}
		else 
        {
            /* manually selected link speed is 10Mbps */
			DPI->myLinkSpeed = SPEED_10;

            /* TODO : set link LED */
		}

    } /* if (SW_PHY_AUTO == DPI->myLinkType) */

	return TRUE;
}


/*******************************************************************************
*
* ks884xEnetAddrGet - get ks884x Ethernet (IEEE station) address
*
* This routine provides an interface for accessing an ks884x device
* Ethernet (IEEE station address) address in the device.
* A handle to the specific device control structure is specified
* in the <pDrvCtrl> parameter.  The 6-byte IEEE station address will be
* copied to the memory location specified by the <enetAdrs> parameter.
*
*
* RETURNS: OK, or ERROR if could not be obtained.
*
* NOMANUAL
*/
STATUS ks884xEnetAddrGet
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
    char *     enetAdrs    /* specifies storage for station address */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    UCHAR         i;

	/* 
     * Get MAC address to the Station registers 
     */

    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xEnetAddrGet(): calling HardwareReadAddress()... \n"));

	/* Get MAC address from the Station registers */
    HardwareReadAddress( phw );

#if (0)
    	printf("ks884xEnetAddrGet() m_bOverrideAddress MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", 
		    phw->m_bOverrideAddress[0], 
		    phw->m_bOverrideAddress[1],
		    phw->m_bOverrideAddress[2],
		    phw->m_bOverrideAddress[3], 
		    phw->m_bOverrideAddress[4],
		    phw->m_bOverrideAddress[5]); 

#endif

#if (0)
    	printf("ks884xEnetAddrGet() m_bPermanentAddress MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", 
		    phw->m_bPermanentAddress[0], 
		    phw->m_bPermanentAddress[1],
		    phw->m_bPermanentAddress[2],
		    phw->m_bPermanentAddress[3], 
		    phw->m_bPermanentAddress[4],
		    phw->m_bPermanentAddress[5]); 
#endif
    memcpy ( enetAdrs, phw->m_bPermanentAddress, ETH_LENGTH_OF_ADDRESS );

    return OK;
}

/*******************************************************************************
*
* ks884xEnetAddrSet - Set ks884x Ethernet (IEEE station) address
*
* This routine provides an interface for accessing an ks884x device
* Ethernet (IEEE station address) address in the device.
* A handle to the specific device control structure is specified
* in the <pDrvCtrl> parameter.  The 6-byte IEEE station address,
* specified by the <enetAdrs> parameter, will be wrote to ks884x device
* MAC address registers .
*
*
* RETURNS: OK, or ERROR if could not be obtained.
*
* NOMANUAL
*/
STATUS ks884xEnetAddrSet
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
    char *     enetAdrs    /* specifies storage for station address */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    UCHAR         i;


	/* 
     * Set MAC address to the Station registers 
     */

    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xEnetAddrSet(): calling HardwareSetAddress()... \n"));

    phw->m_bMacOverrideAddr = TRUE;
    for (i=0; i<MAC_ADDRESS_LENGTH; i++)
         phw->m_bOverrideAddress[i] = enetAdrs[i];
    HardwareSetAddress( phw );


	/* Get MAC address from the Station registers */
    HardwareReadAddress( phw );

    if (memcmp ( enetAdrs, phw->m_bOverrideAddress, ETH_LENGTH_OF_ADDRESS ) != 0)
        return ERROR;

	memcpy(pDrvCtrl->enetAddr, phw->m_bOverrideAddress, ETH_LENGTH_OF_ADDRESS);  

    return OK;
}


/*
 * ks884xSetDefaults
 *	This function is used to set initial defaults before Parse routine.
 *
 * Argument(s)
 *	pDrvCtrl		pointer to DRV_CTRL structure.
 *
 * Return(s)
 *	NONE.
 */

LOCAL void ks884xSetDefaults
(
    DRV_CTRL * pDrvCtrl   /* Driver control structure */
)
{
    ULONG      uPort;
    PHARDWARE  phw = pDrvCtrl->pHW;
	PORT_KS_INFO *DPI;      /* ks884x port related (switch port only) */



	/* Transmit Descriptor Count */
	DI->nTxDescTotal = TXDESC_DEFAULT;		

	/* Receive Descriptor Count */
	DI->nRxDescTotal = RXDESC_DEFAULT;		

	/* Transmit Checksum Enable configuration */
	DI->bTxChecksum = TXCHECKSUM_DEFAULT;	/* enabled */

	/* Receive Checksum Enable */
	DI->bRxChecksum = RXCHECKSUM_DEFAULT;	/* enabled */

    /* Flow Control Enable configuration */
	DI->bRxFlowCtrl = FLOWCONTROL_DEFAULT;	/* enabled */


	/* currently Tx control flow shares the setting of Rx control flow */
	DI->bTxFlowCtrl = DI->bRxFlowCtrl;

	/* Tx coalescing, currently can only be used if buffer unavailable bit is set */
	DI->nTransmitCoalescing = (DI->nTxDescTotal >> 3);

    /* Receive buffer size */
	DI->uRxBufferLen = BUFFER_2044;	

	/* Programmable Burst Length */
	DI->byTxPBL = PBL_DEFAULT;				/* Tx PBL size */
	DI->byRxPBL = PBL_DEFAULT;				/* Rx PBL size */

	/* Perform PHY PowerDown Reset instead of soft reset, the Option function can 
	   be overwritten by user later */
	DI->bPowerDownReset = PHY_POWERDOWN_DEFAULT;

    /* Set device interrupt mask */
 	DI->uIntMask = INT_MASK;

	/* for debug only */
    DI->uDebugDumpPkt = DUMPPACKET_DEFAULT;

    /*
     * Set switch MAC address.
     */

	/* set default mac address for WAN, with Kendin's manufacturer ID 00:10:A1 */
	memcpy(DI->stMacStation, "\0\x10\xa1PCI", ETH_LENGTH_OF_ADDRESS);
#if (0)
	memset(DI->stMacStation, 0, ETH_LENGTH_OF_ADDRESS);
    DI->stMacStation[5] = 0x03;
#endif

    /*
     * Set default values for Port relative.
     */

    for (uPort=0; uPort<MAX_SWITCH_PORT; uPort++)
    {
        DPI = &pDrvCtrl->port[uPort];

	    /* PHY auto-negatiation */
	    DPI->myLinkType = PHY_SPEED_DEFAULT;

        /* disable Spanning Tree */
	    DPI->bySTPenable = PORT_STP_DEFAULT;

        /* enable broadcast storm protection */
	    DPI->byStormProtection = PORT_STORM_DEFAULT;

    } /* for (uPort=0; uPort<MAX_SWITCH_PORT; uPort++) */

    /*
     * Set default values for CPU Port relative.
     */

    /* TODO : */


    /* watchdog periode timer (every 5 sec) */
	pDrvCtrl->nWatchdogDelay = sysClkRateGet() * 5;		

    /* Set default call ks884xEndSend not from CLI command */
    pDrvCtrl->fPollSendFromCLI = FALSE;

    /* Set default for HW struct */
    phw->fDebugDumpTx = FALSE;
    phw->fDebugDumpRx = FALSE;

    phw->f_dircetMode = 1;   /* default transmit to device by MAC lookup mode */
    phw->m_bPortTX = 0;

}

/*
 * ks884xAllocTxBDs
 *	This function is used to allocate pTxMblk and also get Tx descriptor pointer from pShMem.
 *
 * Argument(s)
 *  pDrvCtrl		pointer to pDrvCtrl struct
 *
 * Return(s)
 *  OK		if success
 *	ERROR	if failed
 */
LOCAL STATUS ks884xAllocTxBDs
(
    DRV_CTRL * pDrvCtrl	/* device ID */
)
{
	int      i, oldLevel = 0;
	int      size;
	UINT32	 uPA;
	volatile TXDESC	*pTxDesc;
    UINT32   txFrameCntlMask;


	DRV_LOG(DRV_DEBUG_LOAD, "AllocateTxDescriptors\n", 1, 2, 3, 4, 5, 6);

	/* allocate buffers for keeping track of mblk pointer submitted from Send routine */

	size = sizeof(KS_MBLK) * DI->nTxDescTotal;
	DI->pTxMblk = (KS_MBLK *)calloc(size, 1);
	if (NULL == DI->pTxMblk) 
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xAllocTxBDs: AllocateTxDescriptors failed\n", 1, 2, 3, 4, 5, 6);
		return ERROR;
	}

    /*  Set begin of Transmit decsripotrs address */

	DI->pTxDescBaseAddr = (TXDESC *)((char *)pDrvCtrl->pShMem);

	/* align it to cache line boundary */
	(UINT32)DI->pTxDescBaseAddr &= ~(_CACHE_ALIGN_SIZE - 1);


    /* 
     * Initialize the Tx descriptors 
     */

    oldLevel = intLock();	/* protect it */
	pTxDesc = DI->pTxDescBaseAddr;
	uPA = (UINT32)DI->pTxDescBaseAddr;

	for (i = 0; i < DI->nTxDescTotal - 1; i++, pTxDesc++) 
    {
		uPA += TXBD_SIZE;		             /* pointer to next TxBD */
		pTxDesc->TxDMANextPtr = PCI_SWAP(uPA);
        pTxDesc->TxFrameControl = PCI_SWAP ( txFrameCntlMask ); 
		KS_CACHE_FLUSH(pTxDesc, sizeof(TXDESC));

        DRV_LOG (DRV_DEBUG_BDs, "Tx Desc %d addr: 0x%08X, NextPtr: 0x%08X\n", 
                                  i, pTxDesc, pTxDesc->TxDMANextPtr, 4, 5, 6);
	}
	pTxDesc->TxDMANextPtr = (UINT32)PCI_SWAP(DI->pTxDescBaseAddr);  /* last TxBD point to 1st TxBD */
    pTxDesc->TxFrameControl = PCI_SWAP ( txFrameCntlMask ); 
	KS_CACHE_FLUSH(pTxDesc, sizeof(TXDESC));
	intUnlock(oldLevel);

    DRV_LOG (DRV_DEBUG_BDs, "Tx Desc %d addr: 0x%08X, NextPtr: 0x%08X\n", 
                                  i, pTxDesc, pTxDesc->TxDMANextPtr, 4, 5, 6);


    /* initialize others house work */

	DI->nTxDescAvail = DI->nTxDescTotal;
	DI->nTxDescNextAvail = 0;
	DI->nTxDescUsed = 0;
	DI->nTransmitCount = 0;
	DI->nTxProcessedCount = 0;
	DI->fTxNoResource = 0;

    return OK;
}

/*
 * ks884xAllocRxBDs
 *	This function is used to setup Rx descriptors, also call ks884xAllocRxBuffers()
 *  to allocate receive buffers with new mblk/clblk/cluster to Rx descriptors.
 *
 * Argument(s)
 *  pDrvCtrl		pointer to DRV_CTRL struct
 *
 * Return(s)
 *  OK		if success
 *	ERROR	if failed
 */
LOCAL STATUS ks884xAllocRxBDs
(
    DRV_CTRL * pDrvCtrl	/* device ID */
)
{
	int      i, oldLevel = 0;
	int      size;
	UINT32	 uPA;
	volatile RXDESC	*pRxDesc;

	DRV_LOG(DRV_DEBUG_LOAD, "AllocateRxDescriptors\n", 1, 2, 3, 4, 5, 6);

	/* allocate buffers for keeping track of Rx mblk pointer */

	size = sizeof(KS_MBLK) * DI->nRxDescTotal;
	DI->pRxMblk = (KS_MBLK *)calloc(size, 1);
	if (DI->pRxMblk == NULL) {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xAllocRxBDs: AllocateRxDescriptors failed\n", 1, 2, 3, 4, 5, 6);
		return ERROR;
	}

    /*  Set begin of Receive decsripotrs address */

	size = DI->nTxDescTotal * TXBD_SIZE + _CACHE_ALIGN_SIZE;
	DI->pRxDescBaseAddr = (RXDESC *)((char *)DI->pTxDescBaseAddr + size);
	(UINT32)DI->pRxDescBaseAddr &= ~(_CACHE_ALIGN_SIZE - 1);

    /* 
     * Initialize the Rx descriptors 
     */

    oldLevel = intLock();	/* protect it */
	pRxDesc = DI->pRxDescBaseAddr;
	uPA = (UINT32)DI->pRxDescBaseAddr;
	for (i = 0; i < DI->nRxDescTotal - 1; i++, pRxDesc++) 
    {
	    KS_CACHE_INVALIDATE(pRxDesc, sizeof(RXDESC));     
		uPA += RXBD_SIZE;		               /* pointer to next RxBD */
		pRxDesc->RxDMANextPtr = PCI_SWAP(uPA); /* set Rx decsciptor's next descriptor address */
		KS_CACHE_FLUSH(pRxDesc, sizeof(RXDESC));

        DRV_LOG (DRV_DEBUG_BDs, "Rx Desc %d addr: 0x%08X, NextPtr: 0x%08X\n", 
                                  i, pRxDesc, pRxDesc->RxDMANextPtr, 4, 5, 6);
	}
	pRxDesc->RxDMANextPtr = (UINT32)PCI_SWAP(DI->pRxDescBaseAddr); /* last RxBD point to 1st RxBD */

	KS_CACHE_FLUSH(pRxDesc, sizeof(RXDESC));
	intUnlock(oldLevel);

    DRV_LOG (DRV_DEBUG_BDs, "Rx Desc %d addr: 0x%08X, NextPtr: 0x%08X\n", 
                                  i, pRxDesc, pRxDesc->RxDMANextPtr, 4, 5, 6);

    /* initialize others house work */

	DI->nRxDescNextAvail = 0;
	DI->nRxDescNextToFill = 0;

    /* 
     * Initialize the Rx buffer in the Rx descriptors 
     */
    ks884xAllocRxBuffers (pDrvCtrl);


    return OK;
}

/*
 * ks884xAllocRxBuffers
 *	This function is use to allocate receive buffers with new mblk/clblk/cluster
 *	to Rx descriptors, and call rxDescSet() to set the receive control (RDES1), 
 *  and the receive control (RDES0) OWN bit to be owned by ks884x..
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xAllocRxBuffers
(  
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
	M_BLK_ID pMblk;
	RXDESC	*CurrentDescriptor;
	UCHAR	*pCluster = NULL;
	int i, oldLevel = 0;

	DRV_LOG(DRV_DEBUG_LOAD, "ks884xAllocRxBuffers\n", 1, 2, 3, 4, 5, 6);


	i = DI->nRxDescNextToFill;

    oldLevel = intLock();	/* protect it */
	while (NULL == DI->pRxMblk[i].pMblk) 
    {
		CurrentDescriptor = &DI->pRxDescBaseAddr[i];
	    KS_CACHE_INVALIDATE(CurrentDescriptor, sizeof(RXDESC));

		pMblk = ks884xAllocMblk(pDrvCtrl, &pCluster, NULL);
		if (NULL == pMblk) 
        {
			DRV_LOG(DRV_DEBUG_ERROR, "ks884xAllocRxBuffers failed \n", 1, 2, 3, 4, 5, 6);
			break;
		}

		DI->pRxMblk[i].pMblk = pMblk;
		memset(pCluster, 0, DI->uRxBufferLen);

		/* 
         * setup Rx descriptor 
         */

        /* set Rx buffer to Rx decsciptor */
        rbdBuffAddrSet (pDrvCtrl, CurrentDescriptor, pCluster);
        CurrentDescriptor->RxFrameControl= PCI_SWAP ( DESC_OWN_BIT );
     	CurrentDescriptor->RxDMABufLen   = PCI_SWAP ( DI->uRxBufferLen );
	    KS_CACHE_USER_FLUSH(pCluster, DI->uRxBufferLen);

        DRV_LOG (DRV_DEBUG_BDs, "Rx Desc %d addr:0x%08X, RxDMABufAddr:0x%08X, pCluster:0x%08X\n", 
                                  i, CurrentDescriptor,  rbdBuffAddrGet (pDrvCtrl,CurrentDescriptor), 
                                  pCluster, 5, 6);

		i = ++i % DI->nRxDescTotal;
	}

#if (0)
    i = 31;
	CurrentDescriptor = &DI->pRxDescBaseAddr[i];
    printf("Rx Desc %02d addr:0x%08X, RDES0:0x%08X, RDES1:0x%08X, RDES2:0x%08X, RDES3:0x%08X\n", 
                i, (int)CurrentDescriptor, 
                PCI_SWAP(CurrentDescriptor->RxFrameControl), 
                PCI_SWAP(CurrentDescriptor->RxDMABufLen), 
                (int)rbdBuffAddrGet(pDrvCtrl, CurrentDescriptor), 
                (int)CurrentDescriptor->RxDMANextPtr);
    i=0;
#endif

	DI->nRxDescNextToFill = i;
	intUnlock(oldLevel);

}

/*
 * ks884xFreeTxBDs
 *	This function is used to free Tx descriptors.
 *
 * Argument(s)
 *  pDrvCtrl		pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xFreeTxBDs
(
    DRV_CTRL * pDrvCtrl	/* device ID */
)
{
	DRV_LOG(DRV_DEBUG_UNLOAD, "ks884xFreeTxBDs\n", 1, 2, 3, 4, 5, 6);

	CleanTxRing(pDrvCtrl);

	if (DI->pTxMblk) 
    {
		cfree((char *)DI->pTxMblk);
		DI->pTxMblk = NULL;
	}

	DI->pTxDescBaseAddr = NULL;
}

/*
 * ks884xFreeRxBDs
 *	This function is used to free Rx resources.
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xFreeRxBDs
(
    DRV_CTRL * pDrvCtrl	/* device ID */
)
{
	DRV_LOG(DRV_DEBUG_UNLOAD, "ks884xFreeRxBDs\n", 1, 2, 3, 4, 5, 6);

	CleanRxRing(pDrvCtrl);

	if (DI->pRxMblk) {
		cfree((char *)DI->pRxMblk);
		DI->pRxMblk = NULL;
	}

	DI->pRxDescBaseAddr = NULL;
}


/*
 * CleanTxRing
 *	This function is used to go through Tx descriptor list, 
 *  clean up Tx descripotor, and any pending resources.
 *
 * Argument(s)
 *  pDrvCtrl		pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void CleanTxRing
(
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
	int i;
	volatile TXDESC	*pTxDesc = DI->pTxDescBaseAddr;

	/* free pending mblk if any */
	for (i = 0; i < DI->nTxDescTotal; i++, pTxDesc++) 
    {
		if (NULL != DI->pTxMblk[i].pMblk) 
			netMblkClChainFree(DI->pTxMblk[i].pMblk);

		/* reset corresponding Tx Descriptor structure as well */
		pTxDesc->TxDMABufAddr = 0;
		pTxDesc->TxOwnBit = 0;
		pTxDesc->TxFrameControl = 0;
	}

	DI->nTransmitCount = 0;
	DI->nTxProcessedCount = 0;
	DI->nTxDescAvail = DI->nTxDescTotal;
	DI->nTxDescNextAvail = 0;
	DI->nTxDescUsed = 0;

}

/*
 * CleanRxRing
 *	This function is used to go through Rx descriptor list,
 *  clean up Rx descripotor, and free its associate mblk/clblk/cluster.
 *
 * Argument(s)
 *  pDrvCtrl	pointer to END_DEVICE struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void CleanRxRing
(
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
	int i;
	volatile RXDESC	*pRxDesc = DI->pRxDescBaseAddr;

	/* Free pending sk_buffs if any */
	for (i = 0; i < DI->nRxDescTotal; i++, pRxDesc++) 
    {
		if (DI->pRxMblk[i].pMblk != NULL) 
			netMblkClChainFree(DI->pRxMblk[i].pMblk);

		/* reset corresponding Rx Descriptor structure as well */
		pRxDesc->RxFrameControl = 0;
		pRxDesc->RxDMABufAddr  = 0;
		pRxDesc->RxDMABufLen   = 0;
	}

	DI->nRxDescNextAvail  = 0;
	DI->nRxDescNextToFill = 0;

}


/*
 * ks884xAllocMblk
 *	This function is use to replace used receive buffers with new mblk/clblk/cluster
 *	to Rx descriptors.
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *	pCluster	pointer to cluster data buffer associated with the mblk/clblk/cluster
 *	pSrcMblk	pointer to mblk with source data to copy if not NULL
 *
 * Return(s)
 *	pointer to mblk if successful
 *	NULL	otherwise
 */
M_BLK_ID ks884xAllocMblk
(   
    DRV_CTRL * pDrvCtrl,    /* pointer to DRV_CTRL structure */
    UCHAR    **ppCluster, 
    M_BLK_ID   pSrcMblk
)
{
	UCHAR     *pNewCluster, *pClusterPtr = NULL;
    CL_BLK_ID  pClBlk=NULL;
	M_BLK_ID   pMblk=NULL;

    /*
     * We implicitly are loaning here, if copying is necessary this
     * step may be skipped, but the data must be copied before being
     * passed up to the protocols.
     */
    pNewCluster = netClusterGet(pDrvCtrl->endObj.pNetPool, pDrvCtrl->pClPoolId);
    if (pNewCluster == NULL)
        goto ERREXIT;

    /* Grab a cluster block to marry to the cluster we received. */
    if ((pClBlk = netClBlkGet(pDrvCtrl->endObj.pNetPool, M_DONTWAIT)) == NULL)
        goto ERREXIT;

    if ((pMblk = netMblkGet(pDrvCtrl->endObj.pNetPool, M_DONTWAIT, MT_DATA)) == NULL)
        goto ERREXIT;

	if (NULL == pSrcMblk) 
    {
		pClusterPtr = pNewCluster;
	}
	else 
    {
		/* copy data from source mblk to pClusterPtr */
		DRV_LOG(DRV_DEBUG_RX, "not implemented yet!\n", 1, 2, 3, 4, 5, 6);
		pClusterPtr = pNewCluster;
	}

    /* Join the cluster to the MBlock, note that length should be at least DI->uRxBufferLen */
    netClBlkJoin(pClBlk, pClusterPtr, DI->uRxBufferLen, NULL, 0, 0, 0);
    netMblkClJoin(pMblk, pClBlk);

    /* make the packet data coherent */
    *ppCluster = pClusterPtr;
	return pMblk;

ERREXIT:
    if (pNewCluster != NULL)
        netClFree(pDrvCtrl->endObj.pNetPool, pNewCluster);

    if ( pClBlk != NULL)
        netClBlkFree(pDrvCtrl->endObj.pNetPool, pClBlk);

    if ( pMblk != NULL)
        netMblkFree(pDrvCtrl->endObj.pNetPool, pMblk);

    /*
	DRV_LOG(DRV_DEBUG_ERROR, "ks884xAllocMblk(): Out of mblk/clblk/cluster!\n", 
                              1, 2, 3, 4, 5, 6);
                              */
	END_ERR_ADD(&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
	return NULL;
}

/*
 * ks884xWatchdog
 *	This function is a timer callback routine to check link status on a given port.
 *  If link is down, perform a link auto-negotiation.
 *  Timer periodic is 5 seconds. 
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xWatchdog
(   
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
    ULONG        uPort;
    UINT8	     bLinkActive;
	PORT_KS_INFO *DPI;      /* ks884x port related (switch port only) */


    DRV_LOG(DRV_DEBUG_WATCHDOG, "ks884xWatchdog running ...\n", 1, 2, 3, 4, 5, 6);


    /* Polling KS884x PCI Status Register */
    ks884xPciStatusPolling ( 0 );

	/* Reset the watchdog */
	wdStart(pDrvCtrl->watchdog, pDrvCtrl->nWatchdogDelay, (FUNCPTR)ks884xWatchdog, (int)pDrvCtrl);
}

/*
 *-------------------------------------------------------------------------------------------
 *	The following functions can be called from Shell interface as GUI for 
 *  user to debug.
 *--------------------------------------------------------------------------------------------
 */

/**
 *  Convert device ID to device name. 
 *
 */ 
LOCAL void getDeviceName 
(
    USHORT   ChipID
)
{
    if (ChipID == CHIP_IP_41_PCI)
        fKS884x = 1;
    else if (ChipID == CHIP_IP_42_PCI)
        fKS884x = 2;
    else if (ChipID == CHIP_IP_61_PCI)
        fKS884x = 3;
    else if (ChipID == CHIP_IP_62_PCI)
        fKS884x = 4;
    else
        fKS884x = 0;  /* unknown device */
}

/*
 * showKsBspInfo
 *	  This function is use to show driver's BSP info
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
void showKsBspInfo 
(
)
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
    PHARDWARE  phw = pDrvCtrl->pHW;
    char  hostName[16];
    USHORT   ChipID;
    UCHAR    DevRevisionID;



    /* Display Device Driver info on UART */
    printf( NEWLINE );
	printf("Micrel - Ethernet Group."NEWLINE);	   
    printf("%s PCI BUS Ethernet Driver demo."NEWLINE, deviceName[fKS884x] ); 

    printf(pDescription, deviceName[fKS884x] );
#ifndef  TEST_THROUGHPUT
    printf(ks884xEnd_ver);	   
#else
    printf(ks884xEnd_throughput_ver);	   
#endif /* #ifndef  TEST_THROUGHPUT */

    printf( NEWLINE );
    printf("%s Driver Version: " KS884X_VERSION, deviceName[fKS884x]); /* KS884X Driver Version */
    printf( NEWLINE );

    HardwareReadChipID ( phw, &ChipID, &DevRevisionID );
    printf("%s Chip ID       : %04x", deviceName[fKS884x], ChipID);     /* KS884X Chip ID */
    printf( NEWLINE );
    printf("%s Device Revision: %d", deviceName[fKS884x], DevRevisionID); /* KS884X Device Revision */
    printf( NEWLINE );

    printf( "%s Base Address: 0x%08X"NEWLINE, deviceName[fKS884x], (int)phw->m_ulVIoAddr );

	printf("MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", 
		    pDrvCtrl->enetAddr[0], 
		    pDrvCtrl->enetAddr[1],
		    pDrvCtrl->enetAddr[2],
		    pDrvCtrl->enetAddr[3], 
		    pDrvCtrl->enetAddr[4],
		    pDrvCtrl->enetAddr[5]); 

    memset(hostName, 0, sizeof(hostName));
    sprintf(hostName, "%s%d", KS_DEV_NAME, pDrvCtrl->unit );
    printf("IP  Address: ");	   
    if ( ifAddrGet ( hostName, ksIpAddress ) != OK )
    {
        printf("unknown (fail to get IP address of device %s !)"NEWLINE, hostName );	   
    }
    else
        printf("%s"NEWLINE, ksIpAddress);	   
    printf( NEWLINE );

}

/*
 * showKsLinkStatus
 *	  This function is use to show KS884x Link Status
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
void showKsLinkStatus 
(
)
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
    PHARDWARE  phw = pDrvCtrl->pHW;
    PPORT_INFO pInfo;
    UCHAR      bPort;


    /* Get port's PHY link status */
    SwitchGetLinkStatus( phw );

    /* Display Ethernet Link status on UART */

#ifdef DEF_KS8842
    for ( bPort = 0; bPort < SWITCH_PORT_NUM; bPort++ )
#else
    bPort = 0;
#endif
    {
        pInfo = &phw->m_PortInfo[ bPort ];

        printf("Port %d Link Status:  ", (bPort+1) );	   

        if ( pInfo->ulHardwareState == MediaStateConnected )
        {
	        /* Speed */
            if (pInfo->ulSpeed == 1000000)
            {
	            printf("100 Mbps, ");	   
            }
            else if (pInfo->ulSpeed == 100000)
            {
	            printf("10  Mbps, ");	   
            }
            else 
            {
	            printf("??? Mbps, ");	   
            }

	        /* Duplex */
            if ( pInfo->bDuplex == 2)
            {
	            printf("Full DPX");	   
            }
	        else if ( pInfo->bDuplex == 1)
            {
	            printf("Half DPX");	   
            }	
            else 
            {
	            printf("??? DPX");	   
            }
        }
        else
        {
	        printf("Down ");	   
        }
        printf( NEWLINE );

    } /* for ( bPort = 0; bPort < SWITCH_PORT_NUM; bPort++ ) */

    printf( NEWLINE );
}


/*
 * ksShowTxBDs
 *	  This function is use to show driver's DMA Tx descriptor infomation
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
STATUS ksShowTxBDs( ) 
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
	int      i;
	volatile TXDESC	*pTxDesc;

	pTxDesc = DI->pTxDescBaseAddr;
	for (i = 0; i < DI->nTxDescTotal; i++, pTxDesc++) 
    {
        printf("Tx Desc %02d addr:0x%08X, TDES0:0x%08X, TDES1:0x%08X, TDES2:0x%08X, TDES3:0x%08X\n", 
                i, (int)pTxDesc, 
                PCI_SWAP(pTxDesc->TxOwnBit), 
                PCI_SWAP(pTxDesc->TxFrameControl), 
                (int)tbdBuffAddrGet(pDrvCtrl, pTxDesc), 
                (int)pTxDesc->TxDMANextPtr);
	}
    printf ("nTxDescNextAvail = %d\n", (int)DI->nTxDescNextAvail);
    printf ("nTxDescAvail     = %d\n", (int)DI->nTxDescAvail);
    printf ("fTxNoResource    = %s\n", (int)DI->fTxNoResource ? "TRUE" : "FALSE");
    return OK;
}

/*
 * ksShowRxBDs
 *	  This function is use to show driver's DMA Rx descriptor infomation
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
STATUS ksShowRxBDs( ) 
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
	int      i, oldLevel = 0;
	volatile RXDESC	*pRxDesc;

    /*oldLevel = intLock();*/	/* protect it */
	pRxDesc = DI->pRxDescBaseAddr;
	for (i = 0; i < DI->nRxDescTotal; i++, pRxDesc++) 
    {
 	   KS_CACHE_INVALIDATE(pRxDesc, sizeof(RXDESC));
       printf("Rx Desc %02d addr:0x%08X, RDES0:0x%08X, RDES1:0x%08X, RDES2:0x%08X, RDES3:0x%08X\n", 
                i, (int)pRxDesc, 
                PCI_SWAP(pRxDesc->RxFrameControl), 
                PCI_SWAP(pRxDesc->RxDMABufLen), 
                (int)rbdBuffAddrGet(pDrvCtrl, pRxDesc), 
                (int)pRxDesc->RxDMANextPtr);
	}
    printf ("nRxDescNextAvail = %d\n", (int)DI->nRxDescNextAvail);
	/*intUnlock(oldLevel);*/
    return OK;
}


STATUS ksSetRxBDs
( 
    int i
) 
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
	volatile RXDESC	*pRxDesc;

	pRxDesc = &DI->pRxDescBaseAddr[i];

    /* the clean descriptor to the device's DMA controller  */
    rxDescSet ( pDrvCtrl, pRxDesc );
    KS_CACHE_FLUSH(pRxDesc, sizeof(RXDESC));

    KS_CACHE_INVALIDATE(pRxDesc, sizeof(RXDESC));
    printf("Rx Desc %02d addr:0x%08X, RDES0:0x%08X, RDES1:0x%08X, RDES2:0x%08X, RDES3:0x%08X\n", 
                i, (int)pRxDesc, 
                PCI_SWAP(pRxDesc->RxFrameControl), 
                PCI_SWAP(pRxDesc->RxDMABufLen), 
                (int)rbdBuffAddrGet(pDrvCtrl, pRxDesc), 
                (int)pRxDesc->RxDMANextPtr);
    return OK;
}



/*======================================================================================*/
STATUS ksSetBDBase( ) 
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
    PHARDWARE  phw = pDrvCtrl->pHW;
    UINT32     txDescBase,rxDescBase;


    txDescBase = MEM_TO_PCI_PHYS (KS_CACHE_VIRT_TO_PHYS ((UINT32)DI->pTxDescBaseAddr));
    rxDescBase = MEM_TO_PCI_PHYS (KS_CACHE_VIRT_TO_PHYS ((UINT32)DI->pRxDescBaseAddr));
	HardwareSetDescriptorBase (phw, txDescBase, rxDescBase);

    printf ("DI->pTxDescBaseAddr=0x%08x,txDescBase=0x%08x \n",DI->pTxDescBaseAddr, txDescBase );
    printf ("DI->pRxDescBaseAddr=0x%08x,rxDescBase=0x%08x \n",DI->pRxDescBaseAddr, rxDescBase );

    return OK;
}


/*
 * showKsLinkStatus
 *	  This function is use to get ks884xSH HW structure handler.
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
PHARDWARE ksGetHW 
(
    ULONG deviceIndex 
)
{
    DRV_CTRL * pDrvCtrl; 

    if ( deviceIndex > 1)
        return (0);

    pDrvCtrl = gpDrvCtrl[deviceIndex]; 
    return ( pDrvCtrl->pHW );
}

#endif /* #ifdef KS_PCI_BUS */
