/* ks884xSHEnd.c - END style Micrel KS884X 32bit Generic Bus Ethernet driver */

/* Copyright 1984-2002 Wind River Systems, Inc., and Cetia Inc. */

#ifdef KS_ISA_BUS 

#include "copyright_wrs.h"

/*
modification history
--------------------
21Feb05,pcd  Create for Micrel KS8841 or KS8842 32bit Generic Bus Ethernet driver..
*/

/*
DESCRIPTION
This module implements the Micrel KS8841, or KS8842 Generic bus  
Ethernet 32-bit network interface driver.

The 32bit Generic Bus ethernet controller is inherently little-endian.
The software interface to the driver is divided into two parts.
The first and second part are dealt with in the driver. The second
part of the interface is comprised of the I/O control registers and their
programming. The third part of the interface is comprised of the descriptors
and the buffers. 

This driver is designed to be moderately generic, operating unmodified
across the range of architectures and targets supported by VxWorks.  To
achieve this, the driver must be given several target-specific parameters,
and some external support routines must be provided. These target-specific
values and the external support routines are described below.

This driver supports multiple units per CPU.  The driver can be
configured to support big-endian or little-endian architectures.  It
contains error recovery code to handle known device errata related to DMA
activity. 

Some big-endian processors may be connected to the KS8841, or KS8842 32bit 
generic bus through EPLD which performs byte swapping during data phases.  
On such platforms, the driver need not perform byte swapping during 
transmit\receive Ethernet packets to the device. But, the driver need to 
perform byte swapping during read\write data to device registers.

BOARD LAYOUT
This device is on-board.  No jumpering diagram is necessary. 

EXTERNAL INTERFACE
The drvier provides one standard external interface, ks884xSHEndLoad().  As
input, this routine takes a string of colon-separated parameters.  The
parameters should be specified in hexadecimal (optionally preceded by '0x'
or a minus sign '-').  The parameter string is parsed using strtok_r().

TARGET-SPECIFIC PARAMETERS
The format of the parameter string is:

<unit:devMemAddr:devIoAddr:pciMemBase:vecNum:intLvl:
memAdrs:memSize:memWidth:boardBusEndianMode:offset:flags>

\is
\i <unit>
The unit number of the device.  Unit numbers start at zero and increase for
each device controlled by the same driver.  The driver does not use this
value directly.  The unit number is passed through the MUX API where it is
used to differentiate between multiple instances of a particular driver.

\i <devMemAddr>
This parameter is the memory mapped I/O base address of the device registers
in the memory map of the CPU.  The driver will locate device registers as
offsets from this base address.
This is a required parameter since ks884xSH only support memory mapping.

\i <vecNum>
This parameter is the vector associated with the device interrupt.
This driver configures the PCnet device to generate hardware interrupts
for various events within the device; thus it contains
an interrupt handler routine.  The driver calls pciIntConnect() to connect
its interrupt handler to the interrupt vector generated as a result of the
PCnet interrupt.

\i <intLvl>
Some targets use additional interrupt controller devices to help organize
and service the various interrupt sources.  This driver avoids all
board-specific knowledge of such devices.  During the driver's
initialization, the external routine sysKs884xIntEnable() is called to
perform any board-specific operations required to allow the servicing of a
PCnet interrupt.  For a description of sysKs884xIntEnable(), see "External
Support Requirements" below.

\i <memAdrs>
This parameter gives the driver the memory address to carve out its
buffers and data structures. If this parameter is specified to be
NONE then the driver allocates cache coherent memory for buffers
and descriptors from the system memory pool.
The PCnet device is a DMA type of device and typically shares access to
some region of memory with the CPU.  This driver is designed for systems
that directly share memory between the CPU and the PCnet.  It
assumes that this shared memory is directly available to it
without any arbitration or timing concerns.

\i <memSize>
This parameter can be used to explicitly limit the amount of shared
memory (bytes) this driver will use.  The constant NONE can be used to
indicate no specific size limitation.  This parameter is used only if
a specific memory region is provided to the driver.

\i <memWidth>
Some target hardware that restricts the shared memory region to a
specific location also restricts the access width to this region by
the CPU.  On these targets, performing an access of an invalid width
will cause a bus error.

This parameter can be used to specify the number of bytes of access
width to be used by the driver during access to the shared memory.
The constant NONE can be used to indicate no restrictions.

Current internal support for this mechanism is not robust; implementation 
may not work on all targets requiring these restrictions.

\i <boardBusEndianMode>
This parameter specifies big-endian and little-endian modes of operation.  
When big-endian mode is selected, the PCnet-PCI controller will swap the
order of bytes on the AD bus during a data phase on access to the FIFOs
only:  AD[31:24] is byte 0, AD[23:16] is byte 1, AD[15:8] is byte 2 and
AD[7:0] is byte 3.  In order to select the big-endian mode, set this
parameter to (1).  Most implementations, including natively
big-endian host architectures, should set this parameter to (1) in
order to select little-endian access to the FIFOs, as the driver is
currently designed to perform byte swapping as appropriate to the host
architecture.

\i <offset>
This parameter specifies a memory alignment offset.  Normally this
parameter is zero except for architectures which can only access 32-bit
words on 4-byte aligned address boundaries.  For these architectures the
value of this offset should be 2.

\i <flags>
This is parameter is used for future use.  Currently its value should be
zero.
\ie

EXTERNAL SUPPORT REQUIREMENTS
This driver requires five externally defined support functions that can
be customized by modifying global pointers.  The function pointer types
and default "bindings" are specified below.  To change the defaults, the
BSP should create an appropriate routine and set the function pointer
before first use.  This would normally be done within sysHwInit2().

Note that all of the pointers to externally defined functions <must> be
set to a valid executable code address.  Also, note that
sysKs884xIntEnable(), sysKs884xIntDisable(), and sysKs884xEnetAddrGet()
must be defined in the BSP.  This was done so that the driver would be
compatible with initialization code and support routines in existing BSPs.

The function pointer convention has been introduced to facilitate future
driver versions that do not explicitly reference a named BSP-defined
function.  Among other things, this would allow a BSP designer to define,
for example, one endIntEnable() routine to support multple END drivers.

\is
\i 'ks884xSHIntConnect'
\cs
    IMPORT STATUS (* ks884xSHIntConnect)
        (
        VOIDFUNCPTR * vector,     /@ interrupt vector to attach to     @/
        VOIDFUNCPTR   routine,    /@ routine to be called              @/
        int           parameter   /@ parameter to be passed to routine @/
        );

    /@ default setting @/

    ks884xSHIntConnect = pciIntConnect;
\ce
The ks884xSHIntConnect pointer specifies a function used to connect the
driver interrupt handler to the appropriate vector.  By default it is
the <pciIntLib> routine pciIntConnect().

\i 'ks884xSHIntDisconnect'
\cs
    IMPORT STATUS (* ks884xSHIntDisconnect)
        (
        VOIDFUNCPTR * vector,     /@ interrupt vector to attach to     @/
        VOIDFUNCPTR   routine,    /@ routine to be called              @/
        int           parameter   /@ routine parameter                 @/
        );

    /@ default setting @/

    ks884xSHIntDisconnect = pciIntDisconnect2;
\ce
The ks884xSHIntDisconnect pointer specifies a function used to disconnect
the interrupt handler prior to unloading the driver.  By default it is
the <pciIntLib> routine pciIntDisconnect2().

\i 'ks884xSHIntEnable'
\cs
    IMPORT STATUS (* ks884xSHIntEnable)
        (
        int level                 /@ interrupt level to be enabled @/
        );

    /@ default setting @/

    ks884xSHIntEnable = sysKs884xIntEnable;
\ce
The ks884xSHIntEnable pointer specifies a function used to enable the
interrupt level for the END device.  It is called once during
initialization.  By default it is a BSP routine named sysKs884xIntEnable(). 
The implementation of this routine can vary between architectures, and
even between BSPs for a given architecture family.  Generally, the
parameter to this routine will specify an interrupt <level> defined for an
interrupt controller on the host platform.  For example, MIPS and PowerPC
BSPs may implement this routine by invoking the WRS intEnable() library
routine.  WRS Intel Pentium BSPs may implement this routine via
sysIntEnablePIC().

\i 'ks884xSHIntDisable'
\cs
    IMPORT STATUS (* ks884xSHIntDisable)
        (
        int level                 /@ interrupt level to be disabled @/
        );

    /@ default setting @/

    ks884xSHIntDisable = sysKs884xIntDisable;
\ce
The ks884xSHIntDisable pointer specifies a function used to disable the
interrupt level for the END device.  It is called during stop.  By default
it is a BSP routine named sysKs884xIntDisable().  The implementation of
this routine can vary between architectures, and even between BSPs for a
given architecture family.  Generally, the parameter to this routine will
specify an interrupt <level> defined for an interrupt controller on the
host platform.  For example, MIPS and PowerPC BSPs may implement this
routine by invoking the WRS intDisable() library routine.  WRS Intel
Pentium BSPs may implement this routine via sysIntDisablePIC().

\i 'ks884xSHEnetAddrGet'
\cs
    IMPORT STATUS (* ks884xSHEnetAddrGet)
        (DRV_CTRL * pDrvCtrl, char * pStationAddr);

\ce
The ks884xSHEnetAddrGet pointer specifies a function used to get the Ethernet
(IEEE station) address of the device.  
\ie

SYSTEM RESOURCE USAGE
When implemented, this driver requires the following system resources:

    - one mutual exclusion semaphore
    - one interrupt vector
    - 14240 bytes in text for a PENTIUM3 target
    - 120 bytes in the initialized data section (data)
    - 0 bytes in the uninitialized data section (BSS)

The driver allocates clusters of size 1520 bytes for receive frames and
and transmit frames.

SEE ALSO: muxLib, endLib, netBufLib

\tb "Network Protocol Toolkit User's Guide"

definition TEST_THROUGHPUT        
To test throughput from ks884x chip. In the packets receive routine ks884xRecv(), 
drop the packets right away (not pass to protocol stack) when received packets.

definition BRINGUP_BOARD
For debug purpose, just to bring up the demo board, without software initilize 
the device. All the ks884x device registers are the hardware default value.

definition NOTCHECK_CHIPID
For debug purpose, don't check the ks884x device ID.

definition KS_16BIT_BUS
If ks884x device is 16bit bus interface, we configure sh7751R Area 4 to 16bit 
bus width, otherwise, configure sh7751R Area 4 to 32bit bus width.

definition SH_32BIT_ACCESS_ONLY
Some CPU hardware platforms can only do 32-bit access only 
(can't access ks884x device by BYTE/WORD) because CPU bus controller can't 
gererate t1 setup time and t2 holding time for BxEN (ks884x datasheet Figure 16). 

*/

#include "vxWorks.h"
#include "taskLib.h"
#include "stdio.h"
#include "stdlib.h"
#include "sysLib.h"
#include "intLib.h"
#include "iv.h"
#include "netLib.h"
#include "inetLib.h"

#ifdef WR_IPV6
#include "adv_net.h"
#endif /*WR_IPV6*/

#define END_MACROS

#include "etherMultiLib.h"
#include "endLib.h"
#include "config.h"

#include "hardware.h"
#include "ks884xEnd_shBus.h"
#include "hw_ver.h"
#include "ks_shBus_ver.h"
#include "ks_shBus_ver_put.h"
#include "ks_qc_ver.h"


#if (0)
#define   BRINGUP_BOARD          /* Just to bring up the demo board, without init the device */
#endif
#if (0)
#define   NOTCHECK_CHIPID        /* ignore checking chip id */ 
#endif

#if (0)
#define   KS_16BIT_BUS           /* Set Area 4 to external 16bit bus for ks8842M-16 Non-PCI Interface */
#endif
#if (0)
#define   SDRAM_LIKE_INTERFACE   /* to test ks8842M  SDRAM Like Interface, disable "byte control mode" */
#endif

const char *deviceName[] =
{
/*  0 */  "UNKNOWN",
/*  1 */  "KSZ8841",
/*  2 */  "KSZ8842",
/*  3 */  "KSZ8861",
/*  4 */  "KSZ8862",
	      ""
};

void hw_read_byte  ( PHARDWARE, UCHAR, UINT8 * );
void hw_write_byte ( PHARDWARE, UCHAR, UINT8   );
void hw_read_word  ( PHARDWARE, UCHAR, UINT16 *);
void hw_write_word ( PHARDWARE, UCHAR, UINT16  );


/* defines */


/* Cache macros */

#define KS_CACHE_INVALIDATE(address, len) \
        CACHE_DRV_INVALIDATE (&pDrvCtrl->cacheFuncs, (address), (len))

#define KS_CACHE_FLUSH(address, len)				    \
        CACHE_DRV_FLUSH (&pDrvCtrl->cacheFuncs, (address), (len))

#define KS_CACHE_USER_FLUSH(address, len)				    \
        CACHE_USER_FLUSH((address), (len))

#define KS_CACHE_VIRT_TO_PHYS(address) \
        CACHE_DRV_VIRT_TO_PHYS (&pDrvCtrl->cacheFuncs, (address))

#define KS_CACHE_PHYS_TO_VIRT(address) \
        CACHE_DRV_PHYS_TO_VIRT (&pDrvCtrl->cacheFuncs, (address))

#define END_FLAGS_ISSET(pEnd, setBits) \
            ((pEnd)->flags & (setBits))

#define SYS_OUT_LONG(pDrvCtrl, addr, value)     \
    if (pDrvCtrl->flags & KS_MODE_MEM_IO_MAP)   \
        *((volatile UINT32 *)(addr)) = (value); \
    else                                        \
        sysOutLong ((int)(addr), (value));      \

#define SYS_IN_LONG(pDrvCtrl, addr, data)       \
    if (pDrvCtrl->flags & KS_MODE_MEM_IO_MAP)   \
        ((data) = *((volatile UINT32 *)(addr)));\
    else                                        \
        ((data) = sysInLong ((int) (addr)));    \


/* A shortcut for getting the hardware address from the MIB II stuff. */

/* RFC 1213 (old) */

#define END_HADDR(pEnd) \
    ((pEnd)->mib2Tbl.ifPhysAddress.phyAddress)

#define END_HADDR_LEN(pEnd) \
    ((pEnd)->mib2Tbl.ifPhysAddress.addrLength)

/* RFC 2233 (new - not implemented in this driver yet)
 *
 * #define END_HADDR(pEnd) \
 * ((pEnd)->pMib2Tbl->m2Data.mibIfTbl.ifPhysAddress.phyAddress)
 *
 * #define END_HADDR_LEN(pEnd) \
 * ((pEnd)->pMib2Tbl->m2Data.mibIfTbl.ifPhysAddress.addrLength)
 */

/*** END MACROS ***/	 
#define END_M2_INUCAST(pEnd)			(pEnd)->mib2Tbl.ifInUcastPkts++
#define END_M2_INNUCAST(pEnd)		    (pEnd)->mib2Tbl.ifInNUcastPkts++
#define END_M2_INERRORS(pEnd)		    (pEnd)->mib2Tbl.ifInErrors++
#define END_M2_INDISCARDS(pEnd)		    (pEnd)->mib2Tbl.ifInDiscards++
#define END_M2_INOCTETS(pEnd,bytes)	    (pEnd)->mib2Tbl.ifInOctets += bytes
#define END_M2_OUTUCAST(pEnd)		    (pEnd)->mib2Tbl.ifOutUcastPkts++
#define END_M2_OUTNUCAST(pEnd)		    (pEnd)->mib2Tbl.ifOutNUcastPkts++
#define END_M2_OUTDISCARDS(pEnd)		(pEnd)->mib2Tbl.ifOutDiscards++
#define END_M2_OUTERRORS(pEnd)		    (pEnd)->mib2Tbl.ifOutErrors++
#define END_M2_OUTOCTETS(pEnd,bytes)	(pEnd)->mib2Tbl.ifOutOctets += bytes

/* imports */



/* bind external support interface to default BSP implementation */
IMPORT int    endMultiLstCnt (END_OBJ *);

IMPORT STATUS sysKs884xIntEnable (int);
IMPORT STATUS sysKs884xIntDisable (int);

STATUS (* ks884xSHIntConnect)
       (VOIDFUNCPTR *, VOIDFUNCPTR, int) = intConnect;
STATUS (* ks884xSHIntDisconnect)
       (VOIDFUNCPTR *, VOIDFUNCPTR, int) = NULL;

STATUS (* ks884xSHIntEnable)(int)   = intEnable;
STATUS (* ks884xSHIntDisable)(int)  = intDisable;


#define DRV_DEBUG
/*
 */
#ifdef DRV_DEBUG        /* if debugging driver */
#    include "logLib.h"
#    define DRV_DEBUG_OFF           0x000000
#    define DRV_DEBUG_RX            0x000001
#    define DRV_DEBUG_TX            0x000002
#    define DRV_DEBUG_INT           0x000004
#    define DRV_DEBUG_POLL_RX       0x000008
#    define DRV_DEBUG_POLL_TX       0x000010
#    define DRV_DEBUG_POLL          (DRV_DEBUG_POLL_RX | DRV_DEBUG_POLL_TX)
#    define DRV_DEBUG_LOAD          0x000020
#    define DRV_DEBUG_UNLOAD        0x000040
#    define DRV_DEBUG_IOCTL         0x000080
#    define DRV_DEBUG_RESET         0x000100
#    define DRV_DEBUG_MCAST         0x000200
#    define DRV_DEBUG_CSR           0x000400
#    define DRV_DEBUG_RX_PKT        0x000800
#    define DRV_DEBUG_START         0x001000
#    define DRV_DEBUG_STOP          0x002000
#    define DRV_DEBUG_RXEARLY       0x004000
#    define DRV_DEBUG_TEMP          0x008000
#    define DRV_DEBUG_PHY           0x010000
#    define DRV_DEBUG_TX_MEM        0x020000
#    define DRV_DEBUG_CHIP_INIT     0x040000
#    define DRV_DEBUG_WATCHDOG      0x080000
#    define DRV_DEBUG_32BIT         0x100000
#    define DRV_DEBUG_ERROR         0x800000
#    define DRV_DEBUG_ALL           0xffffff

LOCAL int ksDebug = ( DRV_DEBUG_ERROR | DRV_DEBUG_32BIT /* | DRV_DEBUG_POLL | DRV_DEBUG_CHIP_INIT | DRV_DEBUG_INT */);

#define DRV_LOG(FLG, X0, X1, X2, X3, X4, X5, X6)       \
    if (ksDebug & FLG)                              \
        logMsg((char *) X0, (int)X1, (int)X2, (int)X3, \
                            (int)X4, (int)X5, (int)X6);

#define DRV_PRINT(FLG,X)                                                \
	if (ksDebug & FLG) printf X;


#else /* DRV_DEBUG */

#define DRV_LOG(DBG_SW, X0, X1, X2, X3, X4, X5, X6)
#define DRV_PRINT(FLG,X)                                                \

#endif /* DRV_DEBUG */



/* globals */

IMPORT int ks884xSH_max_if;  

DRV_CTRL 	*gpDrvCtrl[KS884X_MAX_IF];


/* locals */

LOCAL volatile int ks884xSHBusEndianess   = KS_LITTLE_ENDIAN;
LOCAL volatile int ks884xSHInterruptLevel = KS_ACTIVE_LOW;


/* default description for the END device */
LOCAL char * pDescription = "Micrel %s Generic Bus Enhanced Network Driver";
BYTE 	     fKS884x = 0;      /* 1: KS8841, 2: KS8842 */
char         ksIpAddress[INET_ADDR_LEN];


/* forward declarations */

LOCAL STATUS    ks884xSHReset (DRV_CTRL * pDrvCtrl);
LOCAL void      ks8842xEndSHInt (DRV_CTRL * pDrvCtrl);
LOCAL void      ks884xSHHandleRecvInt (DRV_CTRL * pDrvCtrl, UINT16 );
LOCAL BOOL      ks884xSHRecv (DRV_CTRL * pDrvCtrl, char	*, int *, BOOL * );
LOCAL void      ks884xSHHandleTxDoneInt (DRV_CTRL * pDrvCtrl, UINT16 );
LOCAL void      ks884xSHTxDoneProcess ( DRV_CTRL * pDrvCtrl );
LOCAL void      ks884xSHRestart (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xSHRestartSetup (DRV_CTRL * pDrvCtrl);
LOCAL void      ks884xSHAddrFilterSet (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xSHMemInit (DRV_CTRL * pDrvCtrl);

LOCAL void      ks884xSHSetDefaults ( DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ksSetApiFunctions ( DRV_CTRL * pDrvCtrl	);
      M_BLK_ID  ks884xSHAllocMblk ( DRV_CTRL *, UCHAR **, M_BLK_ID );
      BOOL      ks884xSHSend (DRV_CTRL * pDrvCtrl, char	*, int );

LOCAL STATUS    ks884xSHEnetAddrGet ( DRV_CTRL *, char * );
LOCAL STATUS    ks884xSHEnetAddrSet ( DRV_CTRL *, char * );
LOCAL STATUS    ks884xSHGetPhyStatus ( DRV_CTRL * pDrvCtrl, PORT_KS_INFO *, ULONG);
LOCAL void      ks884xSHAutoNegoStart ( DRV_CTRL * pDrvCtrl, PORT_KS_INFO *, ULONG );
LOCAL void      ks884xSHWatchdog ( DRV_CTRL * pDrvCtrl);

LOCAL void      ksReceiveEarlyFrame ( DRV_CTRL  * pDrvCtrl );
LOCAL void      ksReceiveFrame ( DRV_CTRL  * pDrvCtrl, UINT16 );
LOCAL BOOLEAN   QCCheckChipID ( PHARDWARE );
LOCAL void      getDeviceName ( USHORT );
      void      showKsBspInfo ( void );
      void      showKsLinkStatus ( void );
      void      setArea4WaitState ( void );


/* END-Specific interfaces */

LOCAL STATUS    ks884xSHEndStart (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xSHEndStop (DRV_CTRL * pDrvCtrl);
LOCAL STATUS    ks884xSHEndUnload (DRV_CTRL * pDrvCtrl);
LOCAL int       ks884xSHEndIoctl (DRV_CTRL * pDrvCtrl, int cmd, caddr_t data);
LOCAL STATUS    ks884xSHEndMCastAddrAdd (DRV_CTRL* pDrvCtrl, char * pAddress);
LOCAL STATUS    ks884xSHEndMCastAddrDel (DRV_CTRL * pDrvCtrl, char * pAddress);
LOCAL STATUS    ks884xSHMCastAddrGet (DRV_CTRL * pDrvCtrl, MULTI_TABLE * pTable);
      STATUS    ks884xSHEndSend (DRV_CTRL * pDrvCtrl, M_BLK_ID pBuf);
      STATUS    ks884xSHEndPollSend (DRV_CTRL * pDrvCtrl, M_BLK_ID pBuf);
      STATUS    ks884xSHEndPollReceive (DRV_CTRL * pDrvCtrl, M_BLK_ID pBuf);
      STATUS    ks884xSHEndPollStart (DRV_CTRL * pDrvCtrl);
      STATUS    ks884xSHEndPollStop (DRV_CTRL * pDrvCtrl);
      STATUS    ks884xSHInitParse ( DRV_CTRL *  pDrvCtrl, char * );

/*
 * Declare our function table.  This is static across all driver
 * instances.
 */
LOCAL NET_FUNCS ks884xSHFuncTable =
    {
    (FUNCPTR) ks884xSHEndStart,          /* Function to start the device. */
    (FUNCPTR) ks884xSHEndStop,           /* Function to stop the device. */
    (FUNCPTR) ks884xSHEndUnload,         /* Unloading function for the driver. */
    (FUNCPTR) ks884xSHEndIoctl,          /* Ioctl function for the driver. */
    (FUNCPTR) ks884xSHEndSend,           /* Send function for the driver. */
    (FUNCPTR) ks884xSHEndMCastAddrAdd,   /* Multicast address add  */
    (FUNCPTR) ks884xSHEndMCastAddrDel,   /* Multicast address delete */
    (FUNCPTR) ks884xSHMCastAddrGet,      /* Multicast table retrieve */
    (FUNCPTR) ks884xSHEndPollSend,       /* Polling send function  */
    (FUNCPTR) ks884xSHEndPollReceive,    /* Polling receive function */
    endEtherAddressForm,               /* Put address info into a packet.  */
    endEtherPacketDataGet,             /* Get a pointer to packet data. */
    endEtherPacketAddrGet              /* Get packet addresses. */
    };


/* inline functions */



/******************************************************************************
*
* ks884xSHEndLoad - initialize the driver and device
*
* This routine initializes the driver and the device to the operational state.
* All of the device-specific parameters are passed in <initString>, which
* expects a string of the following format:
*
* <unit:devMemAddr:vecNum:intLvl:memAdrs
* :memSize:memWidth:boardBusEndianMode:offset:flags>
*
* This routine can be called in two modes. If it is called with an empty but
* allocated string, it places the name of this device (that is, "ks") into 
* the <initString> and returns 0.
*
* If the string is allocated and not empty, the routine attempts to load
* the driver using the values specified in the string.
*
* CAVEATS
* For ks8842 device (two port switch), the two END interfaces are represented for 
* the two switch port. But, ks8842 only has one set of DMA (same as ks8841).
* So both END interface's "pDrvCtrl->pDMAInfo" point to same DMA structure.
*
* RETURNS: An END object pointer, or NULL on error, or 0 and the name of the
* device if the <initString> was NULL.
*/
END_OBJ * ks884xSHEndLoad
(
    char * initString            /* string to be parse by the driver */
)
{
    DRV_CTRL * pDrvCtrl;
    PHARDWARE  phw;
    struct hw_fn *pks_fn=NULL;
    char       ksEndDescription[100];
    USHORT   ChipID;
    UCHAR    DevRevisionID;



    DRV_LOG(DRV_DEBUG_LOAD, "Loading KS884X  end...\n", 1, 2, 3, 4, 5, 6);

    if (initString == NULL)
    {
        return (NULL);
    }

    if (initString [0] == 0)
    {
        bcopy ((char *)KS_DEV_NAME, initString, KS_DEV_NAME_LEN);
        return ((END_OBJ *) OK);
    }

    /* Change Area 4 Wait State */
    setArea4WaitState(); 

    /* allocate the device structure */

    if ((pDrvCtrl = (DRV_CTRL *) calloc (sizeof (DRV_CTRL), 1))
        == NULL)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): allocate pDrvCtrl failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* parse the init string, filling in the device structure */

    if (ks884xSHInitParse (pDrvCtrl, initString) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): ks884xSHInitParse failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* Record pDrvCtrl as Global */
    gpDrvCtrl[0] = pDrvCtrl;

    /* 
     *  There is only one ks884xSH HW structure. 
     *  Always point to the same HW structure, no matter is is kx8841 or ks8842
     */
#if (0)
    if ( pDrvCtrl->unit == 0)
#endif
    {
        /* allocate the ks884xSH HW structure */
        pDrvCtrl->pHW = (PHARDWARE)calloc (sizeof (HARDWARE), 1);
        if (pDrvCtrl->pHW == NULL)
        {
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): allocate pHW failed\n", 1, 2, 3, 4, 5, 6);
	        goto errorExit;
        }
        memset( pDrvCtrl->pHW , 0, sizeof( HARDWARE ));

        /* allocate the ks884xSH internal set API structure */
        pks_fn = (struct hw_fn *)calloc (sizeof (struct hw_fn), 1);
        if (pks_fn == NULL)
        {
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): allocate pks_fn failed\n", 1, 2, 3, 4, 5, 6);
	        goto errorExit;
        }
        memset( pks_fn , 0, sizeof( struct hw_fn ));
        pDrvCtrl->pHW->m_hwfn = pks_fn;

        /* allocate the ks884xSH DMA structure */
        pDrvCtrl->pDMAInfo = (PDMA_INFO)calloc (sizeof (DMA_INFO), 1);
        if (pDrvCtrl->pDMAInfo == NULL)
        {
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): allocate DMA failed\n", 1, 2, 3, 4, 5, 6);
	        goto errorExit;
        }
        memset( pDrvCtrl->pDMAInfo , 0, sizeof( DMA_INFO ));
    }
#if (0)
    else
    {
        /* it is second END interface, point to the same ks884xSH HW staructure */
        pDrvCtrl->pHW =  gpDrvCtrl[0]->pHW;
        pDrvCtrl->pDMAInfo = gpDrvCtrl[0]->pDMAInfo; 
    }
#endif

    /* 
     * Assign "memory mapped I/O base address of the ks884xSH device registers
     * in the memory map of the CPU" to ks884xSH HW structure.
     * And PCI Conroller endian mode.
     *
     * We only can access ks884xSH registers after here.
     */
    phw = pDrvCtrl->pHW;
    phw->m_ulVIoAddr  = pDrvCtrl->devAdrs; 
    phw->m_boardBusEndianMode = pDrvCtrl->boardBusEndianMode; 

#if (0)
    printf ("m_ulVIoAddr=0x%08x\n", (int)phw->m_ulVIoAddr);
    printf ("m_boardBusEndianMode=%d\n", (int)phw->m_boardBusEndianMode);
#endif
              
	/* set some others default values. */
	ks884xSHSetDefaults( pDrvCtrl );


    /* Have the BSP hand us our address. */

    /* Get ks884xSH station MAC address */
    (* ks884xSHEnetAddrGet) (pDrvCtrl, (char *) &(pDrvCtrl->enetAddr [0]));

	/* if device's MAC address is not assigned, use driver's default MAC address instead */
	if (0 == pDrvCtrl->enetAddr[0] && 0 == pDrvCtrl->enetAddr[1] &&
		0 == pDrvCtrl->enetAddr[2] && 0 == pDrvCtrl->enetAddr[3] &&
		0 == pDrvCtrl->enetAddr[4] && 0 == pDrvCtrl->enetAddr[5]) 
    {
	    DRV_PRINT(DRV_DEBUG_ERROR, 
              ("ks884xSHEndLoad(): Device MAC is not assigned, set driver MAC to device."NEWLINE));

        /* Set driver's default MAC to ks884xSH station MAC address */
        if ( ks884xSHEnetAddrSet(pDrvCtrl, (char *) &DI->stMacStation[0]) == ERROR )
        {
	        DRV_PRINT(DRV_DEBUG_ERROR, 
                     ("ks884xSHEndLoad(): fail to set MAC address to ks884xSH.\n"));
            goto errorExit; 
        }
	}
    else
    {
		memcpy(&DI->stMacStation, pDrvCtrl->enetAddr, ETH_LENGTH_OF_ADDRESS);  
    }

    /* initialize the END and MIB2 parts of the structure */

    HardwareReadChipID ( phw, &ChipID, &DevRevisionID );
    getDeviceName ( ChipID );

    memset(ksEndDescription, 0, sizeof(ksEndDescription));
    sprintf( ksEndDescription, pDescription, deviceName[fKS884x] );

    if (END_OBJ_INIT (&pDrvCtrl->endObj, (DEV_OBJ *)pDrvCtrl, KS_DEV_NAME,
                      pDrvCtrl->unit, &ks884xSHFuncTable, ksEndDescription) == ERROR
     || END_MIB_INIT (&pDrvCtrl->endObj, M2_ifType_ethernet_csmacd,
                      &pDrvCtrl->enetAddr[0], 6, ETHERMTU, LN_SPEED) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): END_OBJ_INIT failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* Perform memory allocation */

    if (ks884xSHMemInit (pDrvCtrl) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): ks884xSHMemInit failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }

    /* set the flags to indicate readiness */

    END_OBJ_READY (&pDrvCtrl->endObj, IFF_NOTRAILERS | IFF_BROADCAST |
                                      IFF_MULTICAST  | IFF_SIMPLEX);

    /* reset and reconfigure the device */

    if (ks884xSHRestartSetup (pDrvCtrl) == ERROR)
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndLoad(): ks884xSHRestartSetup failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
    }


    /* Create a watchdog timer */
	pDrvCtrl->watchdog = wdCreate();
	if (NULL == pDrvCtrl->watchdog) 
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "wdCreate failed\n", 1, 2, 3, 4, 5, 6);
        goto errorExit;
	}

	if (OK != wdStart(pDrvCtrl->watchdog, pDrvCtrl->nWatchdogDelay, (FUNCPTR)ks884xSHWatchdog, (int)pDrvCtrl)) 
    {
	    DRV_LOG(DRV_DEBUG_ERROR, "wdStart failed\n", 1, 2, 3, 4, 5, 6);
		wdDelete(pDrvCtrl->watchdog);
		pDrvCtrl->watchdog = NULL;
	}

 	DRV_LOG(DRV_DEBUG_LOAD, "ks884xSHEndLoad(unit=%d) done.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    return (&pDrvCtrl->endObj);

errorExit:

    if (pks_fn != NULL)
        free ((char *)pks_fn);

    if (pDrvCtrl->pHW != NULL)
        free ((char *)pDrvCtrl->pHW);

    if (pDrvCtrl->pDMAInfo != NULL)
        free ((char *)pDrvCtrl->pDMAInfo);

    if (pDrvCtrl != NULL)
        free ((char *)pDrvCtrl);

    return ((END_OBJ *)NULL);
}

/*******************************************************************************
*
* ks884xSHInitParse - parse the initialization string
*
* Parse the input string. This routine is called from ks884xSHEndLoad() which
* intializes some values in the driver control structure with the values
* passed in the intialization string.
*
* The initialization string format is:
*
* <unit:devMemAddr:devIoAddr:pciMemBase:vecNum:intLvl:memAdrs
* :memSize:memWidth:boardBusEndianMode:offset:flags>
*
* \is
* \i <unit>
* The device unit number.  Unit numbers are integers starting at zero and
* increasing for each device contolled by the driver.
* \i <devMemAddr>
* The device memory mapped I/O register base address.  Device registers
* must be mapped into the host processor address space in order for the
* driver to be functional.  Thus, this is a required parameter.
* \i <vecNum>
* Interrupt vector number.
* \i <intLvl>
* Interrupt level.  Generally, this value specifies an interrupt level
* defined for an external interrupt controller.
* \i <memAdrs>
* Memory pool address or NONE.
* \i <memSize>
* Memory pool size or zero.
* \i <memWidth>
* Memory system size, 1, 2, or 4 bytes (optional).
* \i <boardBusEndianMode>
* Board Bus Endian mode options (1- board bus is big endian mode).
* \i <offset>
* Memory alignment offset.
* \i <flags>
* Device specific flags reserved for future use.
* \ie
*
* RETURNS: OK, or ERROR if any arguments are invalid.
*/
STATUS ks884xSHInitParse
(
    DRV_CTRL *  pDrvCtrl,       /* pointer to the control structure */
    char *      initString      /* initialization string */
)
{
    char *      tok;
    char *      pHolder = NULL;
    UINT32      devMemAddr;



#if (0)
    printf("ks884xSHInitParse(): initString:%s\n",initString);
#endif

    /* unit number */

    if ((tok = strtok_r (initString, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->unit = atoi (tok);
    DRV_LOG (DRV_DEBUG_TEMP, "Unit : %d ...\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    /* memory mapped IO address base */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    devMemAddr = (UINT32) strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_TEMP, "devMemAddr : 0x%X ...\n", devMemAddr,
             2, 3, 4, 5, 6);
    /*
    printf("ks884xSHInitParse(): devMemAddr:0x%08x\n",devMemAddr);
    */

    pDrvCtrl->devAdrs = devMemAddr;
    pDrvCtrl->flags  |= KS_MODE_MEM_IO_MAP;


    /* interrupt vector */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->inum = atoi (tok); 
    DRV_LOG (DRV_DEBUG_TEMP, "inum : 0x%X ...\n", pDrvCtrl->inum,
             2, 3, 4, 5, 6);

    /* interrupt level */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->ilevel = atoi (tok); 
    DRV_LOG (DRV_DEBUG_TEMP, "ilevel : 0x%X ...\n", pDrvCtrl->ilevel,
             2, 3, 4, 5, 6);

    /* caller supplied memory address */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->memAdrs = (char *)strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_TEMP, "memAdrs : 0x%X ...\n", (int)pDrvCtrl->memAdrs,
             2, 3, 4, 5, 6);

    /* caller supplied memory size */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->memSize = strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_TEMP, "memSize : 0x%X ...\n", pDrvCtrl->memSize,
             2, 3, 4, 5, 6);

    /* caller supplied memory width */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->memWidth = atoi (tok);
    DRV_LOG (DRV_DEBUG_TEMP, "memWidth : 0x%X ...\n", pDrvCtrl->memWidth,
             2, 3, 4, 5, 6);

    /* boardBusEndianMode value */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->boardBusEndianMode = atoi (tok);
    DRV_LOG (DRV_DEBUG_TEMP, "boardBusEndianMode : 0x%X ...\n", pDrvCtrl->boardBusEndianMode,
             2, 3, 4, 5, 6);

    /* caller supplied alignment offset */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->offset = atoi (tok);
    DRV_LOG (DRV_DEBUG_TEMP, "Offset : 0x%X ...\n", pDrvCtrl->offset,
             2, 3, 4, 5, 6);

    /* caller supplied flags */

    if ((tok = strtok_r (NULL, ":", &pHolder)) == NULL)
        return ERROR;

    pDrvCtrl->flags |= strtoul (tok, NULL, 16);
    DRV_LOG (DRV_DEBUG_TEMP, "flags : 0x%X ...\n", pDrvCtrl->flags,
             2, 3, 4, 5, 6);


    return OK;
}

/*******************************************************************************
*
* ks884xSHMemInit - initialize memory for Lance chip
*
* Using data in the control structure, setup and initialize the memory
* areas needed.  If the memory address is not already specified, then allocate
* cache safe memory.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xSHMemInit
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    UINT32            sz;       /* temporary size holder */


    /***** Establish size of shared memory region we require *****/

  	sz = (DI->nTxBufTotal) + (DI->nRxBufTotal) + 
          _CACHE_ALIGN_SIZE * 2;


	DRV_LOG (DRV_DEBUG_TEMP, "ks884xSHMemInit(unit=%d): sz=%d\n",
			                  pDrvCtrl->unit, sz, 3, 4, 5, 6);

    /***** Establish a region of shared memory *****/

    /* OK. We now know how much shared memory we need.  If the caller
     * provides a specific memory region, we check to see if the provided
     * region is large enough for our needs.  If the caller did not
     * provide a specific region, then we attempt to allocate the memory
     * from the system, using the cache aware allocation system call.
     */

    switch ((int) pDrvCtrl->memAdrs)
    {
	    default :       /* caller provided memory */
	        if (pDrvCtrl->memSize < sz)     /* not enough space */
		    {
		        DRV_LOG (DRV_DEBUG_ERROR, "ks884xSHMemInit(unit=%d): not enough memory provided\n"
			        "ks884xSH: need %ul got %d\n",
			        pDrvCtrl->unit, sz, pDrvCtrl->memSize, 4, 5, 6);
		        return (ERROR);
            }

	        /* set the beginning of pool */

	        pDrvCtrl->pShMem = pDrvCtrl->memAdrs;

            DRV_LOG (DRV_DEBUG_TEMP, "(default)pShMem : 0x%08X\n", 
                                     pDrvCtrl->pShMem, 2, 3, 4, 5, 6);

	        /* assume pool is cache coherent, copy null structure */

	        pDrvCtrl->cacheFuncs = cacheNullFuncs;
	        break;

        case NONE :     /* get our own memory */

	        /* Because the structures that are shared between the device
	        * and the driver may share cache lines, the possibility exists
	        * that the driver could flush a cache line for a structure and
	        * wipe out an asynchronous change by the device to a neighboring
	        * structure. Therefore, this driver cannot operate with memory
	        * that is not write coherent.  We check for the availability of
	        * such memory here, and abort if the system did not give us what
	        * we need.
	        */

	        if (!CACHE_DMA_IS_WRITE_COHERENT ())
		    {
		        printf ( "ks884xSH: device requires cache coherent memory\n" );
		        return (ERROR);
		    }

	        pDrvCtrl->pShMem = (char *) cacheDmaMalloc (sz);

	        if (pDrvCtrl->pShMem == NULL)
		    {
		        printf ( "ks884xSH: system memory unavailable, size: 0x%08X\n", sz );
		        return (ERROR);
		    }
            DRV_LOG (DRV_DEBUG_TEMP, "(NONE)pShMem: 0x%08X, size: 0x%08X\n", 
                                     pDrvCtrl->pShMem, sz, 3, 4, 5, 6);

	        /* copy the DMA structure */

	        pDrvCtrl->cacheFuncs = cacheDmaFuncs;

            break;
    }


    /* Save some things */

    pDrvCtrl->memBase = (char *)((UINT32) pDrvCtrl->pShMem & 0xff000000);

    if ((int) pDrvCtrl->memAdrs == NONE)
        pDrvCtrl->flags |= KS_MEM_ALLOC_FLAG;

    /* first let's clear memory */

    bzero ((char *) pDrvCtrl->pShMem, (int) sz);

    /* allocate pool structure for mblks, clBlk, and clusters */

    if ((pDrvCtrl->endObj.pNetPool = malloc (sizeof (NET_POOL))) == NULL)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xSHMemInit(unit=%d): malloc for pNetPool failed !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
	    return (ERROR);
    }

    
    pDrvCtrl->clDesc.clNum    = (DI->nRxBufTotal + 2) * 2;
    pDrvCtrl->mClCfg.clBlkNum = pDrvCtrl->clDesc.clNum;
    pDrvCtrl->mClCfg.mBlkNum  = pDrvCtrl->mClCfg.clBlkNum * 2;

    /* total memory size for mBlks and clBlks */
    
    pDrvCtrl->mClCfg.memSize =
        (pDrvCtrl->mClCfg.mBlkNum  * (MSIZE + sizeof (long))) +
        (pDrvCtrl->mClCfg.clBlkNum * (CL_BLK_SZ + sizeof (long)));

    /* total memory for mBlks and clBlks */

    if ((pDrvCtrl->mClCfg.memArea = (char *) memalign
        (sizeof (long), pDrvCtrl->mClCfg.memSize)) == NULL)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xSHMemInit(unit=%d): memalign failed !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return (ERROR);
    }

    DRV_LOG (DRV_DEBUG_TEMP, "mClCfg.memArea: 0x%08X, mClCfg.memSize: 0x%08X\n", 
                             pDrvCtrl->mClCfg.memArea, pDrvCtrl->mClCfg.memSize, 3, 4, 5, 6);

    /* total memory size for all clusters */

    pDrvCtrl->clDesc.clSize  = ROUND_UP(DI->uRxBufferLen, _CACHE_ALIGN_SIZE);
    pDrvCtrl->clDesc.memSize =
        (pDrvCtrl->clDesc.clNum * (pDrvCtrl->clDesc.clSize + 8)) + sizeof (int);

    /* Do we hand over our own memory? */

    if (pDrvCtrl->memAdrs != (char *) NONE)
    {
        pDrvCtrl->clDesc.memArea =
                  (char *)(pDrvCtrl->pShMem + (DI->nTxBufTotal /* * TXBD_SIZE */));
#if (0)
        printf ("pShMem: memArea=%08x\n", pDrvCtrl->clDesc.memArea );
#endif
    }
    else
    {
        pDrvCtrl->clDesc.memArea = cacheDmaMalloc (pDrvCtrl->clDesc.memSize);
#if (0)
        printf ("cacheDmaMalloc: memArea=%08x\n", pDrvCtrl->clDesc.memArea );
#endif
    }

    if (pDrvCtrl->clDesc.memArea == NULL)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xSHMemInit(unit=%d): system memory unavailable !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return (ERROR);
    }

    DRV_LOG (DRV_DEBUG_TEMP, "clDesc.memArea: 0x%08X, clDesc.memSize: 0x%08X\n", 
                             pDrvCtrl->clDesc.memArea, pDrvCtrl->clDesc.memSize, 3, 4, 5, 6);

    /* initialize the device net pool */
    
    if (netPoolInit (pDrvCtrl->endObj.pNetPool, &pDrvCtrl->mClCfg,
                     &pDrvCtrl->clDesc, 1, NULL) == ERROR)
    {
	    DRV_LOG (DRV_DEBUG_ERROR, "ks884xSHMemInit(unit=%d): netPoolInit fail !\n",
			                       pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return (ERROR);
    }
    
    /* Store the cluster pool ID as others need it later. */

    pDrvCtrl->pClPoolId = clPoolIdGet (pDrvCtrl->endObj.pNetPool,
                                       pDrvCtrl->clDesc.clSize, FALSE);

#if (0)
    printf ("ks884xSH: BDs address(pDrvCtrl->pShMem) = 0x%08x\n", (int)pDrvCtrl->pShMem );
    printf ("ks884xSH: BDs buf address(pDrvCtrl->clDesc.memArea) = 0x%08x\n", (int)pDrvCtrl->clDesc.memArea  );
#endif

    return OK;
 }

/*******************************************************************************
*
* ks884xSHEndStart - start the device
*
* This function calls BSP functions to connect interrupts and start the
* device running in interrupt mode.
*
* CAVEATS
* For ks8842 device (two port switch), the two END interfaces are represented for 
* the two switch port. But, ks8842 only has one set of DMA (same as ks8841)  
* need to be started.
*
* RETURNS: OK or ERROR if the interrupt handler could not be connected.
*/
LOCAL STATUS ks884xSHEndStart
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    static int initialized = 0;
    PHARDWARE  phw = pDrvCtrl->pHW;
    STATUS     result = OK;


    pDrvCtrl->txCleaning = FALSE;
    pDrvCtrl->txBlocked  = FALSE;
	pDrvCtrl->rxHandling = FALSE;



    /* only execute the following code once. */
    if(initialized == 0)
    {
 	    DRV_LOG(DRV_DEBUG_START, "ks884xSHEndStart: Start inum=%d, ilevel=%d.\n", 
                                  pDrvCtrl->inum, pDrvCtrl->ilevel, 3, 4, 5, 6);

#ifndef BRINGUP_BOARD 


        result = (* ks884xSHIntConnect)((VOIDFUNCPTR *)INUM_TO_IVEC(pDrvCtrl->inum),
                                 ks8842xEndSHInt, (int)(pDrvCtrl));

        if (result == OK)
        {
            /* Enables the device QMU */
            HardwareEnable( phw );

#ifdef DEF_KS8842

            /* Enable the device Switch */
            SwitchEnable ( phw, TRUE );

#endif /* #ifdef DEF_KS8842 */

            /* initialise the hardware interrupt mask for proper operation. */
            HardwareSetupInterrupt( phw );

#ifdef TEST_THROUGHPUT
            phw->m_wInterruptMask &= ~( INT_TX );
#endif

            /* Clear interrupts on the device-side */
            HardwareAcknowledgeInterrupt( phw, 0xFFFF );

            /* Enables the interrupts of the hardware. */
            HardwareEnableInterrupt( phw );

            /* Enable system level interrupt */
            (* ks884xSHIntEnable) (pDrvCtrl->ilevel);

            CACHE_PIPE_FLUSH (); /* Flush the write pipe */

        }
        else
 	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHEndStart(unit=%d): IntConnect fail.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

#endif /* #ifdef BRINGUP_BOARD */

        initialized ++;

    }   /* if(initialized == 0) */

    if (result == OK)
    {

        /* mark the interface as up */

        END_FLAGS_SET (&pDrvCtrl->endObj, (IFF_UP | IFF_RUNNING));

    }

    DRV_LOG(DRV_DEBUG_START, "ks884xSHEndStart: Stop.\n", 1, 2, 3, 4, 5, 6);

    return (result);
}


/*******************************************************************************
*
* ks8842xEndSHInt - handle controller interrupt
*
* This routine is called at interrupt level in response to an interrupt from
* the controller.
*
* RETURNS: N/A
*/
LOCAL void ks8842xEndSHInt
(
    DRV_CTRL * pDrvCtrl  /* pointer to DRV_CTRL structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
	UINT16     wIntStatus=0;
    int        portIndex = 0;



#ifndef BRINGUP_BOARD

    /* Save address within the TXQ or RXQ frame buffer. */
    /* HardwareSaveRegs( phw ); */


    /* Read the device interrupt status register */
    HardwareReadInterrupt( phw, &wIntStatus );

    DRV_LOG(DRV_DEBUG_INT, "ks8842xEndSHInt(unit=%d), wIntStatus=0x%04X\n", pDrvCtrl->unit, wIntStatus, 3, 4, 5, 6); 

    /* If false interrupt, return. */
    if (! (wIntStatus & phw->m_wInterruptMask))
    {
        return;
    }

    /* Only handle the interrupt bits we enable them */
    wIntStatus &= phw->m_wInterruptMask;

    /* Disable the interrupt bit */
    HardwareDisableInterruptBit( phw, wIntStatus );

    /* Acknowledge the interrupt. */
    HardwareAcknowledgeInterrupt( phw, wIntStatus );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();



    /*
     * Transmit the Packet Status
     */

    if ( (wIntStatus & INT_TX) )
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT ]++;

    /* Transmitted the packet completed */
    if ( (wIntStatus & INT_TX) && !(pDrvCtrl->txCleaning) )
    {
        if (netJobAdd ((FUNCPTR) ks884xSHHandleTxDoneInt,
           (int) pDrvCtrl, (int)INT_TX, 0, 0, 0) == OK)
        {
            pDrvCtrl->txCleaning = TRUE;
        }
    }


    /*  Transmit error */
    if ( ( wIntStatus & INT_TX_UNDERRUN ) ) 
    {
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT_UNDERRUN ]++;

	    DRV_PRINT(DRV_DEBUG_ERROR, 
              ("ks8842xEndSHInt(): Transmit UnderRun!"NEWLINE));

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_TX_UNDERRUN & phw->m_wInterruptMask) );
    }

	if( wIntStatus & INT_TX_STOPPED)
	{
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_XMIT_INT_STOP ]++;

	    DRV_PRINT(DRV_DEBUG_ERROR, 
              ("ks8842xEndSHInt(): Transmit Process Stopped!"NEWLINE));

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_TX_STOPPED & phw->m_wInterruptMask) );
	}



    /*
     * PHY Status
     */

    if ( ( wIntStatus & INT_PHY ) ) 
    {

        SwitchGetLinkStatus( phw );

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_PHY & phw->m_wInterruptMask) );
    }

    /*
     * Receive the Packet Status
     */

    /*  Receive error */
    if ( ( wIntStatus & INT_RX_OVERRUN ) ) 
    {
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_NO_BUFFER ]++;

	    DRV_PRINT(DRV_DEBUG_RX, 
              ("ks8842xEndSHInt(): Receive OverRun!"NEWLINE));

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX_OVERRUN & phw->m_wInterruptMask) );
    }

	if( wIntStatus & INT_RX_ERROR)
	{
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_INT_ERROR ]++;

	    DRV_PRINT(DRV_DEBUG_ERROR, 
              ("ks8842xEndSHInt(): Receive Error Frame!"NEWLINE));

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX_ERROR & phw->m_wInterruptMask) );
	}

	if( wIntStatus & INT_RX_STOPPED)
	{
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_INT_STOP ]++;

	    DRV_PRINT(DRV_DEBUG_ERROR, 
              ("ks8842xEndSHInt(): Receive Process Stopped!"NEWLINE));

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX_STOPPED & phw->m_wInterruptMask) );
	}

    /* Receive the completed packet. */

    if ( ( wIntStatus & INT_RX ) ) 
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_INT ]++;

#ifdef TEST_THROUGHPUT
    if ( ( wIntStatus & INT_RX )  && !(pDrvCtrl->rxHandling) ) 
    {
        /* put a task level input packet handler on the netTask job queue */

        if (netJobAdd ((FUNCPTR) ksReceiveFrame,
           (int) pDrvCtrl, (int)INT_RX,0,0,0) == OK)
        {
            pDrvCtrl->rxHandling = TRUE;
        }

        /* we already received completed packet, not need to process INT_RX_EARLY */
        if ( ( wIntStatus & INT_RX_EARLY ) ) 
        {
            /* Acknowledge the RX_EARLY interrupt. */
            HardwareAcknowledgeEarly( phw );
            wIntStatus &= ~(INT_RX_EARLY);
        }
	}

#else /* #ifdef TEST_THROUGHPUT */

    if ((wIntStatus & INT_RX) && !(pDrvCtrl->rxHandling))
    {
        /* put a task level input packet handler on the netTask job queue */

        if (netJobAdd ((FUNCPTR) ks884xSHHandleRecvInt,
           (int) pDrvCtrl, (int)INT_RX,0,0,0) == OK)
        {
            pDrvCtrl->rxHandling = TRUE;
        }
    }

#endif /* #ifdef TEST_THROUGHPUT */

    /* Receive the Early_Receive packet. */
    if ( ( wIntStatus & INT_RX_EARLY )  && !(pDrvCtrl->rxHandling)) 
    {
        if (netJobAdd ((FUNCPTR) ksReceiveEarlyFrame,
           (int) pDrvCtrl, 0,0,0,0) == OK)
        {
            pDrvCtrl->rxHandling = TRUE;
        }

	}


    if (pDrvCtrl->txBlocked)    /* cause a MUX restart */
    {
        if (netJobAdd ((FUNCPTR) muxTxRestart,
           (int) &pDrvCtrl->endObj, 0, 0, 0, 0) == OK)
        {
            pDrvCtrl->txBlocked = FALSE;
        }
    }

    /* Restore address within the TXQ or RXQ frame buffer. */
    /* HardwareRestoreRegs( phw ); */

#endif /* BRINGUP_BOARD */

}

/*******************************************************************************
*
* ksReceiveEarlyFrame - Receive a Early Receive the Ethernet frame from NIC
*
*	This routine is to test ks884x Early Receive function. Read the 
*   Ethernet frame from ks884x packet memory by its Early Receive Threshold,
*   then read the rest Ethernet frame from ks884x packet memory.
*
* RETURNS: 
*
* NOMANUAL
*/
LOCAL void ksReceiveEarlyFrame 
(
    DRV_CTRL  * pDrvCtrl   /* points to DRV_CTRL structure */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    int         packet_len = phw->m_wReceiveThreshold;
    int         port = 0;
    int         portIndex = 0;
    USHORT      rxCntl;
    USHORT      rxLength;
    UINT8       status=TRUE;
    int         bufOffset;
    M_BLK_ID 	pMblk = NULL;
    UCHAR     * pCluster = NULL;


	DRV_PRINT(DRV_DEBUG_RXEARLY, 
              ("ksReceiveEarlyFrame(): Start."NEWLINE ));

    pDrvCtrl->rxHandling = TRUE;

    if ( (HardwareReceiveMoreDataAvailable ( phw ) ) <= 0  )
    {
        pDrvCtrl->rxHandling = FALSE;

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX_EARLY & phw->m_wInterruptMask) );

        return;
    }

    /* Allocate OS data buffer */
   	pMblk = ks884xSHAllocMblk(pDrvCtrl, &pCluster, NULL);
    if ( pMblk == NULL )
    {
        /* receive packet drop because no buffers in the OS	*/
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_DROPPED ] +=1;

	    /* Free up that memory */
        HardwareReleaseReceive( phw );

        pDrvCtrl->rxHandling = FALSE;

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX_EARLY & phw->m_wInterruptMask) );

        return;
    }

    /* 
     * Read the Early Received data from hardware packet memory to receive buffer 
     * by Early Receive Threshold length
     */

    /* Move rx pointer to data address by read out 4-byte of received packet header from hardware packet memory */
    HardwareReceiveBuffer( phw, pMblk->mBlkHdr.mData, 4 );


    HardwareReceiveBuffer( phw, pMblk->mBlkHdr.mData, packet_len );

    /* dump received data to the console port for debug */
    if ( phw->fDebugDumpRx )
    {
         /* for debug use */
         DBG_PRINT( NEWLINE );
         DBG_PRINT("Received Early Receive Frame: "NEWLINE );
         if ( packet_len > 0 )
             PrintPacketData( (char *)pMblk->mBlkHdr.mData, packet_len, port, 2 );
    }

    /* 
     * Read the rest Ethernet frame from ks884x packet memory   
     */

    bufOffset = packet_len;

    /* Checking Received packet status */
    if ( HardwareReceiveStatus( phw, &rxCntl, &rxLength)  )
    {
        /*
         *  Receive Valid and Good Frame.
         */

        packet_len = rxLength;
        packet_len -= bufOffset;

        /* update the received port */
        port = phw->m_bPortRX;
        portIndex = phw->m_bPortRX;
        if ( portIndex > 0 )
            portIndex -- ;

        /* 
         *  Packet must be checked with no errors, AND 
         *  the packet data size can't be bigger than our receive buffer. 
         */
        if ( (packet_len > 0)  && (packet_len <= (RX_BUF_SIZE - bufOffset)) )
        {
            /*  received Packet with no errors  */

            /* Read the received data from hardware packet memory to receive buffer */
            HardwareReceiveBuffer( phw, &pMblk->mBlkHdr.mData[bufOffset], packet_len );

            /* Update receive statistics. */
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_RCV ]++;
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_RCV ] += bufOffset;
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_RCV ] += packet_len;

            /* dump received data to the console port for debug */
            if ( phw->fDebugDumpRx )
            {
                /* for debug use */
                DBG_PRINT( NEWLINE );
                DBG_PRINT("Rx Cntl Word: 0x%04x, Byte Count: 0x%04x"NEWLINE, 
                          (rxCntl),
                          (rxLength) );
                if ( packet_len > 0 )
                {
                    PrintPacketData( (char *)&pMblk->mBlkHdr.mData[bufOffset], packet_len, port, 2 );
                }
            }

        } 
        else
        {
            /*  received Packet length errors  */
            status = FALSE;
	        DRV_PRINT(DRV_DEBUG_ERROR, 
                     ("ksReceiveEarlyFrame(): Receive packet length errors! packet_len=%d, rxCntl=0x%04x, rxLength=0x%04x)"NEWLINE, 
                        packet_len, rxCntl, rxLength ));

        }
    }
    else
    {
        /*  received Completed Packet with errors  */

        phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_RCV ] -= packet_len;

        status = FALSE;

        DRV_PRINT(DRV_DEBUG_RX, 
                  ("ksReceiveEarlyFrame(): Receive packet with errors! ( rxCntl=0x%04x, rxLength=0x%04x)"NEWLINE, 
                    rxCntl, rxLength ));

    } /* if ( HardwareReceiveStatus( phw, &rxCntl, &rxLength)  ) */


    /* Free Mblk\ClBlk */
    netMblkClChainFree (pMblk);

    /* Free up that memory */
    HardwareReleaseReceive( phw );

    /* 
     * To receive the rest of frame if they are more continuous incoming packets 
     */
    if ( status == TRUE )
        ksReceiveFrame( pDrvCtrl, (UINT16)INT_RX );

    pDrvCtrl->rxHandling = FALSE;

    /* Enable the interrupt bit on the device-side */
    HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX_EARLY & phw->m_wInterruptMask) );


    return;			/* Indicate we got packet or not */

}  /* ksReceiveEarlyFrame */


/******************************************************************************
*
* ksReceiveFrame - 
*
* This routine is to test ks884x throughput perform. It is called from the 
* interrupt service routine to receive packet from device, and drop the packet 
* without pass to protocol stack. 
*/
LOCAL void ksReceiveFrame
(
    DRV_CTRL  * pDrvCtrl,  /* points to DRV_CTRL structure */
    UINT16      wIER       /* holds INT_RX_FRAME bit read from interrupt status */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    M_BLK_ID 	pMblk = NULL;
    UCHAR     * pCluster = NULL;
    int         packetLen = 0;
    USHORT      wMemLength;
    BOOL        fStatus=TRUE;
    int         portIndex = 0;
    BOOL        fBroadcast;


    DRV_LOG(DRV_DEBUG_RX, "ksReceiveFrame (%d), Start. wIER=0x%04X\n", pDrvCtrl->unit, wIER, 3, 4, 5, 6);

    pDrvCtrl->rxHandling = TRUE;

    /* Allocate OS data buffer */
    pMblk = ks884xSHAllocMblk(pDrvCtrl, &pCluster, NULL);
    if ( pMblk == NULL )
    {
        /* receive packet drop because no buffers in the OS	*/
        phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_DROPPED ] +=1;

	    /* Free up that memory */
        HardwareReleaseReceive( phw );

        /* Enable the interrupt bit on the device-side */
        HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX & phw->m_wInterruptMask) );

        pDrvCtrl->rxHandling = FALSE;

        return;
    }

    while  ( ( (wMemLength = HardwareReceiveMoreDataAvailable ( phw )) > 0 ) && ( fStatus ) )
    {
	    DRV_PRINT(DRV_DEBUG_RX, 
                  ("ksReceiveFrame(): wMemLength=%d)"NEWLINE, 
                    wMemLength ));

        /* precess receive packet */

        fStatus = ks884xSHRecv (pDrvCtrl, pMblk->mBlkHdr.mData, &packetLen, &fBroadcast );

	    /*
	     * Update HW registers to indicate that this frame has been processed.
	     */

	    /* Free up that memory */
        HardwareReleaseReceive( phw );

#ifdef TEST_THROUGHPUT
        /*
         *  We loopback received packets to transmit.
         */
        if ( ( phw->fLoopbackStart ) && (fStatus) )
        {
             /* Set "fPollSendFromCLI" to indicate that ks884xSHEndSend from CLI command */
             pDrvCtrl->fPollSendFromCLI = TRUE; 

             /* Set trasnmit destination port */
             phw->m_bPortTX = phw->m_bPortRX;

             /* Send the received packet to the device */
             pMblk->mBlkHdr.mFlags |= M_PKTHDR;
             pMblk->mBlkHdr.mNext = NULL;
	         pMblk->mBlkHdr.mLen = packetLen;

	         if (!fBroadcast ) 
                  pMblk->mBlkHdr.mData[5] = pMblk->mBlkHdr.mData[11] ;  /* DA = SA */
             pMblk->mBlkHdr.mData[11]= DI->stMacStation[5] ;            /* SA = MacStation */

             ks884xSHEndSend ( pDrvCtrl, pMblk );
        }

#endif /* #ifdef TEST_THROUGHPUT */


    } /* while  ( HardwareReceiveMoreDataAvailable ( phw ) > 0 ) */

    /* Free Mblk\ClBlk */
    netMblkClChainFree (pMblk);

    pDrvCtrl->rxHandling = FALSE;

    /* Enable the interrupt bit on the device-side */
    HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX & phw->m_wInterruptMask) );

}



/*******************************************************************************
*
* ks884xSHHandleRecvInt - task level interrupt service for input packets
*
* This routine is called at task level indirectly by the interrupt
* service routine to do any message received processing.
*/
LOCAL void ks884xSHHandleRecvInt
(
    DRV_CTRL  * pDrvCtrl,  /* points to DRV_CTRL structure */
    UINT16      wIER       /* holds INT_RX_FRAME bit read from interrupt status */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    M_BLK_ID 	pMblk = NULL;
    UCHAR     * pCluster = NULL;
    int         packetLen = 0;
    BOOL        fStatus=TRUE;
    USHORT      wMemLength;
    int         portIndex = 0;
    int         maxReadCount = 50;
    BOOL        fBroadcast;


    DRV_LOG(DRV_DEBUG_RX, "ks884xSHHandleRecvInt (%d), Start. wIER=0x%04X\n", pDrvCtrl->unit, wIER, 3, 4, 5, 6);

    while  ( ( (wMemLength = HardwareReceiveMoreDataAvailable ( phw )) > 0 ) && 
             ( (maxReadCount--) > 0  ) && 
             ( fStatus ) )
    {
	    DRV_PRINT(DRV_DEBUG_RX, 
                  ("ks884xSHHandleRecvInt(): wMemLength=%d)"NEWLINE, 
                    wMemLength ));

        /* Allocate OS data buffer */
   	    pMblk = ks884xSHAllocMblk(pDrvCtrl, &pCluster, NULL);

        if ( pMblk != NULL )
        {
            /* precess receive packet */

            DRV_LOG(DRV_DEBUG_RX, "ks884xSHHandleRecvInt(), mData=0x%08X\n", pMblk->mBlkHdr.mData, 2, 3, 4, 5, 6);

            fStatus = ks884xSHRecv (pDrvCtrl, pMblk->mBlkHdr.mData, &packetLen, &fBroadcast );

            if ( fStatus )
            {
	            /* we must align the IP header on a long word (4 byte)
	             * boundary. This is done by shifting the RX packet by 2 bytes.
	             * (The MAC header is 14 bytes).
	             */
	    
	            bcopy (pMblk->mBlkHdr.mData, (pMblk->mBlkHdr.mData + pDrvCtrl->offset), packetLen);
	
	            pMblk->mBlkHdr.mData += pDrvCtrl->offset;

	            /* Make the packet data coherent */
	            KS_CACHE_INVALIDATE (pMblk->mBlkHdr.mData, packetLen);
            
	            /* intialize the mBlk */
                pMblk->mBlkHdr.mLen = packetLen;
                pMblk->mBlkHdr.mFlags |= M_PKTHDR;
                pMblk->mBlkPktHdr.len = packetLen;

                /* Call the upper layer's receive routine. */
                END_RCV_RTN_CALL(&pDrvCtrl->endObj, pMblk);

                /* Add one to our unicast data. */
                END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_UCAST, +1);
            }
            else
            {
                /* Free Mblk\ClBlk */
                netMblkClChainFree (pMblk);
            }
        }
        else
        {
            /* receive packet drop because no buffers in the OS	*/
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_DROPPED ] +=1;
            fStatus = FALSE;

	        DRV_PRINT(DRV_DEBUG_RX, 
                  ("ks884xSHHandleRecvInt(): Get ks884xSHAllocMblk is NULL! "NEWLINE  ));

        } /* if ( pMblk != NULL ) */


	    /*
	     * Update HW registers to indicate that this frame has been processed.
	     */

	    /* Free up that memory */
        HardwareReleaseReceive( phw );

    } /* while  ( HardwareReceiveMoreDataAvailable ( phw ) > 0 ) */

    pDrvCtrl->rxHandling = FALSE;

    /* Enable the interrupt bit on the device-side */
    HardwareEnableInterruptBit ( phw, (USHORT)(INT_RX & phw->m_wInterruptMask) );

}



/*
 * ksBufferCopyAllClusters
 *	This function is use to get a new mblk/clblk/cluster and copy source buffer data to 
 *  this new Mblk buffer in order to make buffer address is 32 bit alignment.

 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *	pSrcMblk	pointer to M_BLK_ID with source mblk to copy data from.
 *	len			lenght of data packet received
 *
 * Return(s)
 *	NONE.
 */
LOCAL M_BLK_ID ksBufferCopyAllClusters
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
    M_BLK_ID   pSrcMblk
)
{
	M_BLK_ID   pMblk;
	UCHAR	  *pCluster = NULL;
    int        len;


	if (NULL == pSrcMblk) 
		return NULL;

	pMblk = ks884xSHAllocMblk(pDrvCtrl, &pCluster, NULL);
	if (NULL == pMblk) 
		return NULL;

	/* shift data to make IP header dword aligned if required */
    len = netMblkToBufCopy(pSrcMblk, pMblk->mBlkHdr.mData, NULL);
	pMblk->mBlkHdr.mLen = len;
	pMblk->mBlkHdr.mFlags |= M_PKTHDR;
	pMblk->mBlkPktHdr.len = len;
    pMblk->mBlkHdr.mNext = NULL;

    /* make the packet data coherent */
    KS_CACHE_INVALIDATE(pMblk->mBlkHdr.mData, len);

    /* Free source pMblk */
    netMblkClChainFree (pSrcMblk);

	return pMblk;
}


/*******************************************************************************
*
* ks884xSHRecv - process the incoming packet
*
* RETURNS: OK or ERROR.
*/
LOCAL BOOL ks884xSHRecv
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
	char	 * packetData,
    int      * packetLen,
    BOOL     * fBroadcast
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    int        packet_len = 0;
    int        port;
    int        portIndex = 0;
    USHORT     rxCntl;
    USHORT     rxLength;
    BOOL       fStatus=TRUE;



    DRV_LOG (DRV_DEBUG_RX, "ks884xSHRecv()\n", 1, 2, 3, 4, 5, 6);

#ifndef BRINGUP_BOARD 

    if ( packetData == NULL )
    {
         DRV_LOG (DRV_DEBUG_ERROR, "ks884xSHRecv(): ERROR! packetData is NULL. \n", 1, 2, 3, 4, 5, 6);
         return (FALSE);
    }


    *fBroadcast = TRUE;

    /* Checking Received packet status */
    if ( HardwareReceiveStatus( phw, &rxCntl, &rxLength)  )
    {

        if ( (rxCntl & RX_UNICAST) == RX_UNICAST )
            *fBroadcast = FALSE;

        /* update the received port */
        port = phw->m_bPortRX;
        portIndex = phw->m_bPortRX;
        if ( portIndex > 0 )
             portIndex -- ;

        /* Get received packet length from hardware packet memory */
        packet_len = HardwareReceiveLength( phw ); 

        /* Packet must be checked with no errors. */
        /* Check if the packet data will fit into our receive buffer. */
        if ( (packet_len > 0) && (packet_len <= DI->uRxBufferLen) )
        {

             /* Copy the received data (not include CRC) from hardware packet memory to system buffer */

             HardwareReceiveBuffer( phw, packetData, packet_len );
             *packetLen = packet_len;

             /* make the packet data coherent */
             KS_CACHE_INVALIDATE (packetData, packet_len);

             /* update the received port */
             port = phw->m_bPortRX;

             /* Update receive statistics. */
             phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_RCV ] +=1;
             phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_RCV ] += packet_len;

             if ( phw->fDebugDumpRx )
             {
                  /* for debug use */
                  DBG_PRINT( NEWLINE );
                  DBG_PRINT("Rx Cntl Word: 0x%04x, Byte Count: 0x%04x"NEWLINE, 
                                   (rxCntl), (rxLength) );
                  if ( packet_len > 0 )
                      PrintPacketData( (UCHAR *)packetData, packet_len, port, 2 );
             }
        } 
        else
        {
                fStatus = FALSE;

                /* Encount an error, set the Length to zero */
                DRV_PRINT(DRV_DEBUG_ERROR, 
                  ("ks884xSHRecv(): Receive packet with errors! ( packet_len=%d, rxCntl=0x%04x, rxLength=0x%04x)"NEWLINE, 
                    packet_len, rxCntl, rxLength ));
        } /* if ( (packet_len > 0) && (packet_len <= MAXIMUM_ETHERNET_PACKET_SIZE) ) */
    }
    else
    {
        fStatus = FALSE;

        /*  received Packet with errors  */
        DRV_PRINT(DRV_DEBUG_ERROR, 
                  ("ks884xSHRecv(): Receive packet with errors! ( rxCntl=0x%04x, rxLength=0x%04x)"NEWLINE, 
                    rxCntl, rxLength ));

    } /* if ( HardwareReceiveStatus( phw, &rxCntl, &rxLength)  ) */

#endif /* #ifdef BRINGUP_BOARD */

    return (fStatus);
}

/*******************************************************************************
*
* ks884xSHSend - Transmit packet to Ethernet device.
*
* RETURNS: OK or ERROR.
*/
BOOL ks884xSHSend
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
	char	 * pData,
    int        dataLen
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    UINT32     uPort;
    ULONG      txCntlAndLength=0;
    USHORT     txCntl;
    USHORT     txLength;



    DRV_LOG (DRV_DEBUG_TX, "ks884xSHSend \n", 1, 2, 3, 4, 5, 6);


    if ( !pDrvCtrl->fPollSendFromCLI )
    {
        /* Trasnmit by Lookup mode */
        phw->m_bPortTX = 0;
    }

    if ( pDrvCtrl->fPollSendFromCLI )
        uPort = phw->m_bPortTX;
    else
        uPort = pDrvCtrl->unit;


    KS_CACHE_USER_FLUSH(pData, dataLen);


#ifndef BRINGUP_BOARD 

    /* Checking ks8842 TXQ memory available for this packet */
    if ( ! HardwareAllocPacket( phw, dataLen   ) ) 
    {
        DRV_LOG(DRV_DEBUG_TX_MEM, "ks884xSHSend(unit=%d) device TXQ memory not available!\n", 
                 pDrvCtrl->unit, 2, 3, 4, 5, 6);

        return (FALSE);
    }

    txCntlAndLength = HardwareSetTransmitLength( phw, dataLen );

	/* Now write the data out to the Ethernet chip */
    /* Continue write packet data (in DWORD) into ks8842 packet memory */
    HW_WRITE_BUFFER( phw, REG_DATA_OFFSET, pData, dataLen );

    if ( phw->fDebugDumpTx )
    {
         /* for debug use */
         DBG_PRINT( NEWLINE );
         txCntl   = (txCntlAndLength & 0x0000ffff);
         txLength = (txCntlAndLength >> 16) & 0x0000ffff;
         DBG_PRINT("Tx Cntl Word: 0x%04x, Byte Count: 0x%04x"NEWLINE, 
                    txCntl, txLength );
         PrintPacketData( (char *)pData, dataLen, uPort, 1 );
    }

#endif /* #ifdef BRINGUP_BOARD */

    return (TRUE);
}

/*******************************************************************************
*
* ks884xSHEndSend - the driver send routine
*
* This routine takes a M_BLK_ID sends off the data in the M_BLK_ID.
* The buffer must already have the addressing information properly installed
* in it.  This is done by a higher layer.
*
*
* RETURNS: OK or ERROR.
*/
STATUS ks884xSHEndSend
(
    DRV_CTRL * pDrvCtrl,  /* device to be initialized */
    M_BLK_ID   pMblk      /* data to send */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    M_BLK*     pNewMblk;
    M_BLK*     pFreeMblk;
	int		   dataLen = 0;
	char	  *pData;
    BOOL       fStatus=TRUE;
    ULONG      portIndex=0;

#ifndef BRINGUP_BOARD 


	DRV_LOG(DRV_DEBUG_TX, "ks884xSHEndSend (unit=%d)\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);


#ifdef  KS_QC_TEST
    if ( !pDrvCtrl->fPollSendFromCLI )
         return (OK);
#endif


    if ( pMblk == NULL )
         return (ERROR);

    /* check device mode */

    if (pDrvCtrl->flags & KS_POLLING)
    {
        netMblkClChainFree (pMblk);
        errno = EINVAL;

        return (ERROR);
    }

    /*
     * Obtain exclusive access to transmitter.  This is necessary because
     * we might have more than one stack transmitting at once.
     */

    END_TX_SEM_TAKE (&pDrvCtrl->endObj, WAIT_FOREVER);

    if ( (UINT32)(pMblk->mBlkHdr.mData) & 0x3 ) 
    {

	    DRV_LOG(DRV_DEBUG_TX, "ks884xSHEndSend () not DWORD Alignment, mData=%08x.\n", 
                pMblk->mBlkHdr.mData, 2, 3, 4, 5, 6);

        /* 
         * Copy all the buffers to a new Mblk to to meet 'buffer address is DWORD alignment',
         * because we write data to device's QMU by DWORD.
         */

        pNewMblk = ksBufferCopyAllClusters ( pDrvCtrl, pMblk );
        if ( pNewMblk == NULL )
        {
             END_TX_SEM_GIVE (&pDrvCtrl->endObj);
             return (ERROR);
        }

        pMblk = pNewMblk;
    }

    pFreeMblk = pMblk;
    while ( pMblk )
    {
        /* get data pointer and data length in this mBlk */ 
	    pData = pMblk->mBlkHdr.mData;
        dataLen = pMblk -> mBlkHdr.mLen;

        /* make the packet data coherent */
        KS_CACHE_INVALIDATE(pData, dataLen);


        /* 
         * Send packet data (from pMblk) to the device.
         */

        fStatus = ks884xSHSend (pDrvCtrl, pData, dataLen );

        /* Checking ks8842 TXQ memory available for this packet */
        if ( ! fStatus ) 
        {
            DRV_LOG(DRV_DEBUG_TX, "ks884xSHSend() fail!\n", 1, 2, 3, 4, 5, 6);

#ifndef TEST_THROUGHPUT

            /* Free pMblk */
            netMblkClChainFree (pFreeMblk);

#endif /* ifndef TEST_THROUGHPUT */

            END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_ERRS, +1);
            END_TX_SEM_GIVE (&pDrvCtrl->endObj);

		    return END_ERR_BLOCK;
        }

        /* Reference next cluster of the Mblk chain */
        pMblk = pMblk -> mBlkHdr.mNext ;

    } /* while ( pMblk ) */

    portIndex = phw->m_bPortTX;

    /* Issue TXQ Command (Enqueue Tx frame from TX buffer into TXQ for transmit) */
    if ( HardwareSendPacket( phw ) ) 
    {

        /* Update driver transmit counters */
        if ( portIndex == 3 )
        {
            while ( portIndex & 3)
            {
               portIndex >>= 1;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;    
             }
        }
        else
        {
               if ( portIndex > 0 ) 
                   portIndex --;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;    
        }

        /* increment the unicast output counter */
        END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_UCAST, +1);
    }
    else
    {
        DRV_LOG(DRV_DEBUG_TX, "ks884xSHEndSend(unit=%d) device Transmit packet fail!\n", 
                 pDrvCtrl->unit, 2, 3, 4, 5, 6);
    }

#ifndef TEST_THROUGHPUT

    /* Free pMblk */
    netMblkClChainFree (pFreeMblk);
#endif

    /* Advance our management index(es) */
    END_TX_SEM_GIVE (&pDrvCtrl->endObj);

#endif /* #ifndef BRINGUP_BOARD */

    return (OK);
}

/*******************************************************************************
*
* ks884xSHEndIoctl - the driver I/O control routine
*
* Process an ioctl request.
*
* RETURNS OK or ERROR value
*/
LOCAL int ks884xSHEndIoctl
(
    DRV_CTRL * pDrvCtrl, /* device to be initialized */
    int        cmd,      /* ioctl command to execute */
    caddr_t    data      /* data to get or set */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    int    error = 0;

#ifndef BRINGUP_BOARD 

    switch ((UINT)cmd)
    {
        case EIOCSADDR:
            if (data == NULL)
                return (EINVAL);

            bcopy ((char *)data, (char *)END_HADDR(&pDrvCtrl->endObj),
                   END_HADDR_LEN (&pDrvCtrl->endObj));
            break;

        case EIOCGADDR:
            if (data == NULL)
                return (EINVAL);

            bcopy ((char *)END_HADDR (&pDrvCtrl->endObj), (char *)data,
                   END_HADDR_LEN (&pDrvCtrl->endObj));
            break;

        case EIOCSFLAGS:
            {
            long value    = (long) data;
            long oldFlags = END_FLAGS_GET (&pDrvCtrl->endObj);
            long newFlags;

            if (value < 0)
            {
                value = -value;
                --value;
                END_FLAGS_CLR (&pDrvCtrl->endObj, value);
            }
            else
            {
                END_FLAGS_SET (&pDrvCtrl->endObj, value);
            }

            /* Handle IFF_XXX flags set or cleared by command */
              
            newFlags = END_FLAGS_GET (&pDrvCtrl->endObj);

            /* IFF_ALLMULTI */

            if ((newFlags & IFF_ALLMULTI) && !(oldFlags & IFF_ALLMULTI))
            {
                /* Enable Multicast frames mode to MAC Rx DMA control */
                HardwareSetMulticast( phw, TRUE );
                pDrvCtrl->flags |= KS_MULTICAST_FLAG; 
            }
            else if ((oldFlags & IFF_ALLMULTI) && !(newFlags & IFF_ALLMULTI))
            {
                /* Disable Multicast frames mode to MAC Rx DMA control */
                HardwareSetMulticast( phw, FALSE );
                pDrvCtrl->flags &= ~KS_MULTICAST_FLAG; 
            }

            /* IFF_PROMISC */
              
            if ((newFlags & IFF_PROMISC) && !(oldFlags & IFF_PROMISC))
            {
                /* Enable Promiscuous mode to device's MAC Rx DMA control */
                HardwareSetPromiscuous( phw, TRUE );
                pDrvCtrl->flags |= KS_PROMISCUOUS_FLAG; 
            }
            else if ((oldFlags & IFF_PROMISC) && !(newFlags & IFF_PROMISC))
            {
                /* Disable Promiscuous mode to device's MAC Rx DMA control */
                HardwareSetPromiscuous( phw, FALSE );
                pDrvCtrl->flags &= ~KS_PROMISCUOUS_FLAG; 
            }

            }
            break;

        case EIOCGFLAGS:
            *(int *)data = END_FLAGS_GET (&pDrvCtrl->endObj);
            break;

        case EIOCMULTIADD:
            error = ks884xSHEndMCastAddrAdd (pDrvCtrl, (char *) data);
            break;

        case EIOCMULTIDEL:
            error = ks884xSHEndMCastAddrDel (pDrvCtrl, (char *) data);
            break;

        case EIOCMULTIGET:
            error = ks884xSHMCastAddrGet (pDrvCtrl, (MULTI_TABLE *) data);
            break;

        case EIOCPOLLSTART:
            error = ks884xSHEndPollStart (pDrvCtrl);
            break;

        case EIOCPOLLSTOP:
            error = ks884xSHEndPollStop (pDrvCtrl);
            break;

        case EIOCGMIB2:
            if (data == NULL)
                return (EINVAL);
            bcopy((char *)&pDrvCtrl->endObj.mib2Tbl, (char *)data,
                  sizeof(pDrvCtrl->endObj.mib2Tbl));
            break;

        case EIOCGFBUF:
            if (data == NULL)
                return (EINVAL);
            *(int *)data = DI->uRxBufferLen;
            break;

        case EIOCGMWIDTH:
            if (data == NULL)
                return (EINVAL);
            *(int *)data = pDrvCtrl->memWidth;
            break;

        case EIOCGHDRLEN:
            if (data == NULL)
                return (EINVAL);
            *(int *)data = SIZEOF_ETHERHEADER;
            break;

        default:
            error = EINVAL;
    }

#endif /* #ifndef BRINGUP_BOARD  */

    return (error);
}

/**
 *	Set internal KS884x API function for ISA bus.
 *
 */ 
/*******************************************************************************
*
* ksSetApiFunctions
*
*	Set internal KS884x API function for ISA bus.
*
* RETURNS: NONE.
*/
LOCAL STATUS ksSetApiFunctions 
(    
    DRV_CTRL * pDrvCtrl	/* device ID */
)
{
    struct hw_fn *pks_fn = pDrvCtrl->pHW->m_hwfn;


    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ksSetApiFunctions(): Start... \n"));

    if ( pks_fn == NULL )
    {
        DRV_PRINT(DRV_DEBUG_ERROR, 
                 ("ksSetApiFunctions(): pks_fn is NULL. \n"));
        return ERROR;
    }

    HardwareSetupFunc( pks_fn );


    return OK;
}

/*******************************************************************************
*
* ks884xSHReset - hardware reset of chip (stop it)
*
* This routine is responsible for resetting the device
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xSHReset
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
#ifndef BRINGUP_BOARD 
    PHARDWARE     phw    = pDrvCtrl->pHW;

    /* Generate a soft-reset of the controller */

    if ( !HardwareReset( phw ) ) 
        return (ERROR);

#endif /* #ifdef BRINGUP_BOARD */

    return (OK);
}


/*******************************************************************************
*
* ks884xSHRestartSetup - 
*
* Setup memory descriptors and initializes hardware chip.
*
* CAVEATS
* For ks8842 device (two port switch), the two END interfaces are represented for 
* the two switch port. But, ks8842 only has one set of DMA (same as ks8841)  
* need to be initialized.
*
* RETURNS OK or ERROR.
*/
LOCAL STATUS ks884xSHRestartSetup
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    static int initialized = 0;
    PHARDWARE  phw = pDrvCtrl->pHW;



    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xSHRestartSetup(): Start initializes hardware chip... \n"));

    /* set promiscuous mode if it's asked for */

    if (END_FLAGS_GET (&pDrvCtrl->endObj) & IFF_PROMISC)
    {
        pDrvCtrl->flags |= KS_PROMISCUOUS_FLAG; 
        DI->bRxPromiscuous = TRUE;
        phw->m_bPromiscuous = TRUE;
    }
    else
    {
        pDrvCtrl->flags &= ~KS_PROMISCUOUS_FLAG; 
        DI->bRxPromiscuous = FALSE;
        phw->m_bPromiscuous = FALSE;
    }

    /* set receive all multicast frames mode if it's asked for */

    if (END_FLAGS_GET (&pDrvCtrl->endObj) & (IFF_ALLMULTI | IFF_MULTICAST))
    {
        pDrvCtrl->flags |= KS_MULTICAST_FLAG; 
        DI->bRxMulticast = TRUE;
        phw->m_bAllMulticast = TRUE;
    }
    else
    {
        pDrvCtrl->flags &= ~KS_MULTICAST_FLAG; 
        DI->bRxMulticast = FALSE;
        phw->m_bAllMulticast = FALSE;
        phw->m_bMulticastListSize = 0;
    }

    /* only execute the following code once. */
    if(initialized == 0)
    {

       /* 
        * Set ks884xSH internal API function point 
        */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling ksSetApiFunctions()... \n"));
        if ( ksSetApiFunctions ( pDrvCtrl) == ERROR )
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                     ("ks884xSHRestartSetup(unit=%d): ERROR on set ks884xSH internal API. \n", pDrvCtrl->unit));
            return ERROR ;
        }
        
#ifndef BRINGUP_BOARD 

       /* 
        * reset the device  
        */
#ifndef KS_QC_TEST
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling ks884xSHReset()... \n"));
        if ( ks884xSHReset (pDrvCtrl) == ERROR )
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                    ("ks884xSHRestartSetup(unit=%d): ERROR on reset ks884xSH. \n", pDrvCtrl->unit ));
            return ERROR ;
        }
#endif /* #ifndef KS_QC_TEST */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling HardwareInitialize()... \n"));
#ifndef KS_QC_TEST
        if ( !HardwareInitialize( phw ) ) 
#else
        if ( !QCCheckChipID( phw ) ) 
#endif /* #ifndef KS_QC_TEST */
        {
            DRV_PRINT(DRV_DEBUG_ERROR, 
                      ("ks884xSHRestartSetup(unit=%d): ERROR on incorrect ks884xSH chip ID. \n", pDrvCtrl->unit ));
#ifndef NOTCHECK_CHIPID 
            return ERROR ;
#endif 
        }

        /* set up address filter for multicasting */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                ("ks884xSHRestartSetup(): calling ks884xSHAddrFilterSet()... \n"));
        if (END_MULTI_LST_CNT (&pDrvCtrl->endObj) > 0)
        {
            ks884xSHAddrFilterSet (pDrvCtrl);
        }

        /* 
         * Clear MIB's Counters 
         */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling HardwareClearCounters(1)... \n"));
        phw->m_bPortSelect = 0;         /* clear Port 1 */
        HardwareClearCounters( phw );
#ifdef DEF_KS8842
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling HardwareClearCounters(2)... \n"));
        phw->m_bPortSelect = 1;         /* clear Port 2 */
        HardwareClearCounters( phw );
#endif
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling HardwareClearCounters(3)... \n"));
        phw->m_bPortSelect = 2;         /* clear Port 3 */
        HardwareClearCounters( phw );

        /* 
         * Initialize device's default configuration setting. 
         */ 

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling HardwareSetup()... \n"));
        HardwareSetup( phw );

#ifndef KS_QC_TEST
        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling HardwareSwitchSetup()... \n"));
        HardwareSwitchSetup ( phw );

        /* Wait for auto negotiation to complete. */
        DelayMillisec( 2000 );
#endif /* #ifndef KS_QC_TEST */

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling SwitchSetAddress()... \n"));
        /* Set Switch function address */
        SwitchSetAddress ( phw, phw->m_bOverrideAddress );

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): calling SwitchGetLinkStatus()... \n"));
        /* Get Port Link Status */
        SwitchGetLinkStatus( phw );

        DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
                  ("ks884xSHRestartSetup(): done ! \n"));

#endif /* #ifdef BRINGUP_BOARD */

        initialized ++;
    }


    CACHE_PIPE_FLUSH ();

    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xSHRestartSetup(): Finished initializes hardware chip. \n"));

    return (OK);
}


/*******************************************************************************
*
* ks884xSHRestart - restart the device after a fatal error
*
* This routine takes care of all the messy details of a restart.  The device
* is reset and re-initialized.  The driver state is re-synchronized.
*
* RETURNS: N/A
* TODO: call this routine if something fail 
*/
LOCAL void ks884xSHRestart
(
    DRV_CTRL * pDrvCtrl  /* device to be initialized */
)
{
    if (ks884xSHRestartSetup (pDrvCtrl) == OK)
    {
        /* set the flags to indicate readiness */

        END_OBJ_READY (&pDrvCtrl->endObj, IFF_UP | IFF_RUNNING |
                       IFF_NOTRAILERS | IFF_BROADCAST | IFF_MULTICAST);
    }
}


/******************************************************************************
*
* ks884xSHAddrFilterSet - set the address filter for multicast addresses
*
* This routine goes through all of the multicast addresses on the list
* of addresses and sets the
* device's logical address filter in the initialization block.
*
* RETURNS: N/A
*
* SEE ALSO:  
*/
LOCAL void ks884xSHAddrFilterSet
(
    DRV_CTRL * pDrvCtrl             /* pointer to DRV_CTRL structure */
)
{
    UINT8             mcast [MAC_ADDRESS_LENGTH];   /* temporary logical address */
    PHARDWARE         phw = pDrvCtrl->pHW;
    ETHER_MULTI *     pCurr = END_MULTI_LST_FIRST (&pDrvCtrl->endObj);



    DRV_LOG(DRV_DEBUG_MCAST, "ks884xSHAddrFilterSet(unit=%d) is called.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

#if (0) /* todo */

    /* clear all the MAC address filters */

    HardwareAddrFilterClear ( phw );

    while (pCurr != NULL)
    {
        /* copy the specified address, as we'll destroy it with shifts below */

        mcast[0] = pCurr->addr[0];
        mcast[1] = pCurr->addr[1];
        mcast[2] = pCurr->addr[2];
        mcast[3] = pCurr->addr[3];
        mcast[4] = pCurr->addr[4];
        mcast[5] = pCurr->addr[5];


        /* Set up MAC address filter. */

        if ( !( HardwareAddrFilterAdd (phw, mcast) ) )
	        DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHAddrFilterSet: exceed MAC address filter table.\n", 1, 2, 3, 4, 5, 6);

        pCurr = END_MULTI_LST_NEXT(pCurr);
    }
#endif
}

/*******************************************************************************
*
* ks884xSHEndPollReceive - routine to receive a packet in polled mode.
*
* This routine is called by a user to try and get a packet from the
* device. This routine return OK if it is successful in getting the packet
*
* RETURNS: OK or EAGAIN.
*/
STATUS ks884xSHEndPollReceive
(
    DRV_CTRL  * pDrvCtrl,  /* points to a device control structure */
    M_BLK_ID    pMblk      /* mBlk chain stores received data */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    int         packetLen = 0;
    BOOL        fStatus=TRUE;
    int         portIndex = 0;
    BOOL        fBroadcast;


#ifndef BRINGUP_BOARD 


    if ((pMblk->mBlkHdr.mFlags & M_EXT) != M_EXT)
    {
        return (EAGAIN);
    }

	DRV_LOG(DRV_DEBUG_POLL_RX, "ks884xSHEndPollReceive (unit=%d)\n", 
                                pDrvCtrl->unit, 2, 3, 4, 5, 6);


    if ( pMblk != NULL )
    {
        /* precess receive packet */

        fStatus = ks884xSHRecv (pDrvCtrl, pMblk->mBlkHdr.mData, &packetLen, &fBroadcast );

        if ( fStatus )
        {
            /* Call the upper layer's receive routine. */
            pMblk->mBlkHdr.mLen = packetLen;
            pMblk->mBlkHdr.mFlags |= M_PKTHDR;
            pMblk->mBlkPktHdr.len = packetLen;

            /* Add one to our unicast data. */
            END_ERR_ADD (&pDrvCtrl->endObj, MIB2_IN_UCAST, +1);
        }
    }
    else
    {
            /* receive packet drop because no buffers in the OS	*/
            phw->m_cnCounter[ portIndex ][ OID_COUNTER_RCV_DROPPED ] +=1;
            fStatus = FALSE;
    } /* if ( pMblk != NULL ) */


    /*
     * Update HW registers to indicate that this frame has been processed.
     */

    /* Free up that memory */
    HardwareReleaseReceive( phw );

#endif /* #ifndef BRINGUP_BOARD */

    return (fStatus);
}

/*******************************************************************************
*
* ks884xSHEndPollSend - send a packet in polled mode.
*
* This routine is called by a user to try and send a packet on the
* device.
*
* INTERNAL
* A return value of OK indicates that the packet has been sent.  In this
* case, the upper level protocol (WDB for example) will free the mBlk and
* associated cluster.
*
* RETURNS: OK or EAGAIN.
*/
STATUS ks884xSHEndPollSend
(
    DRV_CTRL * pDrvCtrl,  /* points to device control structure */
    M_BLK_ID   pMblk      /* mBlk chain stores transmitted data */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
	int		   dataLen;
	char	  *pData;
    BOOL       fStatus=TRUE;
    ULONG      portIndex=0;


#ifndef BRINGUP_BOARD 

	DRV_LOG(DRV_DEBUG_POLL_TX, "ks884xSHEndPollSend (unit=%d)\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    if ( pMblk )
    {
        /* get data pointer and data length in this mBlk */ 
	    pData = pMblk->mBlkHdr.mData;
        dataLen = pMblk -> mBlkHdr.mLen;

        /* Send packet data (from pMblk) to the device. */
        fStatus = ks884xSHSend (pDrvCtrl, pData, dataLen );

        /* Checking ks8842 TXQ memory available for this packet */
        if ( ! fStatus ) 
        {
            DRV_LOG(DRV_DEBUG_POLL_TX, "ks884xSHSend() fail!\n", 1, 2, 3, 4, 5, 6);

            END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_ERRS, +1);
		    return END_ERR_BLOCK;
        }

        /* Issue TXQ Command (Enqueue Tx frame from TX buffer into TXQ for transmit) */
        if ( HardwareSendPacket( phw ) ) 
        {
            /* Update driver transmit counters */
        if ( portIndex == 3 )
        {
            while ( portIndex & 3)
            {
               portIndex >>= 1;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;    
             }
        }
        else
        {
               if ( portIndex > 0 ) 
                   portIndex --;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_BYTES_XMIT ] += dataLen;
               phw->m_cnCounter[ portIndex ][ OID_COUNTER_DIRECTED_FRAMES_XMIT ]++;    
        }

            /* increment the unicast output counter */
            END_ERR_ADD (&pDrvCtrl->endObj, MIB2_OUT_UCAST, +1);
        }
        else
        {
            DRV_LOG(DRV_DEBUG_POLL_TX, "ks884xSHEndPollSend(unit=%d) device Transmit packet fail!\n", 
                    pDrvCtrl->unit, 2, 3, 4, 5, 6);
        }

        /* Free pMblk */
        netMblkClChainFree (pMblk);

    } /* if ( pMblk ) */

#endif /* #ifndef BRINGUP_BOARD */

    return (OK);
}

/*****************************************************************************
*
* ks884xSHEndMCastAddrAdd - add a multicast address for the device
*
* This routine adds a multicast address to whatever the driver
* is already listening for.  It then resets the address filter.
*
* NOTE
* etherMultiAdd() returns EINVAL for invalid multicast addresses.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xSHEndMCastAddrAdd
(
    DRV_CTRL * pDrvCtrl,  /* pointer to DRV_CTRL structure */
    char *     pAddress   /* address to be added */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;

    /*
     * ENETRESET is returned to inform the driver that the list has
     * changed and its reception filter should be adjusted accordingly.
     */
 
    int retVal = etherMultiAdd (&pDrvCtrl->endObj.multiList, pAddress);

    if (retVal == ENETRESET)
    {
        /* edit Logical Addresss Filter Registers */

#if (0) /* todo */
       if ( HardwareAddrFilterAdd (phw, pAddress) )
          retVal = OK;
       else
#endif
          retVal = ERROR;
    }

    return ((retVal == OK) ? OK : ERROR);
}

/*****************************************************************************
*
* ks884xSHEndMCastAddrDel - delete a multicast address for the device
*
* This routine removes a multicast address from whatever the driver
* is listening for.  It then resets the address filter.
*
* NOTE
* etherMultiAdd() returns EINVAL for invalid multicast addresses.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ks884xSHEndMCastAddrDel
(
    DRV_CTRL * pDrvCtrl,  /* pointer to DRV_CTRL structure */
    char *     pAddress   /* address to be deleted */
)
{
    PHARDWARE         phw = pDrvCtrl->pHW;
    /*
     * ENETRESET is returned to inform the driver that the list has
     * changed and its reception filter should be adjusted accordingly.
     */
 
    int retVal = etherMultiDel (&pDrvCtrl->endObj.multiList, pAddress);


    if (retVal == ENETRESET)
    {
        /* edit Logical Addresss Filter Registers */

#if (0) /* TODO */
       if ( HardwareAddrFilterDel (phw, pAddress) )
           retVal = OK;
       else
#endif
           retVal = ERROR;
    }

    return ((retVal == OK) ? OK : ERROR);
}

/*****************************************************************************
*
* ks884xSHMCastAddrGet - get the multicast address list for the device
*
* This routine gets the multicast list of whatever the driver
* is already listening for.
*
* RETURNS: OK/ERROR
*/
LOCAL STATUS ks884xSHMCastAddrGet
    (
    DRV_CTRL * 	pDrvCtrl,  /* pointer to DRV_CTRL structure */
    MULTI_TABLE * 	pTable     /* table where address will be copied */
    )
    {
    return (etherMultiGet (&pDrvCtrl->endObj.multiList, pTable));
    }

/*******************************************************************************
*
* ks884xSHEndStop - stop the device
*
* This function calls BSP functions to disconnect interrupts and stop
* the device from operating in interrupt mode.
*
*
* RETURNS: OK or ERROR if the interrupt handler could not be disconnected.
*/
LOCAL STATUS ks884xSHEndStop
(
    DRV_CTRL * pDrvCtrl  /* device control structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    STATUS     result = OK;



	DRV_LOG(DRV_DEBUG_STOP, "ks884xSHEndStop (unit=%d) is called.\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

	if (NULL != pDrvCtrl->watchdog) 
    {
		wdDelete(pDrvCtrl->watchdog);
		pDrvCtrl->watchdog = NULL;
	}

#ifndef BRINGUP_BOARD 
    /* Disable interrupts on the device-side */
    HardwareDisableInterrupt ( phw );

    /* Disable system level interrupt */
    (* ks884xSHIntDisable) (pDrvCtrl->ilevel);

#ifdef DEF_KS8842

    /* Disable the device swtich */
    SwitchEnable ( phw, FALSE );
#endif /*#ifdef DEF_KS8842 */

    /* Disables the device QMU. */
    HardwareDisable( phw );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    /* disconnect the interrupt handler */

    result = (* ks884xSHIntDisconnect) ((VOIDFUNCPTR *)
             INUM_TO_IVEC (pDrvCtrl->inum), ks8842xEndSHInt, (int)(pDrvCtrl));

#endif /* #ifdef BRINGUP_BOARD */

    /* mark the interface as down */

    END_FLAGS_CLR (&pDrvCtrl->endObj, (IFF_UP | IFF_RUNNING));


    return (result);
}

/******************************************************************************
*
* ks884xSHEndUnload - unload a driver from the system
*
* This function first brings down the device, and then frees any
* stuff that was allocated by the driver in the load function. The controller
* structure should be free by who ever is calling this function.
*
* RETURNS: OK
*/
LOCAL STATUS ks884xSHEndUnload
(
    DRV_CTRL * pDrvCtrl  /* device control structure */
)
{


 	DRV_LOG(DRV_DEBUG_UNLOAD, "ks884xSHEndUnload(unit=%d).\n", pDrvCtrl->unit, 2, 3, 4, 5, 6);

    END_OBJECT_UNLOAD (&pDrvCtrl->endObj);

    /* Free the shared DMA memory. */

    if (pDrvCtrl->flags & KS_MEM_ALLOC_FLAG)
        cacheDmaFree ((void *)(pDrvCtrl->pShMem));

    /* Free the shared DMA memory allocated for clusters */
    
    if (pDrvCtrl->flags & KS_MEM_ALLOC_FLAG)
        cacheDmaFree (pDrvCtrl->clDesc.memArea);

    /* Free the memory allocated for mBlks and clBlks */
    
    if (pDrvCtrl->mClCfg.memArea != NULL)
        free (pDrvCtrl->mClCfg.memArea);

    /* Free the memory allocated for driver pool structure */

    if (pDrvCtrl->endObj.pNetPool != NULL)
        free (pDrvCtrl->endObj.pNetPool);

    return (OK);
}

/*******************************************************************************
*
* ks884xSHEndPollStart - start polled mode operations
*
* RETURNS: OK always.
*/
STATUS ks884xSHEndPollStart
(
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
#ifndef BRINGUP_BOARD 
    PHARDWARE  phw = pDrvCtrl->pHW;

 

    /* Disable interrupts on the device-side */
    HardwareDisableInterrupt ( phw );

    /* Clear interrupts on the device-side */
    HardwareAcknowledgeInterrupt( phw, phw->m_wInterruptMask );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

    pDrvCtrl->flags |= KS_POLLING;

    DRV_LOG (DRV_DEBUG_POLL, "POLLING STARTED\n", 0, 0, 0, 0, 0, 0);

#endif /* #ifndef BRINGUP_BOARD */

    return (OK);
}

/*******************************************************************************
*
* ks884xSHEndPollStop - stop polled mode operations
*
* RETURNS: OK always.
*/
STATUS ks884xSHEndPollStop
(
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;
    int intLevel = intLock();


#ifndef BRINGUP_BOARD 

    /* Clear interrupts on the device-side */
    HardwareAcknowledgeInterrupt( phw, phw->m_wInterruptMask );

    /* Enable interrupts on the device-side */
    HardwareEnableInterrupt( phw );

    /* Flush the write pipe */
    CACHE_PIPE_FLUSH ();

#endif /* #ifdef BRINGUP_BOARD */

    pDrvCtrl->flags &= ~KS_POLLING;

    intUnlock (intLevel);

    DRV_LOG (DRV_DEBUG_POLL, "POLLING STOPPED\n", 0, 0, 0, 0, 0, 0);

    return (OK);
}

/******************************************************************************
*
* ks884xSHHandleTxDoneInt - clean the transmit ring
*
* This routine is called at task level indirectly by the interrupt
* service routine to  clean the transmit ring.

* This routine frees driver clusters associated with Tx descriptor that have used
* the clusters for transmitting packets.  
*
* RETURNS: N/A
*
* NOMANUAL
*/
LOCAL void ks884xSHHandleTxDoneInt
(
    DRV_CTRL * pDrvCtrl,  /* pointer to DRV_CTRL structure */
    UINT16     wIER       /* holds INT_TX_DONE bit read from interrupt status */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;


    DRV_LOG(DRV_DEBUG_TX, "ks884xSHHandleTxDoneInt (%d), wIER=0x%04X\n", pDrvCtrl->unit, wIER, 3, 4, 5, 6);

    pDrvCtrl->txCleaning = TRUE;
    
    /* clean up processed tx descriptors */
	ks884xSHTxDoneProcess(pDrvCtrl);

	/* enable Tx interrupts */
    HardwareEnableInterruptBit ( phw, (USHORT)(INT_TX & phw->m_wInterruptMask) );

    pDrvCtrl->txCleaning = FALSE;
}


/*******************************************************************************
*
* ks884xSHTxDoneProcess - 
*
* Clear the Transmit descriptors when chip has completed the transmit the packets.
*
* RETURNS: N/A.
*/
LOCAL void ks884xSHTxDoneProcess
(
    DRV_CTRL * pDrvCtrl  /* points to device control structure */
)
{
    PHARDWARE  phw = pDrvCtrl->pHW;


    /* Transmit successful */
    HardwareTransmitDone( phw );

}


/*******************************************************************************
*
* ks884xSHParamInit - set global parameters for endianess and interrupt level
*
* ???
*
* RETURNS: OK Always.
*
* NOMANUAL
*/
STATUS ks884xSHParamInit
(
    int pciBusEndianess,
    int intLevel
)
{
    ks884xSHBusEndianess   = pciBusEndianess;
    ks884xSHInterruptLevel = intLevel;

    return (OK);
}


/*
 * ks884xSHAutoNegoStart
 *	This function is used to start auto negotiation
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xSHAutoNegoStart
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
	PORT_KS_INFO *DPI,     /* ks884xSH port related (switch port only) */
    ULONG      uPort
)
{
#ifndef BRINGUP_BOARD 

    PHARDWARE   phw = pDrvCtrl->pHW;


    /* 
     * auto-negotiation mode 
     */
	if (SW_PHY_AUTO == DPI->myLinkType) 
    {
        /* if auto-negotiation not in progress, and Link state has changed */
	    if ( !DPI->bAutoNegoInProgress && DPI->bLinkChanged) 
        {
            /* Start auto-negotiation on this port */
		    SwitchRestartAutoNego( phw );

           	DPI->bAutoNegoInProgress = TRUE;
		    DPI->bLinkChanged = FALSE;
		    DPI->bLinkActive  = FALSE;
        }
    }

	ks884xSHGetPhyStatus ( pDrvCtrl, DPI, uPort );

#endif /* #ifndef BRINGUP_BOARD */

}

/*
 * ks884xSHGetPhyStatus
 *	This function is used to get the status of auto negotiation.
 *
 * Argument(s)
 *	pDrvCtrl	pointer to DRV_CTRL structure.
 *
 * Return(s)
 *	TRUE	link up
 *	FALSE	link down
 */
LOCAL STATUS ks884xSHGetPhyStatus
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
	PORT_KS_INFO *DPI,     /* ks884xSH port related (switch port only) */
    ULONG      uPort
)
{
    PHARDWARE   phw  = pDrvCtrl->pHW;
    PPORT_INFO  pPortInfo = NULL;


#ifndef BRINGUP_BOARD 

    DRV_LOG(DRV_DEBUG_PHY, "ks884xSHGetPhyStatus\n", 1, 2, 3, 4, 5, 6);

    pPortInfo = &phw->m_PortInfo[ uPort ];

    if ( (phw == NULL) || (pPortInfo == NULL) )
    {
        DRV_LOG(DRV_DEBUG_ERROR, "ks884xSHGetPhyStatus(): phw or pPortInfo is NULL.\n", 1, 2, 3, 4, 5, 6);
        return FALSE;
    }

    /* Get port's PHY link status */
    SwitchGetLinkStatus( phw );

	/* 
     * PHY link is down 
     */

    if (pPortInfo->ulHardwareState == MediaStateDisconnected ) 
    {
		DPI->bLinkActive = FALSE;
		DPI->myLinkSpeed = SPEED_UNKNOWN;
		DPI->myLinkDuplex = 0;

        /* TODO : set link LED */

		DRV_LOG(DRV_DEBUG_PHY, "Link down.\n", 1, 2, 3, 4, 5, 6);
		return FALSE;
	}

	/* 
     * PHY link is up
     */

	DPI->bLinkActive = TRUE;

    /* 
     * auto-negotiation mode 
     */
	if (SW_PHY_AUTO == DPI->myLinkType) 
    {

        /* Link is full duplex */ 
		if (pPortInfo->bDuplex == FULL_DUPLEX ) 
        {
            /* force enable flow control, regardless auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, TRUE );

            /* disable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, FALSE );
        } 
        /* Link is half duplex */ 
        else 
        {
            /* flow control is enable based on auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, FALSE );

            /* enable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, TRUE );
		}

		DRV_LOG(DRV_DEBUG_PHY, "Link up.\n", 1, 2, 3, 4, 5, 6);
	}
	/* 
     * manually connection mode 
     */
	else   /* if (SW_PHY_AUTO == DI->myLinkType[uPort]) */
    {
		if ( (SW_PHY_10BASE_T_FD == DPI->myLinkType) || 
             (SW_PHY_100BASE_TX_FD == DPI->myLinkType ) )
        {
            /* manually select link for full duplex */ 
			DPI->myLinkDuplex = FULL_DUPLEX;

            /* force enable flow control, regardless auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, TRUE );

            /* disable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, FALSE );
		}
		else 
        {
            /* manually selected link for half duplex */ 
			DPI->myLinkDuplex = HALF_DUPLEX;

            /* flow control is enable based on auto-negaotiation result */
            PortConfigForceFlowCtrl ( phw, uPort, FALSE );

            /* enable half duplex back pressure on this port */
            PortConfigBackPressure ( phw, uPort, TRUE );
		}

		if  ( (SW_PHY_100BASE_TX_FD == DPI->myLinkType) || 
              (SW_PHY_100BASE_TX    == DPI->myLinkType) )
        {
            /* manually selected link speed is 100Mbps */
			DPI->myLinkSpeed = SPEED_100;

            /* TODO : set link LED */
		}
		else 
        {
            /* manually selected link speed is 10Mbps */
			DPI->myLinkSpeed = SPEED_10;

            /* TODO : set link LED */
		}

    } /* if (SW_PHY_AUTO == DPI->myLinkType) */

#endif /* #ifndef BRINGUP_BOARD */

	return TRUE;
}


/*******************************************************************************
*
* ks884xSHEnetAddrGet - get ks884xSH Ethernet (IEEE station) address
*
* This routine provides an interface for accessing an ks884xSH device
* Ethernet (IEEE station address) address in the device.
* A handle to the specific device control structure is specified
* in the <pDrvCtrl> parameter.  The 6-byte IEEE station address will be
* copied to the memory location specified by the <enetAdrs> parameter.
*
*
* RETURNS: OK, or ERROR if could not be obtained.
*
* NOMANUAL
*/
STATUS ks884xSHEnetAddrGet
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
    char *     enetAdrs    /* specifies storage for station address */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    UCHAR       i;

	/* 
     * Get MAC address to the Station registers 
     */

#ifndef BRINGUP_BOARD 

    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xSHEnetAddrGet(): calling HardwareReadAddress()... \n"));

	/* Get MAC address from the Station registers */
    HardwareReadAddress( phw );

#if (0)
	printf("ks884xSHEnetAddrGet() MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", 
		    phw->m_bPermanentAddress[0], 
		    phw->m_bPermanentAddress[1],
		    phw->m_bPermanentAddress[2],
		    phw->m_bPermanentAddress[3], 
		    phw->m_bPermanentAddress[4],
		    phw->m_bPermanentAddress[5]); 
#endif

    for (i=0; i<ETH_LENGTH_OF_ADDRESS; i++)
         enetAdrs[i] = phw->m_bPermanentAddress[i /* ETH_LENGTH_OF_ADDRESS-i-1 */] ;

#endif /* #ifdef BRINGUP_BOARD */

    return OK;
}

/*******************************************************************************
*
* ks884xSHEnetAddrSet - Set ks884xSH Ethernet (IEEE station) address
*
* This routine provides an interface for accessing an ks884xSH device
* Ethernet (IEEE station address) address in the device.
* A handle to the specific device control structure is specified
* in the <pDrvCtrl> parameter.  The 6-byte IEEE station address,
* specified by the <enetAdrs> parameter, will be wrote to ks884xSH device
* MAC address registers .
*
*
* RETURNS: OK, or ERROR if could not be obtained.
*
* NOMANUAL
*/
STATUS ks884xSHEnetAddrSet
(
    DRV_CTRL * pDrvCtrl,   /* Driver control structure */
    char *     enetAdrs    /* specifies storage for station address */
)
{
    PHARDWARE   phw = pDrvCtrl->pHW;
    UCHAR         i;


	/* 
     * Set MAC address to the Station registers 
     */

#ifndef BRINGUP_BOARD 

    DRV_PRINT(DRV_DEBUG_CHIP_INIT, 
              ("ks884xSHEnetAddrSet(): calling HardwareSetAddress()... \n"));


    phw->m_bMacOverrideAddr = TRUE;
    for (i=0; i<MAC_ADDRESS_LENGTH; i++)
         phw->m_bOverrideAddress[i] = enetAdrs[i];
    HardwareSetAddress( phw );


	/* Get MAC address from the Station registers */
    HardwareReadAddress( phw );

    if (memcmp ( enetAdrs, phw->m_bPermanentAddress, ETH_LENGTH_OF_ADDRESS ) != 0)
        return ERROR;

	memcpy(pDrvCtrl->enetAddr, phw->m_bPermanentAddress, ETH_LENGTH_OF_ADDRESS);  

#endif /* #ifdef BRINGUP_BOARD */

    return OK;

}

/*
 * ks884xSHSetDefaults
 *	This function is used to set initial defaults before Parse routine.
 *
 * Argument(s)
 *	pDrvCtrl		pointer to DRV_CTRL structure.
 *
 * Return(s)
 *	NONE.
 */

LOCAL void ks884xSHSetDefaults
(
    DRV_CTRL * pDrvCtrl   /* Driver control structure */
)
{
    ULONG      uPort;
    PHARDWARE  phw = pDrvCtrl->pHW;
	PORT_KS_INFO *DPI;      /* ks884xSH port related (switch port only) */



	/* Transmit Descriptor Count */
	DI->nTxBufTotal = 4 /* TXBUF_DEFAULT */;		

	/* Receive Descriptor Count */
	DI->nRxBufTotal = 4 /* RXBUF_DEFAULT */;		


	/* currently Tx control flow shares the setting of Rx control flow */
	DI->bTxFlowCtrl = DI->bRxFlowCtrl;

    /* Receive buffer size */
	DI->uRxBufferLen = BUFFER_2044;	

	/* Perform PHY PowerDown Reset instead of soft reset, the Option function can 
	   be overwritten by user later */
	DI->bPowerDownReset = PHY_POWERDOWN_DEFAULT;

    /*
     * Set switch MAC address.
     */

	/* set default mac address for WAN, with Kendin's manufacturer ID 00:10:A1 */
	memcpy(DI->stMacStation, "\0\x10\xa1ISA", ETH_LENGTH_OF_ADDRESS);
#if (0)
	memset(DI->stMacStation, 0, ETH_LENGTH_OF_ADDRESS);
    DI->stMacStation[5] = 0x03;
#endif

    /*
     * Set default values for Port relative.
     */

    for (uPort=0; uPort<MAX_SWITCH_PORT; uPort++)
    {
        DPI = &pDrvCtrl->port[uPort];

	    /* PHY auto-negatiation */
	    DPI->myLinkType = PHY_SPEED_DEFAULT;

        /* disable Spanning Tree */
	    DPI->bySTPenable = PORT_STP_DEFAULT;

        /* enable broadcast storm protection */
	    DPI->byStormProtection = PORT_STORM_DEFAULT;

    } /* for (uPort=0; uPort<MAX_SWITCH_PORT; uPort++) */

    /*
     * Set default values for CPU Port relative.
     */

    /* TODO : */


    /* watchdog periode timer (every 5 sec) */
	pDrvCtrl->nWatchdogDelay = sysClkRateGet() * 5;		

    /* Set default call ks884xSHEndPollSend not from CLI command */
    pDrvCtrl->fPollSendFromCLI = FALSE;

    /* Set default for HW struct */
    phw->fDebugDumpTx = FALSE;
    phw->fDebugDumpRx = FALSE;

    phw->f_dircetMode = 1;   /* default transmit to device by MAC lookup mode */
    phw->m_bPortTX = 0;

}


/*
 * ks884xSHAllocMblk
 *	This function is use to replace used receive buffers with new mblk/clblk/cluster
 *	to Rx descriptors.
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *	pCluster	pointer to cluster data buffer associated with the mblk/clblk/cluster
 *	pSrcMblk	pointer to mblk with source data to copy if not NULL
 *
 * Return(s)
 *	pointer to mblk if successful
 *	NULL	otherwise
 */
M_BLK_ID ks884xSHAllocMblk
(   
    DRV_CTRL * pDrvCtrl,    /* pointer to DRV_CTRL structure */
    UCHAR    **ppCluster, 
    M_BLK_ID   pSrcMblk
)
{
	UCHAR     *pNewCluster, *pClusterPtr = NULL;
    CL_BLK_ID  pClBlk=NULL;
	M_BLK_ID   pMblk=NULL;



    /*
     * We implicitly are loaning here, if copying is necessary this
     * step may be skipped, but the data must be copied before being
     * passed up to the protocols.
     */
    pNewCluster = netClusterGet(pDrvCtrl->endObj.pNetPool, pDrvCtrl->pClPoolId);
    if (pNewCluster == NULL)
        goto ERREXIT;

    /* Grab a cluster block to marry to the cluster we received. */
    if ((pClBlk = netClBlkGet(pDrvCtrl->endObj.pNetPool, M_DONTWAIT)) == NULL)
        goto ERREXIT;

    if ((pMblk = netMblkGet(pDrvCtrl->endObj.pNetPool, M_DONTWAIT, MT_DATA)) == NULL)
        goto ERREXIT;

	if (NULL == pSrcMblk) 
    {
		pClusterPtr = pNewCluster;
	}
	else 
    {
		/* copy data from source mblk to pClusterPtr */
		DRV_LOG(DRV_DEBUG_RX, "not implemented yet!\n", 1, 2, 3, 4, 5, 6);
		pClusterPtr = pNewCluster;
	}

    /* Join the cluster to the MBlock, note that length should be at least DI->uRxBufferLen */
    netClBlkJoin(pClBlk, pClusterPtr, DI->uRxBufferLen, NULL, 0, 0, 0);
    netMblkClJoin(pMblk, pClBlk);

    /* make the packet data coherent */
    *ppCluster = pClusterPtr;
	return pMblk;

ERREXIT:
    if (pNewCluster != NULL)
        netClFree(pDrvCtrl->endObj.pNetPool, pNewCluster);

    if ( pClBlk != NULL)
        netClBlkFree(pDrvCtrl->endObj.pNetPool, pClBlk);

    if ( pMblk != NULL)
        netMblkFree(pDrvCtrl->endObj.pNetPool, pMblk);

	DRV_LOG(DRV_DEBUG_RX, "ks884xSHAllocMblk(): Out of mblk/clblk/cluster!\n", 
                              1, 2, 3, 4, 5, 6);
	END_ERR_ADD(&pDrvCtrl->endObj, MIB2_IN_ERRS, +1);
	return NULL;
}

/*
 * ks884xSHWatchdog
 *	This function is a timer callback routine to check link status on a given port.
 *  If link is down, perform a link auto-negotiation.
 *  Timer periodic is 5 seconds. 
 *
 * Argument(s)
 *  pDrvCtrl	pointer to DRV_CTRL struct
 *
 * Return(s)
 *	NONE.
 */
LOCAL void ks884xSHWatchdog
(   
    DRV_CTRL * pDrvCtrl    /* pointer to DRV_CTRL structure */
)
{
    ULONG        uPort;
    UINT8	     bLinkActive;
	PORT_KS_INFO *DPI;      /* ks884xSH port related (switch port only) */


    DRV_LOG(DRV_DEBUG_WATCHDOG, "ks884xSHWatchdog running ...\n", 1, 2, 3, 4, 5, 6);

}

/*
 *-------------------------------------------------------------------------------------------
 *	The following functions can be called from Shell interface as GUI for 
 *  user to debug.
 *--------------------------------------------------------------------------------------------
 */


/*
 * setArea4WaitState
 *	  This function is set Area 4 the number of idle cycles between bus cycles 
 *    to be inserted from a read access to a write access in the same space 
 *    by SH7751R WCR1 register.
 *    Also set Area 4 the number of wait states (Tw) to be inserted 
 *    by SH7751R WCR2 register
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
void  setArea4WaitState
(
)
{
#define BCR1_ADDR   ((volatile UINT32 *)0xFF800000)	/* Bus  Control Register 1 */
#define BCR2_ADDR   ((volatile UINT16 *)0xFF800004)	/* Bus  Control Register 2 */
#define WCR1_ADDR   ((volatile UINT32 *)0xFF800008)	/* Wait Control Register 1 */
#define WCR2_ADDR   ((volatile UINT32 *)0xFF80000c)	/* Wait Control Register 2 */
#define WCR3_ADDR   ((volatile UINT32 *)0xFF800010)	/* Wait Control Register 3 */

    ULONG  bcr1_data;
    USHORT bcr2_data;
    ULONG  wcr1_data;
    ULONG  wcr2_data;
    ULONG  wcr3_data;


    /* Set Area 4 to 32bit/16Bit data width */
    bcr2_data = *(BCR2_ADDR); 
    bcr2_data &= ~(0x0300);  
#ifdef KS_16BIT_BUS
    bcr2_data |=  (0x0200);     /* 16bit external bus */
#else
    bcr2_data |=  (0x0300);     /* 32bit external bus */
#endif
    *(BCR2_ADDR) = bcr2_data;

    /* Set Area 4 to Byte Control Mode */
    bcr1_data = *(BCR1_ADDR); 
#ifdef SDRAM_LIKE_INTERFACE
    bcr1_data &= ~(0x00100000); /* Normal Mode */  
#else
    bcr1_data |= (0x00100000);  /* Byte Control Mode */
#endif
    *(BCR1_ADDR) = bcr1_data;

    /* Set Area4 to 1 idle cycle between a read access to a write access */
    wcr1_data = *(WCR1_ADDR); 
    wcr1_data &= ~(0x00070000);   
    wcr1_data |=  (0x00010000);   
    *(WCR1_ADDR) = wcr1_data;

    /* Set Area4 to no setup time and holding time  */
    wcr3_data = *(WCR3_ADDR); 
    wcr3_data &= ~(0x000F0000);   
    wcr3_data |=  (0x000F0000);   /* with setup and hold time */
    *(WCR3_ADDR) = wcr3_data;

}

/**
 *  Convert device ID to device name. 
 *
 */ 
LOCAL void getDeviceName 
(
    USHORT   ChipID
)
{
    if (ChipID == CHIP_IP_41_ISA)
        fKS884x = 1;
    else if (ChipID == CHIP_IP_42_ISA)
        fKS884x = 2;
    else if (ChipID == CHIP_IP_61_ISA)
        fKS884x = 3;
    else if (ChipID == CHIP_IP_62_ISA)
        fKS884x = 4;
    else
        fKS884x = 0;  /* unknown device */
}


/*
 * showKsBspInfo
 *	  This function is use to show driver's BSP info
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
void showKsBspInfo 
(
)
{
    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
    PHARDWARE  phw = pDrvCtrl->pHW;
    char  hostName[16];
    USHORT   ChipID;
    UCHAR    DevRevisionID;


#ifndef BRINGUP_BOARD 

#ifndef KS_QC_TEST

    HardwareReadChipID ( phw, &ChipID, &DevRevisionID );

    /* Display Device Driver info on UART */
    printf( NEWLINE );
	printf("Micrel - Ethernet Group."NEWLINE);
    printf("%s 32-bit Generic Bus Ethernet Driver demo."NEWLINE, deviceName[fKS884x] );

    printf(pDescription, deviceName[fKS884x]);
#ifndef  TEST_THROUGHPUT
    printf(ks884xSHEnd_ver);	   
#else
    printf(ks884xSHEnd_throughput_ver);	   
#endif /* #ifndef  TEST_THROUGHPUT */
    printf( NEWLINE );
    printf("%s Driver Version: " KS884X_VERSION, deviceName[fKS884x]); /* KS884X Driver Version */
    printf( NEWLINE );

    DBG_PRINT("%s Chip ID       : %04x", deviceName[fKS884x], ChipID);     /* KS884X Chip ID */
    DBG_PRINT( NEWLINE );
    DBG_PRINT("%s Device Revision: %d", deviceName[fKS884x], DevRevisionID); /* KS884X Device Revision */
    DBG_PRINT( NEWLINE );

    printf( "%s Base Address: 0x%08X"NEWLINE, deviceName[fKS884x], (int)phw->m_ulVIoAddr );

	printf("MAC address %02X:%02X:%02X:%02X:%02X:%02X\n", 
		    pDrvCtrl->enetAddr[0], 
		    pDrvCtrl->enetAddr[1],
		    pDrvCtrl->enetAddr[2],
		    pDrvCtrl->enetAddr[3], 
		    pDrvCtrl->enetAddr[4],
		    pDrvCtrl->enetAddr[5]); 

    memset(hostName, 0, sizeof(hostName));
    sprintf(hostName, "%s%d", KS_DEV_NAME, pDrvCtrl->unit );
    printf("IP  Address: ");	   
    if ( ifAddrGet ( hostName, ksIpAddress ) != OK )
    {
        printf("unknown (fail to get IP address of device %s !)"NEWLINE, hostName );	   
    }
    else
        printf("%s"NEWLINE, ksIpAddress);	   
    printf( NEWLINE );

#else  /* #ifndef KS_QC_TEST */

   /* For QC FAB testing */
    DBG_PRINT( NEWLINE );
    DBG_PRINT( "^[ok : %s]$"NEWLINE, ks884xQC_ver );

#endif  /* #ifndef KS_QC_TEST */

#endif /* #ifdef BRINGUP_BOARD */

}

/*
 * showKsLinkStatus
 *	  This function is use to show KS884x Link Status
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
void showKsLinkStatus 
(
)
{
#ifndef BRINGUP_BOARD 

#ifndef KS_QC_TEST

    DRV_CTRL * pDrvCtrl	= gpDrvCtrl[0]; 
    PHARDWARE  phw = pDrvCtrl->pHW;
    PPORT_INFO pInfo;
    UCHAR      bPort;



    SwitchGetLinkStatus( phw );

    /* Display Ethernet Link status on UART */

#ifdef DEF_KS8842
    for ( bPort = 0; bPort < SWITCH_PORT_NUM; bPort++ )
#else
    bPort = 0;
#endif
    {
        pInfo = &phw->m_PortInfo[ bPort ];

        printf("Port %d Link Status:  ", (bPort+1) );	   

        if ( pInfo->ulHardwareState == MediaStateConnected )
        {
	        /* Speed */
            if (pInfo->ulSpeed == 1000000)
            {
	            printf("100 Mbps, ");	   
            }
            else if (pInfo->ulSpeed == 100000)
            {
	            printf("10  Mbps, ");	   
            }
            else 
            {
	            printf("??? Mbps, ");	   
            }

	        /* Duplex */
            if ( pInfo->bDuplex == 2)
            {
	            printf("Full DPX");	   
            }
	        else if ( pInfo->bDuplex == 1)
            {
	            printf("Half DPX");	   
            }	
            else 
            {
	            printf("??? DPX");	   
            }
        }
        else
        {
	        printf("Down ");	   
        }
        printf( NEWLINE );

    } /* for ( bPort = 0; bPort < SWITCH_PORT_NUM; bPort++ ) */

    printf( NEWLINE );

#endif /* #ifndef KS_QC_TEST */
#endif /* #ifndef BRINGUP_BOARD  */
}


LOCAL BOOLEAN QCCheckChipID
(
    PHARDWARE phw 
)
{
#ifndef BRINGUP_BOARD 
    USHORT wData;


    /*
     * Set Bus Speed to 125MHz
     */

    HardwareSelectBank( phw, REG_BUS_CTRL_BANK );
    HW_WRITE_WORD ( phw, REG_BUS_CTRL_OFFSET, BUS_SPEED_125_MHZ );

    /*
     * Check ks884x Chip ID
     */

    HardwareReadRegWord( phw,
                         (UCHAR)REG_SWITCH_CTRL_BANK,
                         (UCHAR)REG_CHIP_ID_OFFSET,
                         (PUSHORT)&wData
                       );

    if  ( (( wData & SWITCH_CHIP_ID_MASK_41 ) != REG_CHIP_ID_41 ) &&
          (( wData & SWITCH_CHIP_ID_MASK_41 ) != REG_CHIP_ID_42 ) )
    {
        return (FALSE);
    }
#endif /* #ifndef BRINGUP_BOARD  */

    return( TRUE );

}  /* QCCheckChipID */


/*
 * ksGetHW
 *	  This function is use to get ks884xSH HW structure handler.
 *
 * Argument(s)
 *  None.
 * Return(s)
 *	NONE.
 */
PHARDWARE ksGetHW 
(
    ULONG deviceIndex 
)
{
    DRV_CTRL * pDrvCtrl; 

    if ( deviceIndex > 1)
        return (0);

    pDrvCtrl = gpDrvCtrl[deviceIndex]; 
    return ( pDrvCtrl->pHW );
}


void hw_read_byte
(   
    PHARDWARE phwi, 
    UCHAR     addr, 
    UINT8    *data 
)                                   
{                                                                           
    UCHAR   addrByDwordAligned, shiftBit;
    ULONG   dwDataRead;

    addrByDwordAligned = addr & 0x0C;
	shiftBit = (addr &0x03) * 8;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_byte() addr=0x%08x, addrByDwordAligned=0x%08x, shiftBit=%d \n", 
		addr, addrByDwordAligned, shiftBit));

	HW_READ_DWORD( phwi, addrByDwordAligned, &dwDataRead ) ;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_byte() dwDataRead=0x%08x \n", dwDataRead ));

    dwDataRead >>= shiftBit;                                                          
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_byte() dwDataRead=0x%08x \n", dwDataRead ));

    *data = (UINT8)dwDataRead;                                                
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_byte() data=0x%02x \n", *data ));
}

void hw_write_byte
(   
    PHARDWARE phwi, 
    UCHAR     addr, 
    UINT8     data 
)                                   
{                                                                           
    UCHAR   addrByDwordAligned, shiftBit;
    ULONG   dwDataRead, dwDataWrite=0;
    ULONG   dwDataMask=0xFF;

    addrByDwordAligned = addr & 0xC;
	shiftBit = (addr &0x03) * 8;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_byte() data=0x%02x, addr=0x%08x, addrByDwordAligned=0x%08x, shiftBit=%d \n", 
		                     data, addr, addrByDwordAligned, shiftBit));

    dwDataMask <<= shiftBit  ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_byte() dwDataMask=0x%08x \n", dwDataMask ));

	HW_READ_DWORD( phwi, addrByDwordAligned, &dwDataRead ) ;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_byte() dwDataRead=0x%08x \n", dwDataRead ));

    dwDataRead &= ~dwDataMask ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_byte() dwDataRead=0x%08x \n", dwDataRead));

    dwDataWrite = ( data << shiftBit ) ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_byte() dwDataWrite=0x%08x \n", dwDataWrite ));

    dwDataWrite |= dwDataRead ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_byte() dwDataWrite=0x%08x, addrByDwordAligned=0x%08x\n", 
		dwDataWrite, addrByDwordAligned));

    HW_WRITE_DWORD( phwi, addrByDwordAligned, dwDataWrite ) ;

}

void hw_read_word
(   
    PHARDWARE phwi, 
    UCHAR     addr, 
    UINT16   *data 
)                                   
{                                                                           
    UCHAR   addrByDwordAligned, shiftBit;
    ULONG   dwDataRead;

    addrByDwordAligned = addr & 0xC;
	shiftBit = (addr &0x03) * 8;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_word() addr=0x%08x, addrByDwordAligned=0x%08x, shiftBit=%d \n", 
		addr, addrByDwordAligned, shiftBit));

	HW_READ_DWORD( phwi, addrByDwordAligned, &dwDataRead ) ;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_word() dwDataRead=0x%08x \n", dwDataRead ));

    dwDataRead >>= shiftBit;                                                          
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_word() dwDataRead=0x%08x \n", dwDataRead ));

    *data = (UINT16)dwDataRead;                                                
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_read_word() data=0x%04x \n", *data ));
}

void hw_write_word
(   
    PHARDWARE phwi, 
    UCHAR     addr, 
    UINT16    data 
)                                   
{                                                                           
    UCHAR   addrByDwordAligned, shiftBit;
    ULONG   dwDataRead, dwDataWrite=0;
    ULONG   dwDataMask=0xFFFF;

    addrByDwordAligned = addr & 0xC;
	shiftBit = (addr &0x03) * 8;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_word() data=0x%04x, addr=0x%08x, addrByDwordAligned=0x%08x, shiftBit=%d \n", 
		                     data, addr, addrByDwordAligned, shiftBit));

    dwDataMask <<= shiftBit  ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_word() dwDataMask=0x%08x \n", dwDataMask ));

	HW_READ_DWORD( phwi, addrByDwordAligned, &dwDataRead ) ;
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_word() dwDataRead=0x%08x \n", dwDataRead ));

    dwDataRead &= ~dwDataMask ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_word() dwDataRead=0x%08x \n", dwDataRead));

    dwDataWrite = ( data << shiftBit ) ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_word() dwDataWrite=0x%08x \n", dwDataWrite ));

    dwDataWrite |= dwDataRead ; 
	DRV_PRINT(DRV_DEBUG_32BIT, ("hw_write_word() dwDataWrite=0x%08x \n", dwDataWrite ));

    HW_WRITE_DWORD( phwi, addrByDwordAligned, dwDataWrite ) ;

}


#endif /* #ifdef KS_ISA_BUS */
