/* sysLib.c - Hitachi MS7751SE01 system-dependent library */

/* Copyright 1998-2002 Wind River Systems, Inc. */

/*
modification history
--------------------
20Aug04     pcd  added support for Micrel ks884x network controller
01l,07oct03,h_k  removed VM_STAE_GUARDED from PCI IO MMU pages. (SPR #91055)
01k,13may02,frf  cleared diab warnings.
01j,06dec01,zl   removed sysTffs.c inclusion, now done outside of BSP.
01i,29nov01,zl   removed FRQCR init section, done in romInit.
01h,27nov01,zl   componentized virtual PCI support.
01h,16nov01,zl   updated WindML support.
01g,01nov01,zl   corrected NULL usage.
01f,11oct01,h_k  added #ifndef PRJ_BUILD before #ifdef INCLUDE_USB and renamed
                 sysUsb.c to usbPciStub.c (SPR #69828).
01e,24oct01,hk   added virtual PCI memory and I/O support.
01e,07sep01,h_k  added power control support (SPR #69838).
01d,11jun01,h_k  removed UltraIO from sysPhysMemDesc.
01c,28mar01,zl   corrected PCI support, added M1543C initialization.
01b,09jan01,rsh  add PCI and Networking functionality
01a,15dec00,rsh  written for ms7751se01.
*/

/*
DESCRIPTION
This library provides board-specific routines.  The chip drivers included are:

    shScifSio.c - SH7751 On-chip tty driver
    sh7700Timer.c - SH77xx On-chip Timer library
    nvRamToFlash.c - MS7751 On-board flash ROM library
    ataDrv.c - IDE/ATA driver

INCLUDE FILES: sysLib.h

SEE ALSO:
.pG "Configuration"
*/

/* includes */

#include "vxWorks.h"
#include "vme.h"
#include "cacheLib.h"
#include "stdlib.h"
#include "string.h"
#include "sysLib.h"
#include "intLib.h"
#include "logLib.h"
#include "tyLib.h"
#include "vxLib.h"
#include "config.h"
#include "private/vmLibP.h"
#if defined(INCLUDE_WDB) || defined(INCLUDE_DEBUG)
#  include "arch/sh/dbgShLib.h"
#endif
#ifdef  INCLUDE_ATA
#  include "drv/hdisk/ataDrv.h"
#endif  /* INCLUDE_ATA */
#if defined INCLUDE_PCI
#  include "drv/pci/pciConfigLib.h"
#  include "drv/pci/pciIntLib.h"
#  include "aliM1543CPibc.h"
#endif

/* globals */

#if defined(INCLUDE_MMU_BASIC) || defined(INCLUDE_MMU_FULL)

PHYS_MEM_DESC sysPhysMemDesc [] =
    {
    /* adrs and length parameters must be page-aligned (multiples of 0x1000) */

    /* flash reserved for bootrom */
    {
    (void *) ROM_BASE_ADRS,
    (void *) ROM_BASE_ADRS,
    ROM_SIZE,				/* 1MB */
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE_NOT  | VM_STATE_CACHEABLE_NOT
    },

    /* rest of flash */
    {
    (void *) (ROM_BASE_ADRS + ROM_SIZE),
    (void *) (ROM_BASE_ADRS + ROM_SIZE),
    MS7751SE_FLASH_SIZE - ROM_SIZE,	/* 3MB */
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
    },

    /* LED */
    {
    (void *) LED_REG_ADRS_P0,
    (void *) LED_REG_ADRS_P0,
    LED_REG_SIZE,			/* 4KB */
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
    },

    /* S-Dynamic ram (SDRAM)*/
    {
    (void *) LOCAL_MEM_LOCAL_ADRS,
    (void *) LOCAL_MEM_LOCAL_ADRS,
    LOCAL_MEM_SIZE,			/* up to 64MB */
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
#if (0)
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE
#endif
#if (1)
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
#endif
    }

/* #ifdef INCLUDE_KS884X_SHBUS_END */
    /* KS884x on Generic Bus Interface */
    {
    (void *) KS_MBD_ADDR_PHY,
    (void *) KS_MBD_ADDR_PHY,
    KS_REG_SIZE,			/* 4KB */
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
    },
/* #endif *//* ifdef INCLUDE_KS884X_SHBUS_END */

#ifdef INCLUDE_PCI
    ,
#ifndef INCLUDE_SH7751_VIRTUAL_PCI_IO
    /* PCI I/0 window */
    {
    (void *) PCI_MSTR_IO_LOCAL,
    (void *) PCI_MSTR_IO_LOCAL,
    PCI_MSTR_IO_SIZE,
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
    },
#endif /* INCLUDE_SH7751_VIRTUAL_PCI_IO */
 
#ifndef INCLUDE_SH7751_VIRTUAL_PCI_MEMIO
    /* PCI memory window */
    {
    (void *) PCI_MSTR_MEMIO_LOCAL,
    (void *) PCI_MSTR_MEMIO_LOCAL,
    PCI_MSTR_MEMIO_SIZE,
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
    },
#endif /* INCLUDE_SH7751_VIRTUAL_PCI_MEMIO */

    /* PCIC registers */
    {
    (void *) PCI_MSTR_PCIC_LOCAL,
    (void *) PCI_MSTR_PCIC_LOCAL,
    0x1000,
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
    VM_STATE_VALID      | VM_STATE_WRITABLE      | VM_STATE_CACHEABLE_NOT
    }
#endif /* INCLUDE_PCI */

    };

int   sysPhysMemDescNumEnt = NELEMENTS (sysPhysMemDesc);

#endif  /* defined(INCLUDE_MMU_BASIC) || defined(INCLUDE_MMU_FULL) */

int   sysCpu      = CPU;		/* system CPU type (SH7751)          */
char *sysBootLine = BOOT_LINE_ADRS;	/* address of boot line              */
char *sysExcMsg   = EXC_MSG_ADRS;	/* catastrophic message area         */
int   sysProcNum;			/* processor number of this CPU      */
int   sysFlags;				/* boot flags                        */
char  sysBootHost [BOOT_FIELD_LEN];	/* name of host from which we booted */
char  sysBootFile [BOOT_FIELD_LEN];	/* name of file from which we booted */


FUNCPTR sysCacheLibInit = (FUNCPTR) CACHE_ARCH_INIT_RTN;
UINT32  intEvtAdrs = INTEVT_ADRS;	/* INTEVT address (for SH7751) */

/* forward declarations */

void sysDelay (void);
void sysMsDelay (UINT);

#if defined(INCLUDE_WDB) || defined(INCLUDE_DEBUG)
LOCAL void sysUbcInit(UBC *);
#endif

/* driver includes */

#include "timer/sh7700Timer.c"
#include "sysSerial.c"

#ifdef INCLUDE_NVRAM_TO_FLASH
# include "mem/flashMem.c"
# include "mem/nvRamToFlash.c"
#else
# include "mem/nullNvRam.c"
#endif

#if defined(INCLUDE_NETWORK) && defined(INCLUDE_END)    /* network */

#if defined(INCLUDE_LN_97X_END)
STATUS  sysLan97xPciInit (void);
#endif /* INCLUDE_LN_97X_END */
 
#if defined(INCLUDE_KS884X_END)
STATUS  sysKs884xPciInit (void);
#endif /* INCLUDE_KS884X_END */

#endif /* INCLUDE_NETWORK && INCLUDE_END */

#ifdef INCLUDE_ATA
# include "sysAta.c"
#endif  /* INCLUDE_ATA */

#ifndef PRJ_BUILD

#ifdef INCLUDE_USB
# include "usbPciStub.c"
#endif /* INCLUDE_USB */

#endif /* PRJ_BUILD */

#ifdef INCLUDE_WINDML
#include "sysWindML.c"
#endif /* INCLUDE_WINDML */


#if defined (INCLUDE_PCI)
#  include "pci/pciConfigLib.c"		/* PCI config space access */
#  include "pci/pciIntLib.c"		/* PCI int support */
#  ifdef INCLUDE_SHOW_ROUTINES
#    include "pci/pciConfigShow.c"	/* PCI config space display */
#  endif /* INCLUDE_SHOW_ROUTINES */
#  ifdef INCLUDE_PCI_AUTOCONF		/* Use PCI autoconfig */
#    include "pci/pciAutoConfigLib.c"	/* automatic PCI configuration */
#    include "sysBusPci.c"		/* pciAutoConfig BSP support file */
#    include "aliM1543CPibc.c"		/* PCI-ISA bridge */
#  endif /*  INCLUDE_PCI_AUTOCONF */
#endif /* INCLUDE_PCI */

#ifdef INCLUDE_NETWORK
#  include "./sysNet.c"			/* network setup */
#  ifdef INCLUDE_END
#    ifdef INCLUDE_LN_97X_END
#      include "sysLn97xEnd.c"
#    endif /* INCLUDE_LN_97X_END */
#    ifdef INCLUDE_KS884X_END
#      include "sysKS884xEnd.c"
#    endif /* INCLUDE_KS884X_END */
#  endif /* INCLUDE_END */
#endif  /* INCLUDE_NETWORK */


/* external declarations */
 
extern void sysPciConnect();

 
/*******************************************************************************
*
* sysModel - return the model name of the CPU board
*
* This routine returns the model name of the CPU board.
*
* RETURNS: A pointer to the string "HITACHI_ULSI MS7751SE".
*/

char *sysModel (void)
    {
    return (SYS_MODEL);
    }

/*******************************************************************************
*
* sysAbortInt - handle ABORT interrupt
*/

LOCAL void sysAbortInt (void)
    {
    sysToMonitor (BOOT_WARM_NO_AUTOBOOT);		/* see sysLib.h */
    }

/*******************************************************************************
*
* sysBspRev - return the BSP version and revision number
*
* This routine returns a pointer to a BSP version and revision number, for
* example, 1.1/0. BSP_REV is concatenated to BSP_VERSION and returned.
*
* RETURNS: A pointer to the BSP version/revision string.
*/

char * sysBspRev (void)
    {
    return (BSP_VERSION BSP_REV);
    }

/******************************************************************************
*
* sysLedSet - Write to the LED register
*
*
* This routine writes one byte into the LED register
*
* RETURNS: N/A
*
* SEE ALSO: sysLib
*/

void sysLedSet
    (
    UINT8	value
    )
    {
    /* Register access is 16 bit */
    
    *(volatile UINT16 *) LED_REG_ADRS = value << 8;
    }

/*******************************************************************************
*
* sysHwInit - initialize the system hardware
*
* This routine initializes various features of the MS7751SE board.
* It is called from usrInit() in usrConfig.c, with interrupts locked out.
* Also multi-tasking kernel is not started.
*
* NOTE: This routine should not be called directly by the user.
*
* RETURNS: N/A
*/

void sysHwInit (void)
    {

    /* Set LED pattern 1 */
    
    sysLedSet(LED_PATTERN1);


    /* ===================================================================== */
    /* == Initialize SH7751 on-chip int sources (see h/arch/sh/sh7751.h)   = */
    /* ===================================================================== */

    *WDT_WTCSR_WR          = (UINT16)(WTCSR_TIMER_DISABLE | WTCSR_PASSWD);

    *TMU_TOCR		   = (UINT8)TOCR_DEFAULT;	/* force default */
    *TMU_TSTR		   = (UINT8)TSTR_STOP_ALL;
    *TMU_TCR0		   = (UINT16)TCR_DEFAULT;
    *TMU_TCR1		   = (UINT16)TCR_DEFAULT;
    *TMU_TCR2		   = (UINT16)TCR_DEFAULT;

    *SCI_SCR1              = (UINT8)0x00;	/* disable serial1 interrupt */
    *SCI_SCMR              = (UINT8)0x00;	/* disable smart-card i/f */
    *SCIF_SCR2             = (UINT16)0x0000;	/* disable serial2 interrupt */

    *IOP_PCTR              = (UINT32)0x00000000;  /* reset I/O port control  */
    *IOP_SCSPTR1           = (UINT8)0x00;	  /* reset SCI port1 control */
    *IOP_SCSPTR2           = (UINT16)0x0000;	  /* reset SCI port2 control */

    *RTC_ALARM_SEC         = (UINT8)0x00;	/* disable second alarm */
    *RTC_ALARM_MIN         = (UINT8)0x00;	/* disable minute alarm */
    *RTC_ALARM_HOUR        = (UINT8)0x00;	/* disable hour alarm */
    *RTC_ALARM_DAY_OF_WEEK = (UINT8)0x00;	/* disable day-of-week alarm */
    *RTC_ALARM_DAY         = (UINT8)0x00;	/* disable day alarm */
    *RTC_ALARM_MONTH       = (UINT8)0x00;	/* disable month alarm */
    *RTC_CONTROL_1         = (UINT8)0x00;	/* disable carry/alarm ints */
    *RTC_CONTROL_2         = (UINT8)0x09;	/* reset RTC control */

    /* ===================================================================== */
    /* == Initialize SH7751 on-chip INTC  (see h/arch/sh/sh7751.h)        == */
    /* ===================================================================== */

    *INTC_ICR  = (UINT16)ICR_NMI_TRAIL_EDGE;	/* select NMI detect edge */
    *INTC_IPRA = (UINT16)0x0000;
    *INTC_IPRB = (UINT16)0x0000;
    *INTC_IPRC = (UINT16)0x0000;

#ifdef  INCLUDE_PCI
 
    /* ===================================================================== */
    /* == Initialize SH7751 PCI bus                                       == */
    /* ===================================================================== */

    /*  Initialize PCI configuration space support library. */
 
    if (pciConfigLibInit (PCI_MECHANISM_0,
                          (ULONG) sysPciConfigRead,
                          (ULONG) sysPciConfigWrite,
                          (ULONG) sysPciSpecialCycle)
	!= OK)
        {
        /*
         * Note, bailing to sysToMonitor() in here will probally cause
         * an endless loop but there is little alternative at this stage.
         * Watch the bus access lights to help detect this condition.
         */
        sysToMonitor (BOOT_NO_AUTOBOOT);  /* BAIL */
        }
 
    /*  Initialize PCI interrupt library. */
 
    if ((pciIntLibInit ()) != OK)
        {
        sysToMonitor (BOOT_NO_AUTOBOOT);
        }
 
 
    sysPciAutoConfig();         /* The host bridge is init'd herein */

    /* 
     * Disable and clear RTC-6593 interrupts. The RTC is accessed through
     * the M1543C RTC interface.
     */

    sysRTCWrite (11, 0);	/* Ctrl. Reg. B - disable all ints. */
    sysRTCRead  (12);		/* Ctrl. Reg. C - clear int. status */

#endif  /* INCLUDE_PCI */

#ifdef INCLUDE_ATA
    (void) sysAtaInit();	/* Initialize ATA Interface */
#endif

    sysSerialHwInit();          /* Initialize serial devices */

#if defined(INCLUDE_WDB) || defined(INCLUDE_DEBUG)
    _func_wdbUbcInit = sysUbcInit;
#endif

    /*
     * vxPowerModeRegs is used in vxPowerModeSet():vxLib.c
     */

    vxPowerModeRegs.pSTBCR1 = STBCR;
    vxPowerModeRegs.pSTBCR2 = STBCR2;
    vxPowerModeRegs.pSTBCR3 = NULL;

    /*
     * The power management mode is initialized here. Reduced power mode
     * is activated only when the kernel is idle (cf vxPowerDown).
     * Power management mode is selected via vxPowerModeSet().
     * DEFAULT_POWER_MGT_MODE is defined in config.h.
     */

    vxPowerModeSet (DEFAULT_POWER_MGT_MODE);
    }

/*******************************************************************************
*
* sysHwInit2 - additional system configuration and initialization
*
* This routine connects system interrupts and does any additional configuration
* necessary.  This routine is called from sysClkConnect() in the timer
* driver sh7700Timer.c. See usrRoot() for more details.
*
* RETURNS: N/A
*/

void sysHwInit2 (void)
    {
    /* Set LED pattern 2 */
    sysLedSet(LED_PATTERN2);

    /* initialize aux clock interrupts */
    sysAuxClkConnect ((FUNCPTR)NULL, 0);

    /* connect ABORT interrupt */
    intConnect (INT_VEC_ABORT, sysAbortInt, 0);

#ifdef INCLUDE_PCI
    /* Connect PCIC error handlers */
    sysPcicHwInit2();

    /* 
     * Connect the PIBC interrupt controller handler. PCI interrupts are
     * coming through the PICs.
     */
    sysPibcHwInit2 ();
#endif

#ifdef INCLUDE_NETWORK
    sysNetHwInit2 ();
#endif

    /* Initialize serial interface(s) */
    sysSerialHwInit2 ();
    }

/*******************************************************************************
*
* sysPhysMemTop - get the address of the top of memory
*
* This routine returns the address of the first missing byte of memory,
* which indicates the top of memory.
*
* Normally, the user specifies the amount of physical memory with the
* macro LOCAL_MEM_SIZE in config.h.  BSPs that support run-time
* memory sizing do so only if the macro LOCAL_MEM_AUTOSIZE is defined.
* If not defined, then LOCAL_MEM_SIZE is assumed to be, and must be, the
* true size of physical memory.
*
* NOTE: Do no adjust LOCAL_MEM_SIZE to reserve memory for application
* use.  See sysMemTop() for more information on reserving memory.
*
* RETURNS: The address of the top of physical memory.
*
* SEE ALSO: sysMemTop()
*/

char * sysPhysMemTop (void)
    {
    static char * memTop = NULL;

    if (memTop == NULL)
	{
#ifdef LOCAL_MEM_AUTOSIZE

	/* If we could do autosizing, this is the place for it */
#       error "Memory autosizing is not supported with this BSP"

#else

	/* Don't do autosizing, if size is given */
	memTop = (char *)(LOCAL_MEM_LOCAL_ADRS + LOCAL_MEM_SIZE);

#endif /*LOCAL_MEM_AUTOSIZE*/
	}
    
    return memTop;
    }

/*******************************************************************************
*
* sysMemTop - get the address of the top of logical memory
*
* This routine returns the address of the first unusable byte of memory.
* VxWorks will not use any memory at or above this address.
*
* The user can reserve local memory from the board by declaring the
* macro USER_RESERVED_MEM with the amount of memory to reserve. This
* routine will return a pointer to the first byte of the reserved memory
* area.
*
* RETURNS: The address of the top of usable memory.
*/

char *sysMemTop (void)
    {
    static char * memTop = NULL;

    if (memTop == NULL)
	{
	memTop = sysPhysMemTop () - USER_RESERVED_MEM;
	}

    return memTop;
    }

/*******************************************************************************
*
* sysToMonitor - transfer control to the ROM monitor
*
* This routine transfers control to the ROM monitor.  Normally, it is called
* only by reboot()--which services ^X--and by bus errors at interrupt level.
* However, in some circumstances, the user may wish to introduce a
* <startType> to enable special boot ROM facilities.
*
* RETURNS: Does not return.
*/

STATUS sysToMonitor
    (
    int startType    /* parameter passed to ROM to tell it how to boot */
    )
    {
    FUNCPTR pRom = (FUNCPTR) ROM_WARM_ADRS;

    /* disable the MMU */

    VM_ENABLE (FALSE);

    (*pRom) (startType);

    return (OK);	/* in case we ever continue from ROM monitor */
    }

/******************************************************************************
*
* sysProcNumSet - set the processor number
*
* This routine sets the processor number for the CPU board.  Processor numbers
* should be unique on a single backplane.
*
*
* RETURNS: N/A
*
* SEE ALSO: sysProcNumGet()
*/

void sysProcNumSet
    (
    int procNum		/* processor number */
    )
    {
    sysProcNum = procNum;
    }

/******************************************************************************
*
* sysProcNumGet - get the processor number
*
* This routine returns the processor number for the CPU board, which is set
* with sysProcNumSet().
*
* RETURNS: The processor number for the CPU board.
*
* SEE ALSO: sysProcNumSet()
*/

int sysProcNumGet (void)
    {
    return (sysProcNum);
    }

/******************************************************************************
*
* sysBusTas - test and set a location across the bus
*
* This routine simply calls vxTas()
*
* RETURNS: TRUE if the value had not been set (but is now), or FALSE if
* the value was set already.
*
* SEE ALSO: vxTas()
*/

BOOL sysBusTas
    (
    char *adrs		/* address to be tested and set */
    )
    {
    return (vxTas (adrs));
    }

#if defined(INCLUDE_WDB) || defined(INCLUDE_DEBUG)
/******************************************************************************
*
* sysUbcInit - Initialize UBC structure
*
*
* This routine is called when setting the first hardware breakpoint to
* initialize the User Break Controller structure and identify the UBC.
* The following UBC structure members should be set:
*    chanCnt - number of UBC channels (0-4)
*    brcrSize - identify the pUbc-> Supported values are:
*                 BRCR_NONE  - no UBC support
*                 BRCR_0_1   - no BRCR    , 1 channel  (SH7050, SH7000)
*                 BRCR_16_1  - 16-bit BRCR, 1 channel  (SH7055, SH7604)
*                 BRCR_16_2  - 16-bit BRCR, 2 channels (SH7751, SH7709)
*                 BRCR_32_2  - 32-bit BRCR, 2 channels (SH7729, SH7709A)
*                 BRCR_32_4  - 32-bit BRCR, 4 channles (SH7615)
*    brcrInit - BRCR value to initialize
*    pBRCR - address of the BRCR register
*    base[i] - up to four channel base address (user break register adrs)
*
*
* RETURNS: N/A
*
*/

LOCAL void sysUbcInit
    (
    UBC * pUbc
    )
    {
    pUbc->brcrSize = BRCR_16_2;
    pUbc->brcrInit = 0;
    pUbc->pBRCR    = (UINT32) UBC_BRCR;
    pUbc->base[0]  = (UINT32) UBC_BARA;
    pUbc->base[1]  = (UINT32) UBC_BARB;
    }
#endif /* defined(INCLUDE_WDB) || defined(INCLUDE_DEBUG) */

/******************************************************************************
*
* sysMsDelay - delay for the specified amount of time (MS)
*
* This routine will delay for the specified amount of time by counting
* TMU0 ticks.
*
* This routine is not dependent on a particular rollover value for
* the counter, it should work no matter what the rollover value is.
*
* A small amount of count may be lost at the rollover point resulting in
* the sysMsDelay() causing a slightly longer delay than requested. Interrupts 
* can also increase the time if this routine is called with interrupts enabled.
*
* This routine will produce incorrect results if the delay time requested
* causes overflow during the totalDelta calculation. In case of 33.33MHz 
* SYS_PCLK_FREQ, the maximim delay is about 64 seconds.
*
* RETURNS: N/A
*/
void sysMsDelay
    (
    UINT		delay		/* length of time in MS to delay */
    )
    {
    register UINT 	oldval;		/* counter value */
    register UINT 	newval;		/* counter value */
    register UINT 	delta;		/* delta */
    register UINT 	totalDelta;	/* Dec. delta for entire delay period */
    register UINT 	tckElapsed;	/* cumulative ticks */

    /* Calculate delta of ticks for desired elapsed time. */

#define TICK_COUNTER	TMU_TCNT0

    totalDelta = TICK_FREQ / 1000 * delay;	/* TICK_FREQ defined in */
						/* the timer driver */

    /*
     * Now keep grabbing counter value and incrementing "tckElapsed" until
     * we hit the desired delay value.  Compensate for the fact that we may
     * read the counter at 0xffffffff before the interrupt service
     * routine has a chance to set in the rollover value.
     */

    tckElapsed = 0;
    oldval = *TICK_COUNTER;
    while (tckElapsed < totalDelta)
        {
        newval = *TICK_COUNTER;
	delta = abs ((int)oldval - (int)newval);
        if (delta < 1000)
            tckElapsed += delta;			/* no rollover */
        else
            if (newval > oldval)
                tckElapsed += abs ((int)oldval);	/* rollover */
        oldval = newval;
        }
    }

/******************************************************************************
*
* sysDelay - delay for approximately one millisecond
*
* Delay for approximately one milli-second.
*
* RETURNS: N/A
*/

void sysDelay (void)
    {
    sysMsDelay (1);
    }


#ifdef IO_PORTABLE

/*******************************************************************************
*
* sysOutByte - write a byte to the passed address
*
* SEE ALSO:
*/
 
void sysOutByte
    (
    ULONG port,         /* port to be written to */
    UCHAR data          /* data to be copied */
    )
    {
    *(volatile UCHAR *)port = data;
    }
 
/*******************************************************************************
*
* sysOutWord - write a word to the passed address
*
* SEE ALSO:
*/
 
void sysOutWord
    (
    ULONG port,         /* port to be written to */
    UINT16 data         /* data to be copied */
    )
    {
    *(volatile UINT16 *)port = data;
    }
 
 
/*******************************************************************************
*
* sysOutLong - write a long to the passed address
*
* SEE ALSO:
*/
 
void sysOutLong
    (
    ULONG port,         /* port to be written to */
    ULONG data          /* data to be copied */
    )
    {
    *(volatile ULONG *)port = data;
    }
 
/*******************************************************************************
*
* sysInByte - read a byte from the passed address
*
* SEE ALSO:
*/
 
UCHAR sysInByte
    (
    ULONG port          /* port to be read from */
    )
    {
    return (*(volatile UCHAR *)port);
    }
 
/*******************************************************************************
*
* sysInWord - read a word from the passed address
*
* SEE ALSO:
*/
 
USHORT sysInWord
    (
    ULONG port          /* port to be read from */
    )
    {
    return (*(volatile USHORT *)port);
    }
 
/*******************************************************************************
*
* sysInLong - read a long from the passed address
*
* SEE ALSO:
*/
 
ULONG sysInLong
    (
    ULONG port          /* port to be read from */
    )
    {
    return (*(volatile ULONG *)port);
    }

#endif /* IO_PORTABLE */


/*****************************************************************************
*
* Each time VxWorks is rebooted, it starts assigning system-assigned
* port numbers beginning at 1024.  Thus, if vxWorks is rebooted two or
* more times in succession, the same port number will be re-used.
*
* This behavior causes problems when vxWorks is being booted from a
* remote FTP server (particularly one running Solaris), because port
* 1024 goes into a TIME_WAIT state on the server and cannot be reused
* until it times out, typically in 2-4 minutes.
*
* This hack reduces the likelihood of this happening by "wasting"
* a different number of system-assigned port numbers for each boot.
*/

void sysBindFix(void)
{
    UINT8		N;

    sysNvRamGet((char *) &N, 1, NV_OFF_BINDFIX);
    N -= 16;
    sysNvRamSet((char *) &N, 1, NV_OFF_BINDFIX);

    /* This is quite fast even when N=255 */

    while (N--) {
	int s;

	struct sockaddr_in saddr;
	saddr.sin_addr.s_addr = htonl(INADDR_ANY);
	saddr.sin_port = 0;
	saddr.sin_family = AF_INET;
	s = (int)socket(AF_INET, SOCK_STREAM, 0);
	bind(s, (struct sockaddr *) &saddr, sizeof(saddr));
	close(s);
    }
}

