/* sysLn97xEnd.c - system configuration module for ln97xEnd */
 
/* Copyright 1984 - 2001 Wind River Systems, Inc. */
 
/*
modification history
--------------------
01f,08jul02,h_k  fixed MAC address setting for little endian mode. (SPR
                 #79321)
01e,13may02,frf  cleared diab warnings.
01d,07mar02,kab  SPR 70817: *EndLoad returns NULL on failure
01c,23oct01,zl   added ln97xBusEndianess, required for the PCI_SWAP definition.
01b,29mar01,zl   written.
*/


/*
DESCRIPTION
This is the WRS-supplied configuration module for the VxWorks 
ln97xEnd (lnPci) END driver.  It performs the dynamic parameterization 
of the ln97xEnd driver.  This technique of 'just-in-time' 
parameterization allows driver parameter values to be declared 
as any other defined constants rather than as static strings. 
*/
 
#if (defined (INCLUDE_LN_97X_END) && defined (INCLUDE_NETWORK) \
     && defined (INCLUDE_END))

/* includes */

#include "vxWorks.h"
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "end.h"
#include "config.h"

#include "drv/end/ln97xEnd.h" 
 
/* defines */
 
#define LN97X_MAX_UNITS         4
#define LN973_APROM_SIZE	16
 
/* PCI vendor/device Ids and rev mask */
 
#define LN97X_PCI_VENDOR_ID     0x1022  /* AMD */
#define LN97X_PCI_DEVICE_ID     0x2000  /* PCI device ID */
#define LN97X_CSR3_VALUE        0       /* csr3 value */
#define LN97X_OFFSET		2	/* For proper IP header alignment */
#define LN97X_USR_FLAGS		0
 
/* EEPROM (NM93C46) commands, */

#define EEPROM_WEN		0x130	/* 10011xxxx */
#define EEPROM_WDS		0x100	/* 10000xxxx */
#define EEPROM_WRITE		0x140	/* 101aaaaaa */

#define EEN			0x10	/* BCR19 Eprom enable */
#define EECS			0x04	/* BCR19 Eprom chip-select */
#define EESK			0x02	/* BCR19 Eprom serial clock */

#define EEPROM_CS_ON()	ln97xBcrWrite (pDrvCtrl, 19, (EEN | EECS))
#define EEPROM_CS_OFF()	ln97xBcrWrite (pDrvCtrl, 19, EEN)
#define SYS_MS_DELAY(t)	sysMsDelay (t)

#define EEPROM_DCLOCK(data)						\
	{								\
	SYS_MS_DELAY (1);						\
	ln97xBcrWrite (pDrvCtrl, 19, (data) | EEN | EECS);		\
	SYS_MS_DELAY (1);						\
	ln97xBcrWrite (pDrvCtrl, 19, (data) | EEN | EECS | EESK);	\
	SYS_MS_DELAY (1);						\
	ln97xBcrWrite (pDrvCtrl, 19, (data) | EEN | EECS);		\
	}

/* Forward declarations */
 
LOCAL void ln97xBcrWrite (LN_97X_DRV_CTRL * pDrvCtrl, int reg, UINT32 value);
LOCAL void epromWrite (LN_97X_DRV_CTRL * pDrvCtrl, UINT16 adrs, UINT16 word);

/* externs */

IMPORT END_OBJ* ln97xEndLoad (char *);

/* typedefs */
 
typedef struct ln97xPciRsrc             /* LN97X_PCI_RSRC */
    {
    UINT32      iobaseCsr;              /* Base Address Register 0 */
    UINT32      membaseCsr;             /* Base Address Register 1 */
    UINT32      vecNum;                 /* Interrupt Request Level */
    UINT32      intLevel;               /* Interrupt Request vector */
    UINT32      configType;             /* type of configuration */
    UINT32      pciBus;                 /* PCI Bus number */
    UINT32      pciDevice;              /* PCI Device number */
    UINT32      pciFunc;                /* PCI Function number */
    END_OBJ *	pEnd;			/* Pointer to END structure */
    } LN97X_PCI_RSRC;
  
/* locals */

LOCAL  int	ln97XUnits = 0;
 
/*
 * This array stores the board-specific PCI resources, the base address
 * register configuration mode. It's indexed using the device number returned
 * from pciFindDevice().
 */
 
LOCAL LN97X_PCI_RSRC ln97xPciRsrcs [LN97X_MAX_UNITS] =
    {
	{0, 0, 0, 0, PCI_CFG_AUTO, 0, 0, 0, NULL},
	{0, 0, 0, 0, PCI_CFG_AUTO, 0, 0, 0, NULL},
	{0, 0, 0, 0, PCI_CFG_AUTO, 0, 0, 0, NULL},
	{0, 0, 0, 0, PCI_CFG_AUTO, 0, 0, 0, NULL}
    };

#if _BYTE_ORDER == _BIG_ENDIAN
/* This is used in the PCI_SWAP definition */

LOCAL int ln97xBusEndianess = LN_LITTLE_ENDIAN;
#endif

/******************************************************************************
*
* sysLn97xEndLoad - load and create load string, for a ln97x (lnPci) device.
*
* This routine loads the lnPci device with initial parameters. 
*
* RETURNS: pointer to END object or NULL.
*
* SEE ALSO: ln97xEndLoad()
*/
 
END_OBJ * sysLn97xEndLoad
    (
    char * 		pParamStr,   /* ptr to initialization parameter str. */
    void * 		unused       /* unused optional argument */
    )
    {
    LN97X_PCI_RSRC * 	pRsrc;
    int    		unit;
    char * 		cp; 			
    char 		paramStr [END_INIT_STR_MAX];   /* from end.h */
    END_OBJ * 		pEnd;
    static char 	ln97xParamTemplate[] = 
				"0x%x:0x%x:0x%x:%d:%d:-1:-1:-1:0x%x:%d:0x%x";

    /*
     * The ln97xEnd driver END_LOAD_STRING should be:
     * <unit>:<devMemAddr>:<devIoAddr>:<pciMemBase:<vecNum>:<intLvl>:<memAdrs>:
     * <memSize>:<memWidth>:<csr3b>:<offset>:<flags>
     *
     * Note that ln97xEnd unit number is prepended in muxDevLoad, so we 
     * don't put it here!
     */

    if (strlen (pParamStr) == 0)
        {
        /* 
         * muxDevLoad() calls us twice.  If the string is
         * zero length, then this is the first time through
         * this routine, so we just return.
         */

        pEnd = ln97xEndLoad (pParamStr);
        }
    else
	{
        /*
         * On the second pass though here, we actually create 
         * the initialization parameter string on the fly.   
         * Note that we will be handed our unit number on the 
         * second pass through and we need to preserve that information.
         * So we use the unit number handed from the input string.
         */

	/* Get the unit number from the paramStr */

	unit = atoi (pParamStr);
	
	if ((unit < 0) || (unit >= LN97X_MAX_UNITS))
	    return ((END_OBJ *) NULL);

	/* pointer to resources */

	pRsrc = ln97xPciRsrcs + unit;

        cp = strcpy (paramStr, pParamStr); /* cp points to paramStr */

        /* Now, we advance cp, by finding the end the string */

        cp += strlen (paramStr);
        
        /* finish off the initialization parameter string */

	sprintf (cp, ln97xParamTemplate,
                 (UINT) PCI_MEMIO2LOCAL (pRsrc->membaseCsr),  /* memIo base */
                 (UINT) PCI_IO2LOCAL (pRsrc->iobaseCsr),  /* Io base */
                 (UINT) LOCAL2PCI_MEM (0),	/* pciMemBase */
                 pRsrc->vecNum,			/* interrupt IRQ vector */
                 pRsrc->intLevel,		/* interrupt level */
                 LN97X_CSR3_VALUE,		/* csr3 register value */
                 LN97X_OFFSET,			/* offset */
                 LN97X_USR_FLAGS		/* flags */
                );

        if ((pEnd = ln97xEndLoad (paramStr)) == (END_OBJ *)NULL)
	    {
            printf ("Error: device failed ln97xEndLoad routine.\n");
	    }

	pRsrc->pEnd = pEnd;
	}

    return (pEnd);
    }

 
/*******************************************************************************
*
* sysLan97xPciInit - prepare LAN adapter for LN97X initialization
*
* This routine finds out the PCI device, maps its memory and IO address.
* It must be done prior to initializing the LN970, sysLn97XInit().  Also
* must be done prior to MMU initialization, usrMmuInit().
*
* RETURNS: OK/ERROR
*/
 
STATUS sysLan97xPciInit (void)
    {
    UINT32              membaseCsr;
    UINT32              iobaseCsr;
    char                irq;
    int                 pciBus;
    int                 pciDevice;
    int                 pciFunc;
    int                 unit;
    int                 found = 0;
    LN97X_PCI_RSRC * 	pRsrc;

    /* for all the support lan devices find if some of them exist */
 
    for (unit = 0; unit < LN97X_MAX_UNITS; unit++)
        {
        if (pciFindDevice (LN97X_PCI_VENDOR_ID,
                           LN97X_PCI_DEVICE_ID,
                           unit, &pciBus, &pciDevice, &pciFunc) == OK)
            {
             /* board detected */
             /* for now, we're ignoring the possibility of duplicates */

             found = TRUE;
 
             /* load up the PCI device table */
 
             pRsrc = ln97xPciRsrcs + ln97XUnits; /* get the pci entry */
 
             pRsrc->pciBus    = pciBus;
             pRsrc->pciDevice = pciDevice;
             pRsrc->pciFunc   = pciFunc;
 
             ln97XUnits++;      /* number of units found */
             }
        }
 
    if ((found != TRUE) || (pciDevice > PCI_MAX_DEV))
        return (ERROR);
 
    /* Now initialize all the units we found */
 
    for (unit = 0; unit < ln97XUnits; unit++)
        {
        /* Fill in the resource entry */
 
        pRsrc = ln97xPciRsrcs + unit;
 
        /*
         * get memory base address and IO base address
         */
 
        pciConfigInLong (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                         PCI_CFG_BASE_ADDRESS_0, &iobaseCsr);
        pciConfigInLong (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                         PCI_CFG_BASE_ADDRESS_1, &membaseCsr);
        pciConfigInByte (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                         PCI_CFG_DEV_INT_LINE, (UINT8*) &irq);

        /* over write the resource table with values read */
 
        pRsrc->membaseCsr   = membaseCsr & PCI_MEMBASE_MASK;
        pRsrc->iobaseCsr    = iobaseCsr & PCI_IOBASE_MASK;
        pRsrc->vecNum       = IVEC_TO_INUM (PIC_IRQ_TO_IVEC(irq));
        pRsrc->intLevel     = irq;
 
        /* enable mapped memory and IO addresses */
 
        pciConfigOutWord (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                          PCI_CFG_COMMAND, PCI_CMD_IO_ENABLE |
                          PCI_CMD_MEM_ENABLE | PCI_CMD_MASTER_ENABLE);
 
        /* disable sleep mode */
 
        pciConfigOutWord (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
			  PCI_CFG_MODE, SLEEP_MODE_DIS);
        }
 
    return (OK);
    }


/*******************************************************************************
*
* sysLan97xIntEnable - enable Lan97x interrupts
*
* This routine enables ln7997x interrupts.  This may involve operations on
* interrupt control hardware.
*
* RETURNS: OK or ERROR for invalid arguments.
*/
 
STATUS sysLan97xIntEnable
    (
    int level           /* level number */
    )
    {
    /* level here means the PIC IRQ number */

    return (sysPicIntEnable(level));
    }
 
/*******************************************************************************
*
* sysLan97xIntDisable - disable ln970 interrupts
*
* This routine disables ln970 interrupts.  This may involve operations on
* interrupt control hardware.
*
* RETURNS: OK or ERROR for invalid arguments.
*/
 
STATUS sysLan97xIntDisable
    (
    int level           /* level number */
    )
    {
    /* level here means the PIC IRQ number */

    return (sysPicIntDisable(level));
    }
 
/*******************************************************************************
*
* sysLan97xEnetAddrGet - get Ethernet address
*
* This routine provides a target-specific interface for accessing a
* device Ethernet address.
*
* RETURNS: OK or ERROR if could not be obtained.
*/
 
STATUS sysLan97xEnetAddrGet
    (
    LN_97X_DRV_CTRL *  pDrvCtrl,   /* Driver control */
    char *             enetAdrs
    )
    {
    char        aprom [LN973_APROM_SIZE]; /* copy of address prom space */
    char *      ioaddr;
    int         ix;
 
    /* get IO address of unit */
    ioaddr = (char *)(pDrvCtrl->devAdrs);
 
    /* load aprom into an array */
    for (ix = 0; ix < LN973_APROM_SIZE; ix++)
        {
        aprom [ix] = sysInByte((int)(ioaddr + ix));
        }

    bcopy (aprom, enetAdrs, 6);

    /* check for W-s at end of list */
    if ((aprom [0xe] != 'W') || (aprom [0xf] != 'W'))
    {
#if (0)
        logMsg ("sysLn97XEnetAddrGet W's not stored in aprom\n",
                0, 1, 2, 3, 4, 5);
#endif
        return ERROR;
    }

    return (OK);
    }

/*******************************************************************************
*
* sysEnetAddrGet - get Ethernet address
*
* This routine provides a target-specific interface for accessing a
* device Ethernet address.
*
* RETURNS: OK or ERROR if could not be obtained.
*/
 
STATUS sysEnetAddrGet
    (
    int		unit,
    UINT8 *	addr
    )
    {
    LN_97X_DRV_CTRL *  pDrvCtrl;

    /* Check for valid unit number */

    if ((unit < 0) || (unit >= LN97X_MAX_UNITS))
	return (ERROR);

    /* Get the pointer to controller */

    pDrvCtrl = (LN_97X_DRV_CTRL *) ln97xPciRsrcs[unit].pEnd;
    
    if ((pDrvCtrl == NULL) || ((UINT)pDrvCtrl == ERROR))
	return (ERROR);

    if (sysLan97xEnetAddrGet (pDrvCtrl,(char*) addr) == OK)
	{
#if _BYTE_ORDER == _BIG_ENDIAN
	/* Reverse order of bytes */
	UCHAR c;
	c = addr[5]; addr[5] = addr[0]; addr[0] = c;
	c = addr[4]; addr[4] = addr[1]; addr[1] = c;
	c = addr[3]; addr[3] = addr[2]; addr[2] = c;
#endif
	return (OK);
	}
    else 
	{
	return (ERROR);
	}
    }

/*******************************************************************************
*
* sysEnetAddrSet - set Ethernet address
*
* This routine provides a target-specific interface for setting the device
* Ethernet address. This routine stops network activity and it should be
* followed by a hard reset.
*
* RETURNS: OK or ERROR if could not be obtained.
*/
 
STATUS sysEnetAddrSet
    (
    UCHAR byte5,			/* Ethernet address high order byte */
    UCHAR byte4,
    UCHAR byte3,
    UCHAR byte2,
    UCHAR byte1,
    UCHAR byte0				/* Ethernet address low order byte */
    )
    {
    LN_97X_DRV_CTRL *  pDrvCtrl;
 
    /* 
     * get IO address of unit. The API for sysEnetAddrSet() does not 
     * pass the unit number, use unit 0.
     */

    pDrvCtrl = (LN_97X_DRV_CTRL *) ln97xPciRsrcs[0].pEnd;

    if ((pDrvCtrl == NULL) || ((UINT)pDrvCtrl == ERROR))
	return (ERROR);

    /* Enable 32 bit access */

    sysOutLong ((int)pDrvCtrl->pRdp, 0);

    /* Software reset */

    (void) sysInLong ((int)pDrvCtrl->pReset);

    /* Enable EEPROM interface */

    ln97xBcrWrite (pDrvCtrl, 19, EEN);

    /* Write Enet address to NM93C46 EEPROM */

    epromWrite (pDrvCtrl, 0, (byte4 << 8) | byte5);
    epromWrite (pDrvCtrl, 1, (byte2 << 8) | byte3);
    epromWrite (pDrvCtrl, 2, (byte0 << 8) | byte1);
    epromWrite (pDrvCtrl, 7, ('W'   << 8) | 'W');

    ln97xBcrWrite (pDrvCtrl, 19, 0);	/* Disable EEPROM interface */
    
    printf ("Reset the board now to use the new Ethernet Address \n");

    return (OK);
    }

/*******************************************************************************
*
* ln97xBcrWrite - select and write a BCR register
*
* This routine writes the bus configuration register. It first selects the
* BCR register to write through the RAP register and then it writes the value
* to the BDP register.
*
* RETURNS: N/A
*/

LOCAL void ln97xBcrWrite
    (
    LN_97X_DRV_CTRL * 	pDrvCtrl,
    int 		reg,
    UINT32 		value
    )
    {
    int 		level;

    reg &= 0xff;
    value &=0xffff;

    level = intLock ();
    sysOutLong ((int)pDrvCtrl->pRap, PCI_SWAP(reg));
    sysOutLong ((int)pDrvCtrl->pBdp, PCI_SWAP(value));
    intUnlock (level);
    }

/*******************************************************************************
*
* epromWrite - Write a word to the EEPROM
*
* This routine writes one word to the EEPROM through the Ethernet interface.
* The Ethernet must be stopped before calling this routine.
*
* RETURNS: N/A.
*/
 
LOCAL void epromWrite
    (
    LN_97X_DRV_CTRL *  pDrvCtrl,
    UINT16 adrs,
    UINT16 word
    )
    {
    register int ix;

    /* Enable write */

    EEPROM_CS_ON();				/* CS=1 */
    for (ix = 8; ix >= 0; ix--)
	{
        EEPROM_DCLOCK ((EEPROM_WEN >> ix) & 0x0001);
	}
    
    EEPROM_CS_OFF();				/* CS=0 */
    SYS_MS_DELAY (1);

    /* Send write command with address */

    EEPROM_CS_ON();				/* CS=1 */
    for (ix = 8; ix >= 0; ix--)
	{
        EEPROM_DCLOCK (((EEPROM_WRITE + adrs) >> ix) & 0x0001);
	}

    /* Send Data bits */

    for (ix = 15; ix >=0 ; ix--)
	{
	EEPROM_DCLOCK ((word >> ix) & 0x0001);
	}

    EEPROM_CS_OFF();				/* CS=0 */
    SYS_MS_DELAY (20);				/* prg. takes max. 15 ms */

    /* Disable write */

    EEPROM_CS_ON();				/* CS=1 */
    for (ix = 8; ix >= 0; ix--)
	{
        EEPROM_DCLOCK ((EEPROM_WDS >> ix) & 0x0001);
	}

    EEPROM_CS_OFF();				/* CS=0 */
    SYS_MS_DELAY (1);
    }

#endif /* defined (INCLUDE_LN_97X_END) && defined (INCLUDE_NETWORK) */
