/* sysSerial.c - Hitachi MS7751SE01 BSP serial device initialization */

/* Copyright 1998-2001 Wind River Systems, Inc. */

/*
modification history
--------------------
01d,06mar02,h_k  added SCIF lsr2 and ss2r2 and intConnect for IV_SCIF_BRK (SPR
                 #73777).
01c,01nov01,zl   corrected NULL usage.
01b,23apr01,zl   added support for M1543C UARTs.
01a,14dec00,rsh  written.
*/

/*
DESCRIPTION
This module provides code to initialize the serial ports for the MS7750SE.
Two calls exist for the inialization, sysSerialHwInit() and sysSerialHwInit2().
They are called by sysHwInit() and sysHwInit2(), respectively in sysLib.c.
The routine sysSerialHwInit() sets up the necessary data structures and the
tranmit and receive buffers. sysSerialHwInit2() connects the serial interrupt
and enables it.

The routine sysSerialChanGet() is called by bootConfig.c and usrConfig.c as 
part of the generic serial driver setup (when INCLUDE_TTY_DEV is defined).
This instantiates a specific serial device configuration on the target into 
the generic TTY driver.
*/

#include "vxWorks.h"
#include "iv.h"
#include "intLib.h"
#include "config.h"
#include "sysLib.h"
#include "arch/sh/sh7750.h"
#include "arch/sh/sh7751.h"

#ifdef INCLUDE_SCI_SIO
#include "drv/sio/shSciSio.h"
#endif

#ifdef INCLUDE_SCIF_SIO
#include "drv/sio/shScifSio.h"
#endif

#ifdef INCLUDE_NS16550_SIO
#include "drv/sio/ns16552Sio.h"
#endif

#if defined (INCLUDE_SCI_SIO) || defined (INCLUDE_SCIF_SIO)
/*
 * BaudTable is a table of available baud rates, with the appropriate values
 * for smr2/brr2 registers.
 */

SCI_BAUD sysBaudTable [] =
    {
    {75,	SCI_SMR_CKS_75,		SCI_BRR_75},
    {110,	SCI_SMR_CKS_110,	SCI_BRR_110},
    {134,	SCI_SMR_CKS_134,	SCI_BRR_134},
    {150,	SCI_SMR_CKS_150,	SCI_BRR_150},
    {200,	SCI_SMR_CKS_200,	SCI_BRR_200},
    {300,	SCI_SMR_CKS_300,	SCI_BRR_300},
    {600,	SCI_SMR_CKS_600,	SCI_BRR_600},
    {1200,	SCI_SMR_CKS_1200,	SCI_BRR_1200},
    {1800,	SCI_SMR_CKS_1800,	SCI_BRR_1800},
    {2400,	SCI_SMR_CKS_2400,	SCI_BRR_2400},
    {3600,	SCI_SMR_CKS_3600,	SCI_BRR_3600},
    {4800,	SCI_SMR_CKS_4800,	SCI_BRR_4800},
    {7200,	SCI_SMR_CKS_7200,	SCI_BRR_7200},
    {9600,	SCI_SMR_CKS_9600,	SCI_BRR_9600},
    {19200,	SCI_SMR_CKS_19200,	SCI_BRR_19200},
    {31250,	SCI_SMR_CKS_31250,	SCI_BRR_31250},
    {38400,	SCI_SMR_CKS_38400,	SCI_BRR_38400},
    {57600,	SCI_SMR_CKS_57600,	SCI_BRR_57600},
    {115200,	SCI_SMR_CKS_115200,	SCI_BRR_115200},
    {500000,	SCI_SMR_CKS_500000,	SCI_BRR_500000},
    {0,		0,			0}
    };
#endif

#ifdef INCLUDE_SCI_SIO
static SCI_CHAN shSci;
#endif

#ifdef INCLUDE_SCIF_SIO
static SCIF_CHAN shScif;
#endif

#ifdef INCLUDE_NS16550_SIO
typedef struct
    {
    VOIDFUNCPTR* vector;			/* Interrupt vector */
    ULONG	 baseAdrs;			/* Register base address */
    USHORT	 regSpace;			/* Address Interval */
    USHORT	 irq;				/* PIC IRQ */
    } NS16550_CHAN_PARAS;

static NS16550_CHAN_PARAS devParas[NUM_NS16550_CHANNELS] = 
    {
    {PIC_IRQ_TO_IVEC(PIC_UART1),UART1_REG_BASE,SUPERIO_REG_INTERVAL,PIC_UART1}
#if (NUM_NS16550_CHANNELS == 2)
   ,{PIC_IRQ_TO_IVEC(PIC_UART2),UART2_REG_BASE,SUPERIO_REG_INTERVAL,PIC_UART2}
#endif
    };

static NS16550_CHAN ns16550Sio[NUM_NS16550_CHANNELS];

#endif /* INCLUDE_NS16550_SIO */

/* keep track of initialized SIOs */

static SIO_CHAN * sioChan[NUM_TTY];
static int sioChanCount = 0;

/******************************************************************************
*
* sysSerialHwInit - initialize the BSP serial devices to a quiescent state
*
* This routine initializes the BSP serial device descriptors and puts the
* devices in a quiescent state.  It is called from sysHwInit() with
* interrupts locked.
*
* RETURNS: N/A
*
* SEE ALSO: sysHwInit()
*/ 

void sysSerialHwInit (void)
    {
    int i;
#ifdef INCLUDE_SCI_SIO
    /* intialize SCI device descriptors */

    shSci.smr	= SCI_SMR1;
    shSci.brr	= SCI_BRR1;
    shSci.scr	= SCI_SCR1;
    shSci.tdr	= SCI_TDR1;
    shSci.ssr	= SCI_SSR1;
    shSci.rdr	= SCI_RDR1;
    shSci.baud	= 9600;

    /* initialize SCI */

    shSciDevInit (&shSci);
    
    sioChan[sioChanCount++] = (SIO_CHAN *) &shSci;
#endif

#ifdef INCLUDE_SCIF_SIO
    /* initialize SCIF device structure */

    shScif.smr2  = SCIF_SMR2;
    shScif.brr2  = SCIF_BRR2;
    shScif.scr2  = SCIF_SCR2;
    shScif.sfdr2 = SCIF_SFDR2;
    shScif.ssr2  = SCIF_SSR2;
    shScif.frdr2 = SCIF_FRDR2;
    shScif.fcr2  = SCIF_FCR2;
    shScif.fdr2  = SCIF_FDR2;
    shScif.lsr2  = SCIF_LSR2;
    shScif.ss2r2 = NULL;
    shScif.baud  = 9600;

    /* initialize SCIF */

    shScifDevInit (&shScif);

    sioChan[sioChanCount++] = (SIO_CHAN *) &shScif;
#endif

#ifdef INCLUDE_NS16550_SIO

    for (i = 0; i < NUM_NS16550_CHANNELS; i++)
	{
	/* initialize serial device driver */

	ns16550Sio[i].regs		= (UINT8 *) devParas[i].baseAdrs;
	ns16550Sio[i].level		= 0;	/* Not used */
	ns16550Sio[i].pad1		= 1;
	ns16550Sio[i].channelMode	= SIO_MODE_INT;
	ns16550Sio[i].regDelta		= devParas[i].regSpace;
	ns16550Sio[i].baudRate		= 9600;
	ns16550Sio[i].xtal		= (UINT32)1843200;

	/* reset the chip */
	ns16550DevInit (&ns16550Sio[i]);

	sioChan[sioChanCount++] = (SIO_CHAN *) &ns16550Sio[i];
	}

#endif /* INCLUDE_NS16550_SIO */
    }

/******************************************************************************
*
* sysSerialHwInit2 - connect BSP serial device interrupts
*
* This routine connects the BSP serial device interrupts.  It is called from
* sysHwInit2().
*
* Serial device interrupts cannot be connected in sysSerialHwInit() because
* the kernel memory allocator is not initialized at that point, and
* intConnect() calls malloc().
*
* RETURNS: N/A
*
* SEE ALSO: sysHwInit2()
*/ 

void sysSerialHwInit2 (void)
    {
    int i;
    
#ifdef INCLUDE_SCI_SIO
    /* intialize SCI interrupts */

    intConnect (IV_SCI_ERI, shSciIntErr, (int) &shSci);
    intConnect (IV_SCI_RXI, shSciIntRcv, (int) &shSci);
    intConnect (IV_SCI_TXI, shSciIntTx,  (int) &shSci);

    /*
     * enable serial interrupts, onto interrupt controller.
     * Do not enable actual TX/RX interrupts yet.
     */
    *INTC_IPRB = (UINT16)((*INTC_IPRB & 0xff0f) | ((INT_LVL_SCI & 0xf) << 4));

#endif /* INCLUDE_SCI_SIO */

#ifdef INCLUDE_SCIF_SIO
    /* intialize SCI interrupts */

    intConnect (IV_SCIF_ERI, shScifIntErr, (int) &shScif);
    intConnect (IV_SCIF_RXI, shScifIntRcv, (int) &shScif);
    intConnect (IV_SCIF_TXI, shScifIntTx,  (int) &shScif);
    intConnect (IV_SCIF_BRK, shScifIntErr, (int) &shScif);

    /*
     * enable serial interrupts, onto interrupt controller.
     * Do not enable actual TX/RX interrupts yet.
     */

    *INTC_IPRC = (*INTC_IPRC & 0xff0f) | ((INT_LVL_SCIF & 0xf) << 4);

#endif /* INCLUDE_SCIF_SIO */

#ifdef INCLUDE_NS16550_SIO
    /* intialize NS16550 interrupts */

    for (i = 0; i < NUM_NS16550_CHANNELS; i++)
	{
	intConnect (devParas[i].vector, ns16550Int, (int) &ns16550Sio[i]);
	sysPicIntEnable (devParas[i].irq);
	}

#endif /* INCLUDE_NS16550_SIO */
    }

/******************************************************************************
*
* sysSerialChanGet - get the SIO_CHAN device associated with a serial channel
*
* This routine gets the SIO_CHAN device associated with a specified serial
* channel.
*
* RETURNS: A pointer to the SIO_CHAN structure for the channel, or ERROR
* if the channel is invalid.
*/ 

SIO_CHAN * sysSerialChanGet
    (
    int channel		/* serial channel */
    )
    {

    if (channel < sioChanCount)
	return sioChan[channel];
    else
	return ((SIO_CHAN *) ERROR);
    }

