
/* sysWindML.h - WindML BSP specific routines  */

/* Copyright 1984-2002 Wind River Systems, Inc. */

#include "copyright_wrs.h"

/*
modification history
--------------------
01d,13jun02,zl   added type cast for ISA_BASE_LOCAL.
01c,17may02,zl   added pRegBase initialization.
01b,16nov01,jlb  major code review.
01a,08oct01,zl   written based on template sysWindML.c.
*/

/*
DESCRIPTION
This library provides board-specific routines to support WindML.

This API is designed in a general fashion, such that it is applicable to all
processor types and is bus independent.  It provides the support equally for
graphics devices that are integral to a processor (such as with the PPC823)
and graphics devices that are present on a PCI bus.

The API provides support for configurations where a system has multiple
graphics devices and input devices.

A data structure allocated within the BSP provides information describing the
graphics, input (keyboard and pointer), and audio devices.  A serial pointer
that connects to a standard serial port (such as, /tyCo/0) is not covered by
this API.  Those devices use the standard serial drivers.

The data structure to define the graphics device is as follows:

\cs
 typedef struct windml_device
     {
     UINT32      vendorID,          /@ vendor ID @/
     UINT32      deviceID,          /@ device ID @/
     UINT32      instance,          /@ instance of device @/
     UINT32      devType,           /@ type of input device @/
     UINT32      busType;           /@ type of bus @/
     UINT32      regDelta;          /@ distance between adjacent registers @/
     UINT32      intLevel;          /@ interrupt level to be used @/
     VOIDFUNCPTR * intVector;       /@ interrupt vector to be used @/
     void *      pPhysBaseAdrs0;    /@ first base address @/
     void *      pPhysBaseAdrs1;    /@ second base address @/
     void *      pPhysBaseAdrs2;    /@ third base address @/
     void *      pPhysBaseAdrs3;    /@ fourth base address @/
     void *      pPhysBaseAdrs4;    /@ fifth base address @/
     void *      pPhysBaseAdrs5;    /@ sixth base address @/
     void *      pRegBase;          /@ base address of register space @/
     } WINDML_DEVICE;

\ce

 The <vendorID> and the <deviceID> are based upon the PCI bus identifiers.  In
 this case, these identifiers are extended to include the mapping for non-PCI
 devices.  The file sysWindML.h provides the identifier for supported vendor
 and device identifiers.

 The above structure provides space for up to four memory segments that are
 used to access the device (for example, one segment for the frame buffer
 and another for the memory mapped registers).  Typically, a device
 will only have a single memory segment.  The size field identifies the number
 of bytes present within the memory segment.

 The <pRegBase> field identifies the base address to use to access the
 I/O ports.  This is typically the base of the ISA space.  For X86 type
 processors, this field would be set to 0.  For powerPC processors, this
 field would be set according to the memory model used (PRep or CHRP).

INCLUDE FILES: sysWindML.h
*/

/* includes */

#include "vxWorks.h"
#include "memLib.h"
#include "ugl/sysWindML.h"
#include "config.h"
#include "drv/pci/pciConfigLib.h"



/*******************************************************************************
*
* sysWindMLDevGet - configures the device
*
* This routine will determine the presence of the device and perform any
* device configuration required.  The behaviour of this function varies
* depending on the type of device configuration (such as a PCI device,
* integral, device controller, etc.).  A device configuration data
* structure is created for the specified device.  This configuration data
* defines the access mechanism for the device.
*
* The <vendorID> and <deviceID> identify the vendor and device identifiers
* in the PCI environment.  In the case of non-PCI type devices, these
* identifiers provide identifiers of the device outside of the range of PCI
* identifiers.  If these values are set to zero, then the <instance>
* occurrence of a device will be returned.  For example, if the <instance>
* and <vendorID> were set to zero, then the routine will return the first
* occurrence of a device.
*
* The returned data structure provides miscellaneous data items that
* describe the manner in which to access the device.
*
* RETURNS:
* When device has been successfully configured, the address of the
* device data structure; otherwise NULL
*/

WINDML_DEVICE * sysWindMLDevGet
    (
    UINT32 devType,     /* Type of device */
    UINT32 instance,    /* The instance of the device */
    UINT32 vendorID,    /* The identifier for the device manufacturer */
    UINT32 deviceID     /* The identifier for the device */
    )
    {
    WINDML_DEVICE *pDev = NULL;

    switch (devType)
	{
	case WINDML_GRAPHICS_DEVICE:
	    {
	    int busno, devno, funcno;
	    UINT16 data;
	    UINT32 bar = 0;

	    if (vendorID == 0 || deviceID == 0)
		{
		/* If no ID suplied, search for display class */

		int pciClass = PCI_CLASS_DISPLAY_CTLR << 16;

		if (pciFindClass (pciClass, instance, &busno, &devno, 
				  &funcno) != OK)
		    {
		    return (pDev);
		    }

	        if (pciConfigInWord (busno, devno, funcno, PCI_CFG_VENDOR_ID,
				     &data) != OK)
		    {
		    return (pDev);
		    }

		vendorID = data;

	        if (pciConfigInWord (busno, devno, funcno, PCI_CFG_DEVICE_ID,
				     &data) != OK)
		    {
		    return (pDev);
		    }

		deviceID = data;
		}
	    else
		{
		/* Find instance of PCI device */

		if (pciFindDevice (vendorID, deviceID, instance, 
				   &busno, &devno, &funcno) != OK) 
		    {
		    return (pDev);
		    }
		}

	    /* allocate space for device structure */

	    pDev = (WINDML_DEVICE *) calloc(1, sizeof(WINDML_DEVICE));

	    if (pDev == NULL)
		{
		return (pDev);
		}

	    /* board autoconfig'd - extract base address from config space */
	    
            pciConfigInLong (busno, devno, funcno, PCI_CFG_BASE_ADDRESS_0, &bar);
	    bar = (UINT32) PCI_MEMIO2LOCAL (bar & PCI_MEMBASE_MASK);

                   
            pDev->devType        = devType;
	    pDev->vendorID       = vendorID;
	    pDev->deviceID       = deviceID;
	    pDev->instance       = instance;
            pDev->busType        = BUS_TYPE_PCI;
	    pDev->pPhysBaseAdrs0 = (void *) bar;
	    pDev->pRegBase       = (void *) ISA_BASE_LOCAL;
	    break;
	    }

	case WINDML_KEYBOARD_DEVICE:
	    {
	    /* Allocate space for structure */

	    pDev = (WINDML_DEVICE *) calloc(1, sizeof(WINDML_DEVICE));
	    if (pDev == NULL)
		{
		return (pDev);
		}

    	    /* Board uses KBD controller of the M1543C SuperIO */

	    pDev->devType        = devType;
	    pDev->vendorID       = 0;
	    pDev->deviceID       = 0;
	    pDev->instance       = instance;
	    pDev->regDelta       = 4; 		/* IO adrs 60 and 64 */
	    pDev->intLevel       = PIC_KBD;
	    pDev->intVector      = (void *) PIC_IRQ_TO_IVEC(PIC_KBD);
	    pDev->pRegBase       = (void *) KBD_REG_BASE;
	    break;
	    }

	case WINDML_POINTER_DEVICE:
	    {
	    /* Allocate space for structure */

	    pDev = (WINDML_DEVICE *) calloc(1, sizeof(WINDML_DEVICE));
	    if (pDev == NULL)
		{
		return (pDev);
		}

	    /* Board uses Mouse controller of the M1543C SuperIO */

	    pDev->devType        = devType;
	    pDev->vendorID       = 0;
	    pDev->deviceID       = 0;
	    pDev->instance       = instance;
	    pDev->regDelta       = 4; 		/* IO adrs 60 and 64 */
	    pDev->intLevel       = PIC_MOUSE;
	    pDev->intVector      = (void *) PIC_IRQ_TO_IVEC(PIC_MOUSE);
	    pDev->pRegBase       = (void *) MOUSE_REG_BASE;
	    break;
	    }

	default:
	    {
	    /* Unsupported device type, return NULL */

	    break;
	    }
	}

    return (pDev);
    }


/*******************************************************************************
*
* sysWindMLDevCtrl - special control of the device mode
*
* This routine provides special control features for the device.  This
* function is essentially a catch all to provide control of the device
* where the functionality is provided within a PCI configuration header or
* by board specific registers which may be shared by other functions
* implemented on the target board.
*
* The <function> argument defines the type of function that is to be
* performed and the <pArg> parameter provides the details relating to the
* function.
*
* The values for <function> and the interpretation of the <pArg> parameters
* are:
*
*\is
*\i WINDML_ACCESS_MODE_SET
*       Sets the device's access mode as to whether it is to respond to I/O
*       cycles of memory mapped cycles or both.  The accessibility is
*       provided by the <pArg> parameter which is bit mapped containing the
*       flags WINDML_MEM_ENABLE (enable memory mapped access) and
*       WINDML_IO_ENABLE (enable I/O access).
*\i WINDML_LCD_MODE_SET
*       Sets the control mode for an LCD that is controllable by an on board
*       register rather than a graphics device register. The mode information
*       is passed through <pArg>. The flags available are WINDML_LCD_ON
*       WINDML_LCD_OFF, WINDML_BACKLIGHT_ON, WINDML_BACKLIGHT_OFF.
*\i WINDML_BUSWIDTH_SET
*       Some boards allow the LCD bus width to be changed dynamically via
*       an FPGA or other configurable logic. This can be done in a board
*       specific manner. The actual bus width will be passed through <pArg>.
*\i WINDML_PCI_MEMBASE_GET
*       Obtain the base address of CPU memory as seen by PCI devices.
*\ie
*
* RETURNS: OK when control operation was success; otherwise ERROR
*/

STATUS sysWindMLDevCtrl
    (
    WINDML_DEVICE *pDev,        /* device to control */
    int function,               /* Type of operation to perform */
    int *pArg                   /* Control mode */
    )
    {
    STATUS status = ERROR;

    if (pDev == NULL)
	{
	return(status);
	}

    switch (pDev->devType)
	{
	case WINDML_GRAPHICS_DEVICE:
	    {
	    switch (function)
		{
		case WINDML_ACCESS_MODE_SET:
		    {
		    int busno, devno, funcno;

		    if (pciFindDevice (pDev->vendorID, 
				       pDev->deviceID, 
				       pDev->instance, 
				       &busno, &devno, &funcno) != OK) 
			{
			return (ERROR);
			}

		    status = pciConfigOutWord (busno, devno, funcno, 
					       PCI_CFG_COMMAND, *pArg);
		    break;
		    }

                /* Obtain the CPU memory base address as seen by PCI device */
                case WINDML_PCI_MEMBASE_GET:
                    /* PCI memory base is same as CPU, so set to 0 */
                    status = OK;
                    *pArg = 0;
                    break;

		default:
		    {
		    break;
		    }
		}
	    }
	default:
	    {
	    break;
	    }
	}
    return(status);
    }


/*******************************************************************************
*
* sysWindMLDevRelease - release a device configuration
*
* This routine will release any resources that were allocated when a
* device was configured using the <sysWindMLDevGet()> function.  This
* function will free the memory that was allocated for the WINDML_DEVICE
* data structure if it was dynamically allocated.  If the data structure
* was not dynamically allocated, this function will usually be simply a
* stub.
*
* RETURNS: OK when operation was success; otherwise ERROR
*/

STATUS sysWindMLDevRelease
    (
    WINDML_DEVICE *pDev         /* Device to release */
    )
    {
    if (pDev != NULL)
        {
        free ((char *) pDev);
        }

    return (OK);
    }


/*******************************************************************************
*
* sysWindMLIntConnect - Connect the device interrupt.
*
* This routine connects a routine to the interrupt.
*
* RETURNS: OK or ERROR
*/

STATUS sysWindMLIntConnect
    (
    WINDML_DEVICE * pDev,   /* Device to control */
    VOIDFUNCPTR     routine,      /* routine to be called */
    int             parameter     /* parameter to be passed to routine */
    )
    {
    STATUS status = ERROR;

    if (pDev != NULL && pDev->intVector != NULL && routine != NULL)
        {
        if (pDev->busType == BUS_TYPE_PCI)
            status = pciIntConnect(pDev->intVector, routine, parameter);
        else
            status = intConnect (pDev->intVector, routine, parameter);
        }
    return(status);
    }

/******************************************************************************
*
* sysWindMLIntEnable - Enable device interrupts.
*
* This routine enables the interrupt for the device.
*
* RETURNS: OK or ERROR
*/

STATUS sysWindMLIntEnable
    (
    WINDML_DEVICE * pDev         /* Device to control */
    )
    {

    STATUS status = ERROR;

    if (pDev != NULL && pDev->intLevel != 0)
        /* Enable the PIC IRQ */
        status = sysPicIntEnable (pDev->intLevel);

    return status;
    }

/******************************************************************************
*
* sysWindMLIntDisable - Disable device interrupts.
*
* This routine disables the interrupt for the device.
*
* RETURNS: OK or ERROR
*/

STATUS sysWindMLIntDisable
    (
    WINDML_DEVICE * pDev         /* Device to control */
    )
    {

    STATUS status = ERROR;

    if (pDev != NULL && pDev->intLevel != 0)
        /* Disable the PIC IRQ */
        status = sysPicIntDisable (pDev->intLevel);

    return(status);
    }

